/*
 * cellfs.h - CELLFS local header file.
 *
 * Copyright (C) 2005-2008 Nippon Telegraph and Telephone Corporation.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 *
 * Written by Koji Sato <koji@osrg.net>
 *            Ryusuke Konishi <ryusuke@osrg.net>
 */

#ifndef _CELLFS_H
#define _CELLFS_H

#include <linux/kernel.h>
#include <linux/buffer_head.h>
#include <linux/spinlock.h>
#include <linux/blkdev.h>
#include <linux/cellfs_fs.h>
#include "the_cellfs.h"
#include "bmap.h"

/**
 * struct cellfs_inode_info - cellfs inode data in memory
 * @i_flags: inode flags
 * @i_state: dynamic state flags
 * @i_bmap: pointer on i_bmap_data
 * @i_bmap_data: raw block mapping
 * @i_xattr: <TODO>
 * @i_dir_start_lookup: page index of last successful search
 * @i_cno: checkpoint number for GC inode
 * @i_btnode_cache: cached pages of b-tree nodes
 * @i_dirty: list for connecting dirty files
 * @xattr_sem: semaphore for extended attributes processing
 * @i_bh: buffer contains disk inode
 * @i_root: root object of the current filesystem tree
 * @vfs_inode: VFS inode object
 */
struct cellfs_inode_info {
	__u32 i_flags;
	unsigned long  i_state;		/* Dynamic state flags */
	struct cellfs_bmap *i_bmap;
	struct cellfs_bmap i_bmap_data;
	__u64 i_xattr;	/* sector_t ??? */
	__u32 i_dir_start_lookup;
	__u64 i_cno;		/* check point number for GC inode */
	struct address_space i_btnode_cache;
	struct list_head i_dirty;	/* List for connecting dirty files */

#ifdef CONFIG_CELLFS_XATTR
	/*
	 * Extended attributes can be read independently of the main file
	 * data. Taking i_sem even when reading would cause contention
	 * between readers of EAs and writers of regular file data, so
	 * instead we synchronize on xattr_sem when reading or changing
	 * EAs.
	 */
	struct rw_semaphore xattr_sem;
#endif
	struct buffer_head *i_bh;	/* i_bh contains a new or dirty
					   disk inode */
	struct cellfs_root *i_root;
	struct inode vfs_inode;
};

static inline struct cellfs_inode_info *CELLFS_I(const struct inode *inode)
{
	return container_of(inode, struct cellfs_inode_info, vfs_inode);
}

static inline struct cellfs_inode_info *
CELLFS_BMAP_I(const struct cellfs_bmap *bmap)
{
	return container_of(bmap, struct cellfs_inode_info, i_bmap_data);
}

static inline struct inode *CELLFS_BTNC_I(struct address_space *btnc)
{
	struct cellfs_inode_info *ii =
		container_of(btnc, struct cellfs_inode_info, i_btnode_cache);
	return &ii->vfs_inode;
}

/*
 * Dynamic state flags of CELLFS on-memory inode (i_state)
 */
enum {
	CELLFS_I_NEW = 0,		/* Inode is newly created */
	CELLFS_I_DIRTY,			/* The file is dirty */
	CELLFS_I_QUEUED,			/* inode is in dirty_files list */
	CELLFS_I_BUSY,			/* inode is grabbed by a segment
					   constructor */
	CELLFS_I_COLLECTED,		/* All dirty blocks are collected */
	CELLFS_I_UPDATED,		/* The file has been written back */
	CELLFS_I_INODE_DIRTY,		/* write_inode is requested */
	CELLFS_I_BMAP,			/* has bmap and btnode_cache */
	CELLFS_I_GCINODE,		/* inode for GC, on memory only */
};

/*
 * commit flags for cellfs_commit_super and cellfs_sync_super
 */
enum {
	CELLFS_SB_COMMIT = 0,	/* Commit a super block alternately */
	CELLFS_SB_COMMIT_ALL	/* Commit both super blocks */
};

/*
 * Macros to check inode numbers
 */
#define CELLFS_MDT_INO_BITS   \
	((unsigned int)(1 << CELLFS_DAT_INO | 1 << CELLFS_CPFILE_INO |	\
			1 << CELLFS_SUFILE_INO | 1 << CELLFS_IFILE_INO |	\
			1 << CELLFS_ATIME_INO | 1 << CELLFS_SKETCH_INO))

#define CELLFS_SYS_INO_BITS   \
	((unsigned int)(1 << CELLFS_ROOT_INO) | CELLFS_MDT_INO_BITS)

#define CELLFS_FIRST_INO(sb) (((struct the_cellfs *)sb->s_fs_info)->ns_first_ino)

#define CELLFS_MDT_INODE(sb, ino) \
	((ino) < CELLFS_FIRST_INO(sb) && (CELLFS_MDT_INO_BITS & (1 << (ino))))
#define CELLFS_VALID_INODE(sb, ino) \
	((ino) >= CELLFS_FIRST_INO(sb) || (CELLFS_SYS_INO_BITS & (1 << (ino))))

/**
 * struct cellfs_transaction_info: context information for synchronization
 * @ti_magic: Magic number
 * @ti_save: Backup of journal_info field of task_struct
 * @ti_flags: Flags
 * @ti_count: Nest level
 * @ti_garbage:	List of inode to be put when releasing semaphore
 */
struct cellfs_transaction_info {
	u32			ti_magic;
	void		       *ti_save;
				/* This should never used. If this happens,
				   one of other filesystems has a bug. */
	unsigned short		ti_flags;
	unsigned short		ti_count;
	struct list_head	ti_garbage;
};

/* ti_magic */
#define CELLFS_TI_MAGIC		0xd9e392fb

/* ti_flags */
#define CELLFS_TI_DYNAMIC_ALLOC	0x0001  /* Allocated from slab */
#define CELLFS_TI_SYNC		0x0002	/* Force to construct segment at the
					   end of transaction. */
#define CELLFS_TI_GC		0x0004	/* GC context */
#define CELLFS_TI_COMMIT		0x0008	/* Change happened or not */
#define CELLFS_TI_WRITER		0x0010	/* Constructor context */


int cellfs_transaction_begin(struct super_block *,
			    struct cellfs_transaction_info *, int);
int cellfs_transaction_commit(struct super_block *);
void cellfs_transaction_abort(struct super_block *);

static inline void cellfs_set_transaction_flag(unsigned int flag)
{
	struct cellfs_transaction_info *ti = current->journal_info;

	ti->ti_flags |= flag;
}

static inline int cellfs_test_transaction_flag(unsigned int flag)
{
	struct cellfs_transaction_info *ti = current->journal_info;

	if (ti == NULL || ti->ti_magic != CELLFS_TI_MAGIC)
		return 0;
	return !!(ti->ti_flags & flag);
}

static inline int cellfs_doing_gc(void)
{
	return cellfs_test_transaction_flag(CELLFS_TI_GC);
}

static inline int cellfs_doing_construction(void)
{
	return cellfs_test_transaction_flag(CELLFS_TI_WRITER);
}

/*
 * function prototype
 */
#ifdef CONFIG_CELLFS_POSIX_ACL
#error "CELLFS: not yet supported POSIX ACL"
extern int cellfs_acl_chmod(struct inode *);
extern int cellfs_init_acl(struct inode *, struct inode *);
#else
static inline int cellfs_acl_chmod(struct inode *inode)
{
	return 0;
}

static inline int cellfs_init_acl(struct inode *inode, struct inode *dir)
{
	inode->i_mode &= ~current_umask();
	return 0;
}
#endif

#define CELLFS_ATIME_DISABLE

/* Flags that should be inherited by new inodes from their parent. */
#define CELLFS_FL_INHERITED						\
	(FS_SECRM_FL | FS_UNRM_FL | FS_COMPR_FL | FS_SYNC_FL |		\
	 FS_IMMUTABLE_FL | FS_APPEND_FL | FS_NODUMP_FL | FS_NOATIME_FL |\
	 FS_COMPRBLK_FL | FS_NOCOMP_FL | FS_NOTAIL_FL | FS_DIRSYNC_FL)

/* Mask out flags that are inappropriate for the given type of inode. */
static inline __u32 cellfs_mask_flags(umode_t mode, __u32 flags)
{
	if (S_ISDIR(mode))
		return flags;
	else if (S_ISREG(mode))
		return flags & ~(FS_DIRSYNC_FL | FS_TOPDIR_FL);
	else
		return flags & (FS_NODUMP_FL | FS_NOATIME_FL);
}

/* dir.c */
extern int cellfs_add_link(struct dentry *, struct inode *);
extern ino_t cellfs_inode_by_name(struct inode *, const struct qstr *);
extern int cellfs_make_empty(struct inode *, struct inode *);
extern struct cellfs_dir_entry *
cellfs_find_entry(struct inode *, const struct qstr *, struct page **);
extern int cellfs_delete_entry(struct cellfs_dir_entry *, struct page *);
extern int cellfs_empty_dir(struct inode *);
extern struct cellfs_dir_entry *cellfs_dotdot(struct inode *, struct page **);
extern void cellfs_set_link(struct inode *, struct cellfs_dir_entry *,
			   struct page *, struct inode *);

/* file.c */
extern int cellfs_sync_file(struct file *, loff_t, loff_t, int);

/* ioctl.c */
long cellfs_ioctl(struct file *, unsigned int, unsigned long);
long cellfs_compat_ioctl(struct file *file, unsigned int cmd, unsigned long arg);
int cellfs_ioctl_prepare_clean_segments(struct the_cellfs *, struct cellfs_argv *,
				       void **);

/* inode.c */
void cellfs_inode_add_blocks(struct inode *inode, int n);
void cellfs_inode_sub_blocks(struct inode *inode, int n);
extern struct inode *cellfs_new_inode(struct inode *, umode_t);
extern void cellfs_free_inode(struct inode *);
extern int cellfs_get_block(struct inode *, sector_t, struct buffer_head *, int);
extern void cellfs_set_inode_flags(struct inode *);
extern int cellfs_read_inode_common(struct inode *, struct cellfs_inode *);
extern void cellfs_write_inode_common(struct inode *, struct cellfs_inode *, int);
struct inode *cellfs_ilookup(struct super_block *sb, struct cellfs_root *root,
			    unsigned long ino);
struct inode *cellfs_iget_locked(struct super_block *sb, struct cellfs_root *root,
				unsigned long ino);
struct inode *cellfs_iget(struct super_block *sb, struct cellfs_root *root,
			 unsigned long ino);
extern struct inode *cellfs_iget_for_gc(struct super_block *sb,
				       unsigned long ino, __u64 cno);
extern void cellfs_update_inode(struct inode *, struct buffer_head *);
extern void cellfs_truncate(struct inode *);
extern void cellfs_evict_inode(struct inode *);
extern int cellfs_setattr(struct dentry *, struct iattr *);
extern void cellfs_write_failed(struct address_space *mapping, loff_t to);
int cellfs_permission(struct inode *inode, int mask);
int cellfs_load_inode_block(struct inode *inode, struct buffer_head **pbh);
extern int cellfs_inode_dirty(struct inode *);
int cellfs_set_file_dirty(struct inode *inode, unsigned nr_dirty);
extern int cellfs_mark_inode_dirty(struct inode *);
extern void cellfs_dirty_inode(struct inode *, int flags);
int cellfs_fiemap(struct inode *inode, struct fiemap_extent_info *fieinfo,
		 __u64 start, __u64 len);

/* super.c */
extern struct inode *cellfs_alloc_inode(struct super_block *);
extern void cellfs_destroy_inode(struct inode *);
extern __printf(3, 4)
void cellfs_error(struct super_block *, const char *, const char *, ...);
extern __printf(3, 4)
void cellfs_warning(struct super_block *, const char *, const char *, ...);
extern struct cellfs_super_block *
cellfs_read_super_block(struct super_block *, u64, int, struct buffer_head **);
extern int cellfs_store_magic_and_option(struct super_block *,
					struct cellfs_super_block *, char *);
extern int cellfs_check_feature_compatibility(struct super_block *,
					     struct cellfs_super_block *);
extern void cellfs_set_log_cursor(struct cellfs_super_block *,
				 struct the_cellfs *);
struct cellfs_super_block **cellfs_prepare_super(struct super_block *sb,
					       int flip);
int cellfs_commit_super(struct super_block *sb, int flag);
int cellfs_cleanup_super(struct super_block *sb);
int cellfs_resize_fs(struct super_block *sb, __u64 newsize);
int cellfs_attach_checkpoint(struct super_block *sb, __u64 cno, int curr_mnt,
			    struct cellfs_root **root);
int cellfs_checkpoint_is_mounted(struct super_block *sb, __u64 cno);

/* gcinode.c */
int cellfs_gccache_submit_read_data(struct inode *, sector_t, sector_t, __u64,
				   struct buffer_head **);
int cellfs_gccache_submit_read_node(struct inode *, sector_t, __u64,
				   struct buffer_head **);
int cellfs_gccache_wait_and_mark_dirty(struct buffer_head *);
int cellfs_init_gcinode(struct inode *inode);
void cellfs_remove_all_gcinodes(struct the_cellfs *cellfs);

/*
 * Inodes and files operations
 */
extern const struct file_operations cellfs_dir_operations;
extern const struct inode_operations cellfs_file_inode_operations;
extern const struct file_operations cellfs_file_operations;
extern const struct address_space_operations cellfs_aops;
extern const struct inode_operations cellfs_dir_inode_operations;
extern const struct inode_operations cellfs_special_inode_operations;
extern const struct inode_operations cellfs_symlink_inode_operations;

/*
 * filesystem type
 */
extern struct file_system_type cellfs_fs_type;


#endif	/* _CELLFS_H */
