/*
 * sysfs.h - sysfs support declarations.
 *
 * Copyright (C) 2005-2014 Nippon Telegraph and Telephone Corporation.
 * Copyright (C) 2014 HGST, Inc., a Western Digital Company.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * Written by Vyacheslav Dubeyko <Vyacheslav.Dubeyko@hgst.com>
 */

#ifndef _CELLFS_SYSFS_H
#define _CELLFS_SYSFS_H

#include <linux/sysfs.h>

#define CELLFS_ROOT_GROUP_NAME	"cellfs"

/*
 * struct cellfs_sysfs_dev_subgroups - device subgroup kernel objects
 * @sg_superblock_kobj: /sys/fs/<cellfs>/<device>/superblock
 * @sg_superblock_kobj_unregister: completion state
 * @sg_segctor_kobj: /sys/fs/<cellfs>/<device>/segctor
 * @sg_segctor_kobj_unregister: completion state
 * @sg_mounted_snapshots_kobj: /sys/fs/<cellfs>/<device>/mounted_snapshots
 * @sg_mounted_snapshots_kobj_unregister: completion state
 * @sg_checkpoints_kobj: /sys/fs/<cellfs>/<device>/checkpoints
 * @sg_checkpoints_kobj_unregister: completion state
 * @sg_segments_kobj: /sys/fs/<cellfs>/<device>/segments
 * @sg_segments_kobj_unregister: completion state
 */
struct cellfs_sysfs_dev_subgroups {
	/* /sys/fs/<cellfs>/<device>/superblock */
	struct kobject sg_superblock_kobj;
	struct completion sg_superblock_kobj_unregister;

	/* /sys/fs/<cellfs>/<device>/segctor */
	struct kobject sg_segctor_kobj;
	struct completion sg_segctor_kobj_unregister;

	/* /sys/fs/<cellfs>/<device>/mounted_snapshots */
	struct kobject sg_mounted_snapshots_kobj;
	struct completion sg_mounted_snapshots_kobj_unregister;

	/* /sys/fs/<cellfs>/<device>/checkpoints */
	struct kobject sg_checkpoints_kobj;
	struct completion sg_checkpoints_kobj_unregister;

	/* /sys/fs/<cellfs>/<device>/segments */
	struct kobject sg_segments_kobj;
	struct completion sg_segments_kobj_unregister;
};

#define CELLFS_COMMON_ATTR_STRUCT(name) \
struct cellfs_##name##_attr { \
	struct attribute attr; \
	ssize_t (*show)(struct kobject *, struct attribute *, \
			char *); \
	ssize_t (*store)(struct kobject *, struct attribute *, \
			 const char *, size_t); \
}

CELLFS_COMMON_ATTR_STRUCT(feature);

#define CELLFS_DEV_ATTR_STRUCT(name) \
struct cellfs_##name##_attr { \
	struct attribute attr; \
	ssize_t (*show)(struct cellfs_##name##_attr *, struct the_cellfs *, \
			char *); \
	ssize_t (*store)(struct cellfs_##name##_attr *, struct the_cellfs *, \
			 const char *, size_t); \
}

CELLFS_DEV_ATTR_STRUCT(dev);
CELLFS_DEV_ATTR_STRUCT(segments);
CELLFS_DEV_ATTR_STRUCT(mounted_snapshots);
CELLFS_DEV_ATTR_STRUCT(checkpoints);
CELLFS_DEV_ATTR_STRUCT(superblock);
CELLFS_DEV_ATTR_STRUCT(segctor);

#define CELLFS_CP_ATTR_STRUCT(name) \
struct cellfs_##name##_attr { \
	struct attribute attr; \
	ssize_t (*show)(struct cellfs_##name##_attr *, struct cellfs_root *, \
			char *); \
	ssize_t (*store)(struct cellfs_##name##_attr *, struct cellfs_root *, \
			 const char *, size_t); \
}

CELLFS_CP_ATTR_STRUCT(snapshot);

#define CELLFS_ATTR(type, name, mode, show, store) \
	static struct cellfs_##type##_attr cellfs_##type##_attr_##name = \
		__ATTR(name, mode, show, store)

#define CELLFS_INFO_ATTR(type, name) \
	CELLFS_ATTR(type, name, 0444, NULL, NULL)
#define CELLFS_RO_ATTR(type, name) \
	CELLFS_ATTR(type, name, 0444, cellfs_##type##_##name##_show, NULL)
#define CELLFS_RW_ATTR(type, name) \
	CELLFS_ATTR(type, name, 0644, \
		    cellfs_##type##_##name##_show, \
		    cellfs_##type##_##name##_store)

#define CELLFS_FEATURE_INFO_ATTR(name) \
	CELLFS_INFO_ATTR(feature, name)
#define CELLFS_FEATURE_RO_ATTR(name) \
	CELLFS_RO_ATTR(feature, name)
#define CELLFS_FEATURE_RW_ATTR(name) \
	CELLFS_RW_ATTR(feature, name)

#define CELLFS_DEV_INFO_ATTR(name) \
	CELLFS_INFO_ATTR(dev, name)
#define CELLFS_DEV_RO_ATTR(name) \
	CELLFS_RO_ATTR(dev, name)
#define CELLFS_DEV_RW_ATTR(name) \
	CELLFS_RW_ATTR(dev, name)

#define CELLFS_SEGMENTS_RO_ATTR(name) \
	CELLFS_RO_ATTR(segments, name)
#define CELLFS_SEGMENTS_RW_ATTR(name) \
	CELLFS_RW_ATTR(segs_info, name)

#define CELLFS_MOUNTED_SNAPSHOTS_RO_ATTR(name) \
	CELLFS_RO_ATTR(mounted_snapshots, name)

#define CELLFS_CHECKPOINTS_RO_ATTR(name) \
	CELLFS_RO_ATTR(checkpoints, name)
#define CELLFS_CHECKPOINTS_RW_ATTR(name) \
	CELLFS_RW_ATTR(checkpoints, name)

#define CELLFS_SNAPSHOT_INFO_ATTR(name) \
	CELLFS_INFO_ATTR(snapshot, name)
#define CELLFS_SNAPSHOT_RO_ATTR(name) \
	CELLFS_RO_ATTR(snapshot, name)
#define CELLFS_SNAPSHOT_RW_ATTR(name) \
	CELLFS_RW_ATTR(snapshot, name)

#define CELLFS_SUPERBLOCK_RO_ATTR(name) \
	CELLFS_RO_ATTR(superblock, name)
#define CELLFS_SUPERBLOCK_RW_ATTR(name) \
	CELLFS_RW_ATTR(superblock, name)

#define CELLFS_SEGCTOR_INFO_ATTR(name) \
	CELLFS_INFO_ATTR(segctor, name)
#define CELLFS_SEGCTOR_RO_ATTR(name) \
	CELLFS_RO_ATTR(segctor, name)
#define CELLFS_SEGCTOR_RW_ATTR(name) \
	CELLFS_RW_ATTR(segctor, name)

#define CELLFS_FEATURE_ATTR_LIST(name) \
	(&cellfs_feature_attr_##name.attr)
#define CELLFS_DEV_ATTR_LIST(name) \
	(&cellfs_dev_attr_##name.attr)
#define CELLFS_SEGMENTS_ATTR_LIST(name) \
	(&cellfs_segments_attr_##name.attr)
#define CELLFS_MOUNTED_SNAPSHOTS_ATTR_LIST(name) \
	(&cellfs_mounted_snapshots_attr_##name.attr)
#define CELLFS_CHECKPOINTS_ATTR_LIST(name) \
	(&cellfs_checkpoints_attr_##name.attr)
#define CELLFS_SNAPSHOT_ATTR_LIST(name) \
	(&cellfs_snapshot_attr_##name.attr)
#define CELLFS_SUPERBLOCK_ATTR_LIST(name) \
	(&cellfs_superblock_attr_##name.attr)
#define CELLFS_SEGCTOR_ATTR_LIST(name) \
	(&cellfs_segctor_attr_##name.attr)

#endif /* _CELLFS_SYSFS_H */
