// SPDX-License-Identifier: GPL-2.0+
/*
 * direct.c - CELLFS direct block pointer.
 *
 * Copyright (C) 2006-2008 Nippon Telegraph and Telephone Corporation.
 *
 * Written by Koji Sato.
 */

#include <linux/errno.h>
#include "cellfs.h"
#include "page.h"
#include "direct.h"
#include "alloc.h"
#include "dat.h"

static inline __le64 *cellfs_direct_dptrs(const struct cellfs_bmap *direct)
{
	return (__le64 *)
		((struct cellfs_direct_node *)direct->b_u.u_data + 1);
}

static inline __u64
cellfs_direct_get_ptr(const struct cellfs_bmap *direct, __u64 key)
{
	return le64_to_cpu(*(cellfs_direct_dptrs(direct) + key));
}

static inline void cellfs_direct_set_ptr(struct cellfs_bmap *direct,
					__u64 key, __u64 ptr)
{
	*(cellfs_direct_dptrs(direct) + key) = cpu_to_le64(ptr);
}

static int cellfs_direct_lookup(const struct cellfs_bmap *direct,
			       __u64 key, int level, __u64 *ptrp)
{
	__u64 ptr;

	if (key > CELLFS_DIRECT_KEY_MAX || level != 1)
		return -ENOENT;
	ptr = cellfs_direct_get_ptr(direct, key);
	if (ptr == CELLFS_BMAP_INVALID_PTR)
		return -ENOENT;

	*ptrp = ptr;
	return 0;
}

static int cellfs_direct_lookup_contig(const struct cellfs_bmap *direct,
				      __u64 key, __u64 *ptrp,
				      unsigned int maxblocks)
{
	struct inode *dat = NULL;
	__u64 ptr, ptr2;
	sector_t blocknr;
	int ret, cnt;

	if (key > CELLFS_DIRECT_KEY_MAX)
		return -ENOENT;
	ptr = cellfs_direct_get_ptr(direct, key);
	if (ptr == CELLFS_BMAP_INVALID_PTR)
		return -ENOENT;

	if (CELLFS_BMAP_USE_VBN(direct)) {
		dat = cellfs_bmap_get_dat(direct);
		ret = cellfs_dat_translate(dat, ptr, &blocknr);
		if (ret < 0)
			return ret;
		ptr = blocknr;
	}

	maxblocks = min_t(unsigned int, maxblocks,
			  CELLFS_DIRECT_KEY_MAX - key + 1);
	for (cnt = 1; cnt < maxblocks &&
		     (ptr2 = cellfs_direct_get_ptr(direct, key + cnt)) !=
		     CELLFS_BMAP_INVALID_PTR;
	     cnt++) {
		if (dat) {
			ret = cellfs_dat_translate(dat, ptr2, &blocknr);
			if (ret < 0)
				return ret;
			ptr2 = blocknr;
		}
		if (ptr2 != ptr + cnt)
			break;
	}
	*ptrp = ptr;
	return cnt;
}

static __u64
cellfs_direct_find_target_v(const struct cellfs_bmap *direct, __u64 key)
{
	__u64 ptr;

	ptr = cellfs_bmap_find_target_seq(direct, key);
	if (ptr != CELLFS_BMAP_INVALID_PTR)
		/* sequential access */
		return ptr;

	/* block group */
	return cellfs_bmap_find_target_in_group(direct);
}

static int cellfs_direct_insert(struct cellfs_bmap *bmap, __u64 key, __u64 ptr)
{
	union cellfs_bmap_ptr_req req;
	struct inode *dat = NULL;
	struct buffer_head *bh;
	int ret;

	if (key > CELLFS_DIRECT_KEY_MAX)
		return -ENOENT;
	if (cellfs_direct_get_ptr(bmap, key) != CELLFS_BMAP_INVALID_PTR)
		return -EEXIST;

	if (CELLFS_BMAP_USE_VBN(bmap)) {
		req.bpr_ptr = cellfs_direct_find_target_v(bmap, key);
		dat = cellfs_bmap_get_dat(bmap);
	}
	ret = cellfs_bmap_prepare_alloc_ptr(bmap, &req, dat);
	if (!ret) {
		/* ptr must be a pointer to a buffer head. */
		bh = (struct buffer_head *)((unsigned long)ptr);
		set_buffer_cellfs_volatile(bh);

		cellfs_bmap_commit_alloc_ptr(bmap, &req, dat);
		cellfs_direct_set_ptr(bmap, key, req.bpr_ptr);

		if (!cellfs_bmap_dirty(bmap))
			cellfs_bmap_set_dirty(bmap);

		if (CELLFS_BMAP_USE_VBN(bmap))
			cellfs_bmap_set_target_v(bmap, key, req.bpr_ptr);

		cellfs_inode_add_blocks(bmap->b_inode, 1);
	}
	return ret;
}

static int cellfs_direct_delete(struct cellfs_bmap *bmap, __u64 key)
{
	union cellfs_bmap_ptr_req req;
	struct inode *dat;
	int ret;

	if (key > CELLFS_DIRECT_KEY_MAX ||
	    cellfs_direct_get_ptr(bmap, key) == CELLFS_BMAP_INVALID_PTR)
		return -ENOENT;

	dat = CELLFS_BMAP_USE_VBN(bmap) ? cellfs_bmap_get_dat(bmap) : NULL;
	req.bpr_ptr = cellfs_direct_get_ptr(bmap, key);

	ret = cellfs_bmap_prepare_end_ptr(bmap, &req, dat);
	if (!ret) {
		cellfs_bmap_commit_end_ptr(bmap, &req, dat);
		cellfs_direct_set_ptr(bmap, key, CELLFS_BMAP_INVALID_PTR);
		cellfs_inode_sub_blocks(bmap->b_inode, 1);
	}
	return ret;
}

static int cellfs_direct_seek_key(const struct cellfs_bmap *direct, __u64 start,
				 __u64 *keyp)
{
	__u64 key;

	for (key = start; key <= CELLFS_DIRECT_KEY_MAX; key++) {
		if (cellfs_direct_get_ptr(direct, key) !=
		    CELLFS_BMAP_INVALID_PTR) {
			*keyp = key;
			return 0;
		}
	}
	return -ENOENT;
}

static int cellfs_direct_last_key(const struct cellfs_bmap *direct, __u64 *keyp)
{
	__u64 key, lastkey;

	lastkey = CELLFS_DIRECT_KEY_MAX + 1;
	for (key = CELLFS_DIRECT_KEY_MIN; key <= CELLFS_DIRECT_KEY_MAX; key++)
		if (cellfs_direct_get_ptr(direct, key) !=
		    CELLFS_BMAP_INVALID_PTR)
			lastkey = key;

	if (lastkey == CELLFS_DIRECT_KEY_MAX + 1)
		return -ENOENT;

	*keyp = lastkey;

	return 0;
}

static int cellfs_direct_check_insert(const struct cellfs_bmap *bmap, __u64 key)
{
	return key > CELLFS_DIRECT_KEY_MAX;
}

static int cellfs_direct_gather_data(struct cellfs_bmap *direct,
				    __u64 *keys, __u64 *ptrs, int nitems)
{
	__u64 key;
	__u64 ptr;
	int n;

	if (nitems > CELLFS_DIRECT_NBLOCKS)
		nitems = CELLFS_DIRECT_NBLOCKS;
	n = 0;
	for (key = 0; key < nitems; key++) {
		ptr = cellfs_direct_get_ptr(direct, key);
		if (ptr != CELLFS_BMAP_INVALID_PTR) {
			keys[n] = key;
			ptrs[n] = ptr;
			n++;
		}
	}
	return n;
}

int cellfs_direct_delete_and_convert(struct cellfs_bmap *bmap,
				    __u64 key, __u64 *keys, __u64 *ptrs, int n)
{
	__le64 *dptrs;
	int ret, i, j;

	/* no need to allocate any resource for conversion */

	/* delete */
	ret = bmap->b_ops->bop_delete(bmap, key);
	if (ret < 0)
		return ret;

	/* free resources */
	if (bmap->b_ops->bop_clear != NULL)
		bmap->b_ops->bop_clear(bmap);

	/* convert */
	dptrs = cellfs_direct_dptrs(bmap);
	for (i = 0, j = 0; i < CELLFS_DIRECT_NBLOCKS; i++) {
		if ((j < n) && (i == keys[j])) {
			dptrs[i] = (i != key) ?
				cpu_to_le64(ptrs[j]) :
				CELLFS_BMAP_INVALID_PTR;
			j++;
		} else
			dptrs[i] = CELLFS_BMAP_INVALID_PTR;
	}

	cellfs_direct_init(bmap);
	return 0;
}

static int cellfs_direct_propagate(struct cellfs_bmap *bmap,
				  struct buffer_head *bh)
{
	struct cellfs_palloc_req oldreq, newreq;
	struct inode *dat;
	__u64 key;
	__u64 ptr;
	int ret;

	if (!CELLFS_BMAP_USE_VBN(bmap))
		return 0;

	dat = cellfs_bmap_get_dat(bmap);
	key = cellfs_bmap_data_get_key(bmap, bh);
	ptr = cellfs_direct_get_ptr(bmap, key);
	if (!buffer_cellfs_volatile(bh)) {
		oldreq.pr_entry_nr = ptr;
		newreq.pr_entry_nr = ptr;
		ret = cellfs_dat_prepare_update(dat, &oldreq, &newreq);
		if (ret < 0)
			return ret;
		cellfs_dat_commit_update(dat, &oldreq, &newreq,
					bmap->b_ptr_type == CELLFS_BMAP_PTR_VS);
		set_buffer_cellfs_volatile(bh);
		cellfs_direct_set_ptr(bmap, key, newreq.pr_entry_nr);
	} else
		ret = cellfs_dat_mark_dirty(dat, ptr);

	return ret;
}

static int cellfs_direct_assign_v(struct cellfs_bmap *direct,
				 __u64 key, __u64 ptr,
				 struct buffer_head **bh,
				 sector_t blocknr,
				 union cellfs_binfo *binfo)
{
	struct inode *dat = cellfs_bmap_get_dat(direct);
	union cellfs_bmap_ptr_req req;
	int ret;

	req.bpr_ptr = ptr;
	ret = cellfs_dat_prepare_start(dat, &req.bpr_req);
	if (!ret) {
		cellfs_dat_commit_start(dat, &req.bpr_req, blocknr);
		binfo->bi_v.bi_vblocknr = cpu_to_le64(ptr);
		binfo->bi_v.bi_blkoff = cpu_to_le64(key);
	}
	return ret;
}

static int cellfs_direct_assign_p(struct cellfs_bmap *direct,
				 __u64 key, __u64 ptr,
				 struct buffer_head **bh,
				 sector_t blocknr,
				 union cellfs_binfo *binfo)
{
	cellfs_direct_set_ptr(direct, key, blocknr);

	binfo->bi_dat.bi_blkoff = cpu_to_le64(key);
	binfo->bi_dat.bi_level = 0;

	return 0;
}

static int cellfs_direct_assign(struct cellfs_bmap *bmap,
			       struct buffer_head **bh,
			       sector_t blocknr,
			       union cellfs_binfo *binfo)
{
	__u64 key;
	__u64 ptr;

	key = cellfs_bmap_data_get_key(bmap, *bh);
	if (unlikely(key > CELLFS_DIRECT_KEY_MAX)) {
		cellfs_msg(bmap->b_inode->i_sb, KERN_CRIT,
			  "%s (ino=%lu): invalid key: %llu", __func__,
			  bmap->b_inode->i_ino, (unsigned long long)key);
		return -EINVAL;
	}
	ptr = cellfs_direct_get_ptr(bmap, key);
	if (unlikely(ptr == CELLFS_BMAP_INVALID_PTR)) {
		cellfs_msg(bmap->b_inode->i_sb, KERN_CRIT,
			  "%s (ino=%lu): invalid pointer: %llu", __func__,
			  bmap->b_inode->i_ino, (unsigned long long)ptr);
		return -EINVAL;
	}

	return CELLFS_BMAP_USE_VBN(bmap) ?
		cellfs_direct_assign_v(bmap, key, ptr, bh, blocknr, binfo) :
		cellfs_direct_assign_p(bmap, key, ptr, bh, blocknr, binfo);
}

static const struct cellfs_bmap_operations cellfs_direct_ops = {
	.bop_lookup		=	cellfs_direct_lookup,
	.bop_lookup_contig	=	cellfs_direct_lookup_contig,
	.bop_insert		=	cellfs_direct_insert,
	.bop_delete		=	cellfs_direct_delete,
	.bop_clear		=	NULL,

	.bop_propagate		=	cellfs_direct_propagate,

	.bop_lookup_dirty_buffers	=	NULL,

	.bop_assign		=	cellfs_direct_assign,
	.bop_mark		=	NULL,

	.bop_seek_key		=	cellfs_direct_seek_key,
	.bop_last_key		=	cellfs_direct_last_key,

	.bop_check_insert	=	cellfs_direct_check_insert,
	.bop_check_delete	=	NULL,
	.bop_gather_data	=	cellfs_direct_gather_data,
};


int cellfs_direct_init(struct cellfs_bmap *bmap)
{
	bmap->b_ops = &cellfs_direct_ops;
	return 0;
}
