// SPDX-License-Identifier: GPL-2.0+
/*
 * segment.c - CELLFS segment constructor.
 *
 * Copyright (C) 2005-2008 Nippon Telegraph and Telephone Corporation.
 *
 * Written by Ryusuke Konishi.
 *
 */

#include <linux/pagemap.h>
#include <linux/buffer_head.h>
#include <linux/writeback.h>
#include <linux/bitops.h>
#include <linux/bio.h>
#include <linux/completion.h>
#include <linux/blkdev.h>
#include <linux/backing-dev.h>
#include <linux/freezer.h>
#include <linux/kthread.h>
#include <linux/crc32.h>
#include <linux/pagevec.h>
#include <linux/slab.h>
#include <linux/sched/signal.h>

#include "cellfs.h"
#include "btnode.h"
#include "page.h"
#include "segment.h"
#include "sufile.h"
#include "cpfile.h"
#include "ifile.h"
#include "segbuf.h"


/*
 * Segment constructor
 */
#define SC_N_INODEVEC	16   /* Size of locally allocated inode vector */

#define SC_MAX_SEGDELTA 64   /*
			      * Upper limit of the number of segments
			      * appended in collection retry loop
			      */

/* Construction mode */
enum {
	SC_LSEG_SR = 1,	/* Make a logical segment having a super root */
	SC_LSEG_DSYNC,	/*
			 * Flush data blocks of a given file and make
			 * a logical segment without a super root.
			 */
	SC_FLUSH_FILE,	/*
			 * Flush data files, leads to segment writes without
			 * creating a checkpoint.
			 */
	SC_FLUSH_DAT,	/*
			 * Flush DAT file.  This also creates segments
			 * without a checkpoint.
			 */
};

/* Stage numbers of dirty block collection */
enum {
	CELLFS_ST_INIT = 0,
	CELLFS_ST_GC,		/* Collecting dirty blocks for GC */
	CELLFS_ST_FILE,
	CELLFS_ST_IFILE,
	CELLFS_ST_CPFILE,
	CELLFS_ST_SUFILE,
	CELLFS_ST_DAT,
	CELLFS_ST_SR,		/* Super root */
	CELLFS_ST_DSYNC,		/* Data sync blocks */
	CELLFS_ST_DONE,
};

#define CREATE_TRACE_POINTS
#include <trace/events/cellfs.h>

/*
 * cellfs_sc_cstage_inc(), cellfs_sc_cstage_set(), cellfs_sc_cstage_get() are
 * wrapper functions of stage count (cellfs_sc_info->sc_stage.scnt). Users of
 * the variable must use them because transition of stage count must involve
 * trace events (trace_cellfs_collection_stage_transition).
 *
 * cellfs_sc_cstage_get() isn't required for the above purpose because it doesn't
 * produce tracepoint events. It is provided just for making the intention
 * clear.
 */
static inline void cellfs_sc_cstage_inc(struct cellfs_sc_info *sci)
{
	sci->sc_stage.scnt++;
	trace_cellfs_collection_stage_transition(sci);
}

static inline void cellfs_sc_cstage_set(struct cellfs_sc_info *sci, int next_scnt)
{
	sci->sc_stage.scnt = next_scnt;
	trace_cellfs_collection_stage_transition(sci);
}

static inline int cellfs_sc_cstage_get(struct cellfs_sc_info *sci)
{
	return sci->sc_stage.scnt;
}

/* State flags of collection */
#define CELLFS_CF_NODE		0x0001	/* Collecting node blocks */
#define CELLFS_CF_IFILE_STARTED	0x0002	/* IFILE stage has started */
#define CELLFS_CF_SUFREED	0x0004	/* segment usages has been freed */
#define CELLFS_CF_HISTORY_MASK	(CELLFS_CF_IFILE_STARTED | CELLFS_CF_SUFREED)

/* Operations depending on the construction mode and file type */
struct cellfs_sc_operations {
	int (*collect_data)(struct cellfs_sc_info *, struct buffer_head *,
			    struct inode *);
	int (*collect_node)(struct cellfs_sc_info *, struct buffer_head *,
			    struct inode *);
	int (*collect_bmap)(struct cellfs_sc_info *, struct buffer_head *,
			    struct inode *);
	void (*write_data_binfo)(struct cellfs_sc_info *,
				 struct cellfs_segsum_pointer *,
				 union cellfs_binfo *);
	void (*write_node_binfo)(struct cellfs_sc_info *,
				 struct cellfs_segsum_pointer *,
				 union cellfs_binfo *);
};

/*
 * Other definitions
 */
static void cellfs_segctor_start_timer(struct cellfs_sc_info *);
static void cellfs_segctor_do_flush(struct cellfs_sc_info *, int);
static void cellfs_segctor_do_immediate_flush(struct cellfs_sc_info *);
static void cellfs_dispose_list(struct the_cellfs *, struct list_head *, int);

#define cellfs_cnt32_gt(a, b)   \
	(typecheck(__u32, a) && typecheck(__u32, b) && \
	 ((__s32)(b) - (__s32)(a) < 0))
#define cellfs_cnt32_ge(a, b)   \
	(typecheck(__u32, a) && typecheck(__u32, b) && \
	 ((__s32)(a) - (__s32)(b) >= 0))
#define cellfs_cnt32_lt(a, b)  cellfs_cnt32_gt(b, a)
#define cellfs_cnt32_le(a, b)  cellfs_cnt32_ge(b, a)

static int cellfs_prepare_segment_lock(struct super_block *sb,
				      struct cellfs_transaction_info *ti)
{
	struct cellfs_transaction_info *cur_ti = current->journal_info;
	void *save = NULL;

	if (cur_ti) {
		if (cur_ti->ti_magic == CELLFS_TI_MAGIC)
			return ++cur_ti->ti_count;

		/*
		 * If journal_info field is occupied by other FS,
		 * it is saved and will be restored on
		 * cellfs_transaction_commit().
		 */
		cellfs_msg(sb, KERN_WARNING, "journal info from a different FS");
		save = current->journal_info;
	}
	if (!ti) {
		ti = kmem_cache_alloc(cellfs_transaction_cachep, GFP_NOFS);
		if (!ti)
			return -ENOMEM;
		ti->ti_flags = CELLFS_TI_DYNAMIC_ALLOC;
	} else {
		ti->ti_flags = 0;
	}
	ti->ti_count = 0;
	ti->ti_save = save;
	ti->ti_magic = CELLFS_TI_MAGIC;
	current->journal_info = ti;
	return 0;
}

/**
 * cellfs_transaction_begin - start indivisible file operations.
 * @sb: super block
 * @ti: cellfs_transaction_info
 * @vacancy_check: flags for vacancy rate checks
 *
 * cellfs_transaction_begin() acquires a reader/writer semaphore, called
 * the segment semaphore, to make a segment construction and write tasks
 * exclusive.  The function is used with cellfs_transaction_commit() in pairs.
 * The region enclosed by these two functions can be nested.  To avoid a
 * deadlock, the semaphore is only acquired or released in the outermost call.
 *
 * This function allocates a cellfs_transaction_info struct to keep context
 * information on it.  It is initialized and hooked onto the current task in
 * the outermost call.  If a pre-allocated struct is given to @ti, it is used
 * instead; otherwise a new struct is assigned from a slab.
 *
 * When @vacancy_check flag is set, this function will check the amount of
 * free space, and will wait for the GC to reclaim disk space if low capacity.
 *
 * Return Value: On success, 0 is returned. On error, one of the following
 * negative error code is returned.
 *
 * %-ENOMEM - Insufficient memory available.
 *
 * %-ENOSPC - No space left on device
 */
int cellfs_transaction_begin(struct super_block *sb,
			    struct cellfs_transaction_info *ti,
			    int vacancy_check)
{
	struct the_cellfs *cellfs;
	int ret = cellfs_prepare_segment_lock(sb, ti);
	struct cellfs_transaction_info *trace_ti;

	if (unlikely(ret < 0))
		return ret;
	if (ret > 0) {
		trace_ti = current->journal_info;

		trace_cellfs_transaction_transition(sb, trace_ti,
				    trace_ti->ti_count, trace_ti->ti_flags,
				    TRACE_CELLFS_TRANSACTION_BEGIN);
		return 0;
	}

	sb_start_intwrite(sb);

	cellfs = sb->s_fs_info;
	down_read(&cellfs->ns_segctor_sem);
	if (vacancy_check && cellfs_near_disk_full(cellfs)) {
		up_read(&cellfs->ns_segctor_sem);
		ret = -ENOSPC;
		goto failed;
	}

	trace_ti = current->journal_info;
	trace_cellfs_transaction_transition(sb, trace_ti, trace_ti->ti_count,
					    trace_ti->ti_flags,
					    TRACE_CELLFS_TRANSACTION_BEGIN);
	return 0;

 failed:
	ti = current->journal_info;
	current->journal_info = ti->ti_save;
	if (ti->ti_flags & CELLFS_TI_DYNAMIC_ALLOC)
		kmem_cache_free(cellfs_transaction_cachep, ti);
	sb_end_intwrite(sb);
	return ret;
}

/**
 * cellfs_transaction_commit - commit indivisible file operations.
 * @sb: super block
 *
 * cellfs_transaction_commit() releases the read semaphore which is
 * acquired by cellfs_transaction_begin(). This is only performed
 * in outermost call of this function.  If a commit flag is set,
 * cellfs_transaction_commit() sets a timer to start the segment
 * constructor.  If a sync flag is set, it starts construction
 * directly.
 */
int cellfs_transaction_commit(struct super_block *sb)
{
	struct cellfs_transaction_info *ti = current->journal_info;
	struct the_cellfs *cellfs = sb->s_fs_info;
	int err = 0;

	BUG_ON(ti == NULL || ti->ti_magic != CELLFS_TI_MAGIC);
	ti->ti_flags |= CELLFS_TI_COMMIT;
	if (ti->ti_count > 0) {
		ti->ti_count--;
		trace_cellfs_transaction_transition(sb, ti, ti->ti_count,
			    ti->ti_flags, TRACE_CELLFS_TRANSACTION_COMMIT);
		return 0;
	}
	if (cellfs->ns_writer) {
		struct cellfs_sc_info *sci = cellfs->ns_writer;

		if (ti->ti_flags & CELLFS_TI_COMMIT)
			cellfs_segctor_start_timer(sci);
		if (atomic_read(&cellfs->ns_ndirtyblks) > sci->sc_watermark)
			cellfs_segctor_do_flush(sci, 0);
	}
	up_read(&cellfs->ns_segctor_sem);
	trace_cellfs_transaction_transition(sb, ti, ti->ti_count,
			    ti->ti_flags, TRACE_CELLFS_TRANSACTION_COMMIT);

	current->journal_info = ti->ti_save;

	if (ti->ti_flags & CELLFS_TI_SYNC)
		err = cellfs_construct_segment(sb);
	if (ti->ti_flags & CELLFS_TI_DYNAMIC_ALLOC)
		kmem_cache_free(cellfs_transaction_cachep, ti);
	sb_end_intwrite(sb);
	return err;
}

void cellfs_transaction_abort(struct super_block *sb)
{
	struct cellfs_transaction_info *ti = current->journal_info;
	struct the_cellfs *cellfs = sb->s_fs_info;

	BUG_ON(ti == NULL || ti->ti_magic != CELLFS_TI_MAGIC);
	if (ti->ti_count > 0) {
		ti->ti_count--;
		trace_cellfs_transaction_transition(sb, ti, ti->ti_count,
			    ti->ti_flags, TRACE_CELLFS_TRANSACTION_ABORT);
		return;
	}
	up_read(&cellfs->ns_segctor_sem);

	trace_cellfs_transaction_transition(sb, ti, ti->ti_count,
		    ti->ti_flags, TRACE_CELLFS_TRANSACTION_ABORT);

	current->journal_info = ti->ti_save;
	if (ti->ti_flags & CELLFS_TI_DYNAMIC_ALLOC)
		kmem_cache_free(cellfs_transaction_cachep, ti);
	sb_end_intwrite(sb);
}

void cellfs_relax_pressure_in_lock(struct super_block *sb)
{
	struct the_cellfs *cellfs = sb->s_fs_info;
	struct cellfs_sc_info *sci = cellfs->ns_writer;

	if (!sci || !sci->sc_flush_request)
		return;

	set_bit(CELLFS_SC_PRIOR_FLUSH, &sci->sc_flags);
	up_read(&cellfs->ns_segctor_sem);

	down_write(&cellfs->ns_segctor_sem);
	if (sci->sc_flush_request &&
	    test_bit(CELLFS_SC_PRIOR_FLUSH, &sci->sc_flags)) {
		struct cellfs_transaction_info *ti = current->journal_info;

		ti->ti_flags |= CELLFS_TI_WRITER;
		cellfs_segctor_do_immediate_flush(sci);
		ti->ti_flags &= ~CELLFS_TI_WRITER;
	}
	downgrade_write(&cellfs->ns_segctor_sem);
}

static void cellfs_transaction_lock(struct super_block *sb,
				   struct cellfs_transaction_info *ti,
				   int gcflag)
{
	struct cellfs_transaction_info *cur_ti = current->journal_info;
	struct the_cellfs *cellfs = sb->s_fs_info;
	struct cellfs_sc_info *sci = cellfs->ns_writer;

	WARN_ON(cur_ti);
	ti->ti_flags = CELLFS_TI_WRITER;
	ti->ti_count = 0;
	ti->ti_save = cur_ti;
	ti->ti_magic = CELLFS_TI_MAGIC;
	current->journal_info = ti;

	for (;;) {
		trace_cellfs_transaction_transition(sb, ti, ti->ti_count,
			    ti->ti_flags, TRACE_CELLFS_TRANSACTION_TRYLOCK);

		down_write(&cellfs->ns_segctor_sem);
		if (!test_bit(CELLFS_SC_PRIOR_FLUSH, &sci->sc_flags))
			break;

		cellfs_segctor_do_immediate_flush(sci);

		up_write(&cellfs->ns_segctor_sem);
		cond_resched();
	}
	if (gcflag)
		ti->ti_flags |= CELLFS_TI_GC;

	trace_cellfs_transaction_transition(sb, ti, ti->ti_count,
			    ti->ti_flags, TRACE_CELLFS_TRANSACTION_LOCK);
}

static void cellfs_transaction_unlock(struct super_block *sb)
{
	struct cellfs_transaction_info *ti = current->journal_info;
	struct the_cellfs *cellfs = sb->s_fs_info;

	BUG_ON(ti == NULL || ti->ti_magic != CELLFS_TI_MAGIC);
	BUG_ON(ti->ti_count > 0);

	up_write(&cellfs->ns_segctor_sem);
	current->journal_info = ti->ti_save;

	trace_cellfs_transaction_transition(sb, ti, ti->ti_count,
			    ti->ti_flags, TRACE_CELLFS_TRANSACTION_UNLOCK);
}

static void *cellfs_segctor_map_segsum_entry(struct cellfs_sc_info *sci,
					    struct cellfs_segsum_pointer *ssp,
					    unsigned int bytes)
{
	struct cellfs_segment_buffer *segbuf = sci->sc_curseg;
	unsigned int blocksize = sci->sc_super->s_blocksize;
	void *p;

	if (unlikely(ssp->offset + bytes > blocksize)) {
		ssp->offset = 0;
		BUG_ON(CELLFS_SEGBUF_BH_IS_LAST(ssp->bh,
					       &segbuf->sb_segsum_buffers));
		ssp->bh = CELLFS_SEGBUF_NEXT_BH(ssp->bh);
	}
	p = ssp->bh->b_data + ssp->offset;
	ssp->offset += bytes;
	return p;
}

/**
 * cellfs_segctor_reset_segment_buffer - reset the current segment buffer
 * @sci: cellfs_sc_info
 */
static int cellfs_segctor_reset_segment_buffer(struct cellfs_sc_info *sci)
{
	struct cellfs_segment_buffer *segbuf = sci->sc_curseg;
	struct buffer_head *sumbh;
	unsigned int sumbytes;
	unsigned int flags = 0;
	int err;

	if (cellfs_doing_gc())
		flags = CELLFS_SS_GC;
	err = cellfs_segbuf_reset(segbuf, flags, sci->sc_seg_ctime, sci->sc_cno);
	if (unlikely(err))
		return err;

	sumbh = CELLFS_SEGBUF_FIRST_BH(&segbuf->sb_segsum_buffers);
	sumbytes = segbuf->sb_sum.sumbytes;
	sci->sc_finfo_ptr.bh = sumbh;  sci->sc_finfo_ptr.offset = sumbytes;
	sci->sc_binfo_ptr.bh = sumbh;  sci->sc_binfo_ptr.offset = sumbytes;
	sci->sc_blk_cnt = sci->sc_datablk_cnt = 0;
	return 0;
}

static int cellfs_segctor_feed_segment(struct cellfs_sc_info *sci)
{
	sci->sc_nblk_this_inc += sci->sc_curseg->sb_sum.nblocks;
	if (CELLFS_SEGBUF_IS_LAST(sci->sc_curseg, &sci->sc_segbufs))
		return -E2BIG; /*
				* The current segment is filled up
				* (internal code)
				*/
	sci->sc_curseg = CELLFS_NEXT_SEGBUF(sci->sc_curseg);
	return cellfs_segctor_reset_segment_buffer(sci);
}

static int cellfs_segctor_add_super_root(struct cellfs_sc_info *sci)
{
	struct cellfs_segment_buffer *segbuf = sci->sc_curseg;
	int err;

	if (segbuf->sb_sum.nblocks >= segbuf->sb_rest_blocks) {
		err = cellfs_segctor_feed_segment(sci);
		if (err)
			return err;
		segbuf = sci->sc_curseg;
	}
	err = cellfs_segbuf_extend_payload(segbuf, &segbuf->sb_super_root);
	if (likely(!err))
		segbuf->sb_sum.flags |= CELLFS_SS_SR;
	return err;
}

/*
 * Functions for making segment summary and payloads
 */
static int cellfs_segctor_segsum_block_required(
	struct cellfs_sc_info *sci, const struct cellfs_segsum_pointer *ssp,
	unsigned int binfo_size)
{
	unsigned int blocksize = sci->sc_super->s_blocksize;
	/* Size of finfo and binfo is enough small against blocksize */

	return ssp->offset + binfo_size +
		(!sci->sc_blk_cnt ? sizeof(struct cellfs_finfo) : 0) >
		blocksize;
}

static void cellfs_segctor_begin_finfo(struct cellfs_sc_info *sci,
				      struct inode *inode)
{
	sci->sc_curseg->sb_sum.nfinfo++;
	sci->sc_binfo_ptr = sci->sc_finfo_ptr;
	cellfs_segctor_map_segsum_entry(
		sci, &sci->sc_binfo_ptr, sizeof(struct cellfs_finfo));

	if (CELLFS_I(inode)->i_root &&
	    !test_bit(CELLFS_SC_HAVE_DELTA, &sci->sc_flags))
		set_bit(CELLFS_SC_HAVE_DELTA, &sci->sc_flags);
	/* skip finfo */
}

static void cellfs_segctor_end_finfo(struct cellfs_sc_info *sci,
				    struct inode *inode)
{
	struct cellfs_finfo *finfo;
	struct cellfs_inode_info *ii;
	struct cellfs_segment_buffer *segbuf;
	__u64 cno;

	if (sci->sc_blk_cnt == 0)
		return;

	ii = CELLFS_I(inode);

	if (test_bit(CELLFS_I_GCINODE, &ii->i_state))
		cno = ii->i_cno;
	else if (CELLFS_ROOT_METADATA_FILE(inode->i_ino))
		cno = 0;
	else
		cno = sci->sc_cno;

	finfo = cellfs_segctor_map_segsum_entry(sci, &sci->sc_finfo_ptr,
						 sizeof(*finfo));
	finfo->fi_ino = cpu_to_le64(inode->i_ino);
	finfo->fi_nblocks = cpu_to_le32(sci->sc_blk_cnt);
	finfo->fi_ndatablk = cpu_to_le32(sci->sc_datablk_cnt);
	finfo->fi_cno = cpu_to_le64(cno);

	segbuf = sci->sc_curseg;
	segbuf->sb_sum.sumbytes = sci->sc_binfo_ptr.offset +
		sci->sc_super->s_blocksize * (segbuf->sb_sum.nsumblk - 1);
	sci->sc_finfo_ptr = sci->sc_binfo_ptr;
	sci->sc_blk_cnt = sci->sc_datablk_cnt = 0;
}

static int cellfs_segctor_add_file_block(struct cellfs_sc_info *sci,
					struct buffer_head *bh,
					struct inode *inode,
					unsigned int binfo_size)
{
	struct cellfs_segment_buffer *segbuf;
	int required, err = 0;

 retry:
	segbuf = sci->sc_curseg;
	required = cellfs_segctor_segsum_block_required(
		sci, &sci->sc_binfo_ptr, binfo_size);
	if (segbuf->sb_sum.nblocks + required + 1 > segbuf->sb_rest_blocks) {
		cellfs_segctor_end_finfo(sci, inode);
		err = cellfs_segctor_feed_segment(sci);
		if (err)
			return err;
		goto retry;
	}
	if (unlikely(required)) {
		err = cellfs_segbuf_extend_segsum(segbuf);
		if (unlikely(err))
			goto failed;
	}
	if (sci->sc_blk_cnt == 0)
		cellfs_segctor_begin_finfo(sci, inode);

	cellfs_segctor_map_segsum_entry(sci, &sci->sc_binfo_ptr, binfo_size);
	/* Substitution to vblocknr is delayed until update_blocknr() */
	cellfs_segbuf_add_file_buffer(segbuf, bh);
	sci->sc_blk_cnt++;
 failed:
	return err;
}

/*
 * Callback functions that enumerate, mark, and collect dirty blocks
 */
static int cellfs_collect_file_data(struct cellfs_sc_info *sci,
				   struct buffer_head *bh, struct inode *inode)
{
	int err;

	err = cellfs_bmap_propagate(CELLFS_I(inode)->i_bmap, bh);
	if (err < 0)
		return err;

	err = cellfs_segctor_add_file_block(sci, bh, inode,
					   sizeof(struct cellfs_binfo_v));
	if (!err)
		sci->sc_datablk_cnt++;
	return err;
}

static int cellfs_collect_file_node(struct cellfs_sc_info *sci,
				   struct buffer_head *bh,
				   struct inode *inode)
{
	return cellfs_bmap_propagate(CELLFS_I(inode)->i_bmap, bh);
}

static int cellfs_collect_file_bmap(struct cellfs_sc_info *sci,
				   struct buffer_head *bh,
				   struct inode *inode)
{
	WARN_ON(!buffer_dirty(bh));
	return cellfs_segctor_add_file_block(sci, bh, inode, sizeof(__le64));
}

static void cellfs_write_file_data_binfo(struct cellfs_sc_info *sci,
					struct cellfs_segsum_pointer *ssp,
					union cellfs_binfo *binfo)
{
	struct cellfs_binfo_v *binfo_v = cellfs_segctor_map_segsum_entry(
		sci, ssp, sizeof(*binfo_v));
	*binfo_v = binfo->bi_v;
}

static void cellfs_write_file_node_binfo(struct cellfs_sc_info *sci,
					struct cellfs_segsum_pointer *ssp,
					union cellfs_binfo *binfo)
{
	__le64 *vblocknr = cellfs_segctor_map_segsum_entry(
		sci, ssp, sizeof(*vblocknr));
	*vblocknr = binfo->bi_v.bi_vblocknr;
}

static const struct cellfs_sc_operations cellfs_sc_file_ops = {
	.collect_data = cellfs_collect_file_data,
	.collect_node = cellfs_collect_file_node,
	.collect_bmap = cellfs_collect_file_bmap,
	.write_data_binfo = cellfs_write_file_data_binfo,
	.write_node_binfo = cellfs_write_file_node_binfo,
};

static int cellfs_collect_dat_data(struct cellfs_sc_info *sci,
				  struct buffer_head *bh, struct inode *inode)
{
	int err;

	err = cellfs_bmap_propagate(CELLFS_I(inode)->i_bmap, bh);
	if (err < 0)
		return err;

	err = cellfs_segctor_add_file_block(sci, bh, inode, sizeof(__le64));
	if (!err)
		sci->sc_datablk_cnt++;
	return err;
}

static int cellfs_collect_dat_bmap(struct cellfs_sc_info *sci,
				  struct buffer_head *bh, struct inode *inode)
{
	WARN_ON(!buffer_dirty(bh));
	return cellfs_segctor_add_file_block(sci, bh, inode,
					    sizeof(struct cellfs_binfo_dat));
}

static void cellfs_write_dat_data_binfo(struct cellfs_sc_info *sci,
				       struct cellfs_segsum_pointer *ssp,
				       union cellfs_binfo *binfo)
{
	__le64 *blkoff = cellfs_segctor_map_segsum_entry(sci, ssp,
							  sizeof(*blkoff));
	*blkoff = binfo->bi_dat.bi_blkoff;
}

static void cellfs_write_dat_node_binfo(struct cellfs_sc_info *sci,
				       struct cellfs_segsum_pointer *ssp,
				       union cellfs_binfo *binfo)
{
	struct cellfs_binfo_dat *binfo_dat =
		cellfs_segctor_map_segsum_entry(sci, ssp, sizeof(*binfo_dat));
	*binfo_dat = binfo->bi_dat;
}

static const struct cellfs_sc_operations cellfs_sc_dat_ops = {
	.collect_data = cellfs_collect_dat_data,
	.collect_node = cellfs_collect_file_node,
	.collect_bmap = cellfs_collect_dat_bmap,
	.write_data_binfo = cellfs_write_dat_data_binfo,
	.write_node_binfo = cellfs_write_dat_node_binfo,
};

static const struct cellfs_sc_operations cellfs_sc_dsync_ops = {
	.collect_data = cellfs_collect_file_data,
	.collect_node = NULL,
	.collect_bmap = NULL,
	.write_data_binfo = cellfs_write_file_data_binfo,
	.write_node_binfo = NULL,
};

static size_t cellfs_lookup_dirty_data_buffers(struct inode *inode,
					      struct list_head *listp,
					      size_t nlimit,
					      loff_t start, loff_t end)
{
	struct address_space *mapping = inode->i_mapping;
	struct pagevec pvec;
	pgoff_t index = 0, last = ULONG_MAX;
	size_t ndirties = 0;
	int i;

	if (unlikely(start != 0 || end != LLONG_MAX)) {
		/*
		 * A valid range is given for sync-ing data pages. The
		 * range is rounded to per-page; extra dirty buffers
		 * may be included if blocksize < pagesize.
		 */
		index = start >> PAGE_SHIFT;
		last = end >> PAGE_SHIFT;
	}
	pagevec_init(&pvec);
 repeat:
	if (unlikely(index > last) ||
	    !pagevec_lookup_range_tag(&pvec, mapping, &index, last,
				PAGECACHE_TAG_DIRTY))
		return ndirties;

	for (i = 0; i < pagevec_count(&pvec); i++) {
		struct buffer_head *bh, *head;
		struct page *page = pvec.pages[i];

		lock_page(page);
		if (!page_has_buffers(page))
			create_empty_buffers(page, i_blocksize(inode), 0);
		unlock_page(page);

		bh = head = page_buffers(page);
		do {
			if (!buffer_dirty(bh) || buffer_async_write(bh))
				continue;
			get_bh(bh);
			list_add_tail(&bh->b_assoc_buffers, listp);
			ndirties++;
			if (unlikely(ndirties >= nlimit)) {
				pagevec_release(&pvec);
				cond_resched();
				return ndirties;
			}
		} while (bh = bh->b_this_page, bh != head);
	}
	pagevec_release(&pvec);
	cond_resched();
	goto repeat;
}

static void cellfs_lookup_dirty_node_buffers(struct inode *inode,
					    struct list_head *listp)
{
	struct cellfs_inode_info *ii = CELLFS_I(inode);
	struct address_space *mapping = &ii->i_btnode_cache;
	struct pagevec pvec;
	struct buffer_head *bh, *head;
	unsigned int i;
	pgoff_t index = 0;

	pagevec_init(&pvec);

	while (pagevec_lookup_tag(&pvec, mapping, &index,
					PAGECACHE_TAG_DIRTY)) {
		for (i = 0; i < pagevec_count(&pvec); i++) {
			bh = head = page_buffers(pvec.pages[i]);
			do {
				if (buffer_dirty(bh) &&
						!buffer_async_write(bh)) {
					get_bh(bh);
					list_add_tail(&bh->b_assoc_buffers,
						      listp);
				}
				bh = bh->b_this_page;
			} while (bh != head);
		}
		pagevec_release(&pvec);
		cond_resched();
	}
}

static void cellfs_dispose_list(struct the_cellfs *cellfs,
			       struct list_head *head, int force)
{
	struct cellfs_inode_info *ii, *n;
	struct cellfs_inode_info *ivec[SC_N_INODEVEC], **pii;
	unsigned int nv = 0;

	while (!list_empty(head)) {
		spin_lock(&cellfs->ns_inode_lock);
		list_for_each_entry_safe(ii, n, head, i_dirty) {
			list_del_init(&ii->i_dirty);
			if (force) {
				if (unlikely(ii->i_bh)) {
					brelse(ii->i_bh);
					ii->i_bh = NULL;
				}
			} else if (test_bit(CELLFS_I_DIRTY, &ii->i_state)) {
				set_bit(CELLFS_I_QUEUED, &ii->i_state);
				list_add_tail(&ii->i_dirty,
					      &cellfs->ns_dirty_files);
				continue;
			}
			ivec[nv++] = ii;
			if (nv == SC_N_INODEVEC)
				break;
		}
		spin_unlock(&cellfs->ns_inode_lock);

		for (pii = ivec; nv > 0; pii++, nv--)
			iput(&(*pii)->vfs_inode);
	}
}

static void cellfs_iput_work_func(struct work_struct *work)
{
	struct cellfs_sc_info *sci = container_of(work, struct cellfs_sc_info,
						 sc_iput_work);
	struct the_cellfs *cellfs = sci->sc_super->s_fs_info;

	cellfs_dispose_list(cellfs, &sci->sc_iput_queue, 0);
}

static int cellfs_test_metadata_dirty(struct the_cellfs *cellfs,
				     struct cellfs_root *root)
{
	int ret = 0;

	if (cellfs_mdt_fetch_dirty(root->ifile))
		ret++;
	if (cellfs_mdt_fetch_dirty(cellfs->ns_cpfile))
		ret++;
	if (cellfs_mdt_fetch_dirty(cellfs->ns_sufile))
		ret++;
	if ((ret || cellfs_doing_gc()) && cellfs_mdt_fetch_dirty(cellfs->ns_dat))
		ret++;
	return ret;
}

static int cellfs_segctor_clean(struct cellfs_sc_info *sci)
{
	return list_empty(&sci->sc_dirty_files) &&
		!test_bit(CELLFS_SC_DIRTY, &sci->sc_flags) &&
		sci->sc_nfreesegs == 0 &&
		(!cellfs_doing_gc() || list_empty(&sci->sc_gc_inodes));
}

static int cellfs_segctor_confirm(struct cellfs_sc_info *sci)
{
	struct the_cellfs *cellfs = sci->sc_super->s_fs_info;
	int ret = 0;

	if (cellfs_test_metadata_dirty(cellfs, sci->sc_root))
		set_bit(CELLFS_SC_DIRTY, &sci->sc_flags);

	spin_lock(&cellfs->ns_inode_lock);
	if (list_empty(&cellfs->ns_dirty_files) && cellfs_segctor_clean(sci))
		ret++;

	spin_unlock(&cellfs->ns_inode_lock);
	return ret;
}

static void cellfs_segctor_clear_metadata_dirty(struct cellfs_sc_info *sci)
{
	struct the_cellfs *cellfs = sci->sc_super->s_fs_info;

	cellfs_mdt_clear_dirty(sci->sc_root->ifile);
	cellfs_mdt_clear_dirty(cellfs->ns_cpfile);
	cellfs_mdt_clear_dirty(cellfs->ns_sufile);
	cellfs_mdt_clear_dirty(cellfs->ns_dat);
}

static int cellfs_segctor_create_checkpoint(struct cellfs_sc_info *sci)
{
	struct the_cellfs *cellfs = sci->sc_super->s_fs_info;
	struct buffer_head *bh_cp;
	struct cellfs_checkpoint *raw_cp;
	int err;

	/* XXX: this interface will be changed */
	err = cellfs_cpfile_get_checkpoint(cellfs->ns_cpfile, cellfs->ns_cno, 1,
					  &raw_cp, &bh_cp);
	if (likely(!err)) {
		/*
		 * The following code is duplicated with cpfile.  But, it is
		 * needed to collect the checkpoint even if it was not newly
		 * created.
		 */
		mark_buffer_dirty(bh_cp);
		cellfs_mdt_mark_dirty(cellfs->ns_cpfile);
		cellfs_cpfile_put_checkpoint(
			cellfs->ns_cpfile, cellfs->ns_cno, bh_cp);
	} else
		WARN_ON(err == -EINVAL || err == -ENOENT);

	return err;
}

static int cellfs_segctor_fill_in_checkpoint(struct cellfs_sc_info *sci)
{
	struct the_cellfs *cellfs = sci->sc_super->s_fs_info;
	struct buffer_head *bh_cp;
	struct cellfs_checkpoint *raw_cp;
	int err;

	err = cellfs_cpfile_get_checkpoint(cellfs->ns_cpfile, cellfs->ns_cno, 0,
					  &raw_cp, &bh_cp);
	if (unlikely(err)) {
		WARN_ON(err == -EINVAL || err == -ENOENT);
		goto failed_ibh;
	}
	raw_cp->cp_snapshot_list.ssl_next = 0;
	raw_cp->cp_snapshot_list.ssl_prev = 0;
	raw_cp->cp_inodes_count =
		cpu_to_le64(atomic64_read(&sci->sc_root->inodes_count));
	raw_cp->cp_blocks_count =
		cpu_to_le64(atomic64_read(&sci->sc_root->blocks_count));
	raw_cp->cp_nblk_inc =
		cpu_to_le64(sci->sc_nblk_inc + sci->sc_nblk_this_inc);
	raw_cp->cp_create = cpu_to_le64(sci->sc_seg_ctime);
	raw_cp->cp_cno = cpu_to_le64(cellfs->ns_cno);

	if (test_bit(CELLFS_SC_HAVE_DELTA, &sci->sc_flags))
		cellfs_checkpoint_clear_minor(raw_cp);
	else
		cellfs_checkpoint_set_minor(raw_cp);

	cellfs_write_inode_common(sci->sc_root->ifile,
				 &raw_cp->cp_ifile_inode, 1);
	cellfs_cpfile_put_checkpoint(cellfs->ns_cpfile, cellfs->ns_cno, bh_cp);
	return 0;

 failed_ibh:
	return err;
}

static void cellfs_fill_in_file_bmap(struct inode *ifile,
				    struct cellfs_inode_info *ii)

{
	struct buffer_head *ibh;
	struct cellfs_inode *raw_inode;

	if (test_bit(CELLFS_I_BMAP, &ii->i_state)) {
		ibh = ii->i_bh;
		BUG_ON(!ibh);
		raw_inode = cellfs_ifile_map_inode(ifile, ii->vfs_inode.i_ino,
						  ibh);
		cellfs_bmap_write(ii->i_bmap, raw_inode);
		cellfs_ifile_unmap_inode(ifile, ii->vfs_inode.i_ino, ibh);
	}
}

static void cellfs_segctor_fill_in_file_bmap(struct cellfs_sc_info *sci)
{
	struct cellfs_inode_info *ii;

	list_for_each_entry(ii, &sci->sc_dirty_files, i_dirty) {
		cellfs_fill_in_file_bmap(sci->sc_root->ifile, ii);
		set_bit(CELLFS_I_COLLECTED, &ii->i_state);
	}
}

static void cellfs_segctor_fill_in_super_root(struct cellfs_sc_info *sci,
					     struct the_cellfs *cellfs)
{
	struct buffer_head *bh_sr;
	struct cellfs_super_root *raw_sr;
	unsigned int isz, srsz;

	bh_sr = CELLFS_LAST_SEGBUF(&sci->sc_segbufs)->sb_super_root;
	raw_sr = (struct cellfs_super_root *)bh_sr->b_data;
	isz = cellfs->ns_inode_size;
	srsz = CELLFS_SR_BYTES(isz);

	raw_sr->sr_bytes = cpu_to_le16(srsz);
	raw_sr->sr_nongc_ctime
		= cpu_to_le64(cellfs_doing_gc() ?
			      cellfs->ns_nongc_ctime : sci->sc_seg_ctime);
	raw_sr->sr_flags = 0;

	cellfs_write_inode_common(cellfs->ns_dat, (void *)raw_sr +
				 CELLFS_SR_DAT_OFFSET(isz), 1);
	cellfs_write_inode_common(cellfs->ns_cpfile, (void *)raw_sr +
				 CELLFS_SR_CPFILE_OFFSET(isz), 1);
	cellfs_write_inode_common(cellfs->ns_sufile, (void *)raw_sr +
				 CELLFS_SR_SUFILE_OFFSET(isz), 1);
	memset((void *)raw_sr + srsz, 0, cellfs->ns_blocksize - srsz);
}

static void cellfs_redirty_inodes(struct list_head *head)
{
	struct cellfs_inode_info *ii;

	list_for_each_entry(ii, head, i_dirty) {
		if (test_bit(CELLFS_I_COLLECTED, &ii->i_state))
			clear_bit(CELLFS_I_COLLECTED, &ii->i_state);
	}
}

static void cellfs_drop_collected_inodes(struct list_head *head)
{
	struct cellfs_inode_info *ii;

	list_for_each_entry(ii, head, i_dirty) {
		if (!test_and_clear_bit(CELLFS_I_COLLECTED, &ii->i_state))
			continue;

		clear_bit(CELLFS_I_INODE_SYNC, &ii->i_state);
		set_bit(CELLFS_I_UPDATED, &ii->i_state);
	}
}

static int cellfs_segctor_apply_buffers(struct cellfs_sc_info *sci,
				       struct inode *inode,
				       struct list_head *listp,
				       int (*collect)(struct cellfs_sc_info *,
						      struct buffer_head *,
						      struct inode *))
{
	struct buffer_head *bh, *n;
	int err = 0;

	if (collect) {
		list_for_each_entry_safe(bh, n, listp, b_assoc_buffers) {
			list_del_init(&bh->b_assoc_buffers);
			err = collect(sci, bh, inode);
			brelse(bh);
			if (unlikely(err))
				goto dispose_buffers;
		}
		return 0;
	}

 dispose_buffers:
	while (!list_empty(listp)) {
		bh = list_first_entry(listp, struct buffer_head,
				      b_assoc_buffers);
		list_del_init(&bh->b_assoc_buffers);
		brelse(bh);
	}
	return err;
}

static size_t cellfs_segctor_buffer_rest(struct cellfs_sc_info *sci)
{
	/* Remaining number of blocks within segment buffer */
	return sci->sc_segbuf_nblocks -
		(sci->sc_nblk_this_inc + sci->sc_curseg->sb_sum.nblocks);
}

static int cellfs_segctor_scan_file(struct cellfs_sc_info *sci,
				   struct inode *inode,
				   const struct cellfs_sc_operations *sc_ops)
{
	LIST_HEAD(data_buffers);
	LIST_HEAD(node_buffers);
	int err;

	if (!(sci->sc_stage.flags & CELLFS_CF_NODE)) {
		size_t n, rest = cellfs_segctor_buffer_rest(sci);

		n = cellfs_lookup_dirty_data_buffers(
			inode, &data_buffers, rest + 1, 0, LLONG_MAX);
		if (n > rest) {
			err = cellfs_segctor_apply_buffers(
				sci, inode, &data_buffers,
				sc_ops->collect_data);
			BUG_ON(!err); /* always receive -E2BIG or true error */
			goto break_or_fail;
		}
	}
	cellfs_lookup_dirty_node_buffers(inode, &node_buffers);

	if (!(sci->sc_stage.flags & CELLFS_CF_NODE)) {
		err = cellfs_segctor_apply_buffers(
			sci, inode, &data_buffers, sc_ops->collect_data);
		if (unlikely(err)) {
			/* dispose node list */
			cellfs_segctor_apply_buffers(
				sci, inode, &node_buffers, NULL);
			goto break_or_fail;
		}
		sci->sc_stage.flags |= CELLFS_CF_NODE;
	}
	/* Collect node */
	err = cellfs_segctor_apply_buffers(
		sci, inode, &node_buffers, sc_ops->collect_node);
	if (unlikely(err))
		goto break_or_fail;

	cellfs_bmap_lookup_dirty_buffers(CELLFS_I(inode)->i_bmap, &node_buffers);
	err = cellfs_segctor_apply_buffers(
		sci, inode, &node_buffers, sc_ops->collect_bmap);
	if (unlikely(err))
		goto break_or_fail;

	cellfs_segctor_end_finfo(sci, inode);
	sci->sc_stage.flags &= ~CELLFS_CF_NODE;

 break_or_fail:
	return err;
}

static int cellfs_segctor_scan_file_dsync(struct cellfs_sc_info *sci,
					 struct inode *inode)
{
	LIST_HEAD(data_buffers);
	size_t n, rest = cellfs_segctor_buffer_rest(sci);
	int err;

	n = cellfs_lookup_dirty_data_buffers(inode, &data_buffers, rest + 1,
					    sci->sc_dsync_start,
					    sci->sc_dsync_end);

	err = cellfs_segctor_apply_buffers(sci, inode, &data_buffers,
					  cellfs_collect_file_data);
	if (!err) {
		cellfs_segctor_end_finfo(sci, inode);
		BUG_ON(n > rest);
		/* always receive -E2BIG or true error if n > rest */
	}
	return err;
}

static int cellfs_segctor_collect_blocks(struct cellfs_sc_info *sci, int mode)
{
	struct the_cellfs *cellfs = sci->sc_super->s_fs_info;
	struct list_head *head;
	struct cellfs_inode_info *ii;
	size_t ndone;
	int err = 0;
	#pragma GCC diagnostic push
	#pragma GCC diagnostic ignored "-Wimplicit-fallthrough"
	switch (cellfs_sc_cstage_get(sci)) {
	case CELLFS_ST_INIT:
		/* Pre-processes */
		sci->sc_stage.flags = 0;

		if (!test_bit(CELLFS_SC_UNCLOSED, &sci->sc_flags)) {
			sci->sc_nblk_inc = 0;
			sci->sc_curseg->sb_sum.flags = CELLFS_SS_LOGBGN;
			if (mode == SC_LSEG_DSYNC) {
				cellfs_sc_cstage_set(sci, CELLFS_ST_DSYNC);
				goto dsync_mode;
			}
		}

		sci->sc_stage.dirty_file_ptr = NULL;
		sci->sc_stage.gc_inode_ptr = NULL;
		if (mode == SC_FLUSH_DAT) {
			cellfs_sc_cstage_set(sci, CELLFS_ST_DAT);
			goto dat_stage;
		}
		cellfs_sc_cstage_inc(sci);  /* Fall through */
	case CELLFS_ST_GC:
		if (cellfs_doing_gc()) {
			head = &sci->sc_gc_inodes;
			ii = list_prepare_entry(sci->sc_stage.gc_inode_ptr,
						head, i_dirty);
			list_for_each_entry_continue(ii, head, i_dirty) {
				err = cellfs_segctor_scan_file(
					sci, &ii->vfs_inode,
					&cellfs_sc_file_ops);
				if (unlikely(err)) {
					sci->sc_stage.gc_inode_ptr = list_entry(
						ii->i_dirty.prev,
						struct cellfs_inode_info,
						i_dirty);
					goto break_or_fail;
				}
				set_bit(CELLFS_I_COLLECTED, &ii->i_state);
			}
			sci->sc_stage.gc_inode_ptr = NULL;
		}
		cellfs_sc_cstage_inc(sci);  /* Fall through */
	case CELLFS_ST_FILE:
		head = &sci->sc_dirty_files;
		ii = list_prepare_entry(sci->sc_stage.dirty_file_ptr, head,
					i_dirty);
		list_for_each_entry_continue(ii, head, i_dirty) {
			clear_bit(CELLFS_I_DIRTY, &ii->i_state);

			err = cellfs_segctor_scan_file(sci, &ii->vfs_inode,
						      &cellfs_sc_file_ops);
			if (unlikely(err)) {
				sci->sc_stage.dirty_file_ptr =
					list_entry(ii->i_dirty.prev,
						   struct cellfs_inode_info,
						   i_dirty);
				goto break_or_fail;
			}
			/* sci->sc_stage.dirty_file_ptr = CELLFS_I(inode); */
			/* XXX: required ? */
		}
		sci->sc_stage.dirty_file_ptr = NULL;
		if (mode == SC_FLUSH_FILE) {
			cellfs_sc_cstage_set(sci, CELLFS_ST_DONE);
			return 0;
		}
		cellfs_sc_cstage_inc(sci);
		sci->sc_stage.flags |= CELLFS_CF_IFILE_STARTED;
		/* Fall through */
	case CELLFS_ST_IFILE:
		err = cellfs_segctor_scan_file(sci, sci->sc_root->ifile,
					      &cellfs_sc_file_ops);
		if (unlikely(err))
			break;
		cellfs_sc_cstage_inc(sci);
		/* Creating a checkpoint */
		err = cellfs_segctor_create_checkpoint(sci);
		if (unlikely(err))
			break;
		/* Fall through */
	case CELLFS_ST_CPFILE:
		err = cellfs_segctor_scan_file(sci, cellfs->ns_cpfile,
					      &cellfs_sc_file_ops);
		if (unlikely(err))
			break;
		cellfs_sc_cstage_inc(sci);  /* Fall through */
	case CELLFS_ST_SUFILE:
		err = cellfs_sufile_freev(cellfs->ns_sufile, sci->sc_freesegs,
					 sci->sc_nfreesegs, &ndone);
		if (unlikely(err)) {
			cellfs_sufile_cancel_freev(cellfs->ns_sufile,
						  sci->sc_freesegs, ndone,
						  NULL);
			break;
		}
		sci->sc_stage.flags |= CELLFS_CF_SUFREED;

		err = cellfs_segctor_scan_file(sci, cellfs->ns_sufile,
					      &cellfs_sc_file_ops);
		if (unlikely(err))
			break;
		cellfs_sc_cstage_inc(sci);  /* Fall through */
	case CELLFS_ST_DAT:
 dat_stage:
		err = cellfs_segctor_scan_file(sci, cellfs->ns_dat,
					      &cellfs_sc_dat_ops);
		if (unlikely(err))
			break;
		if (mode == SC_FLUSH_DAT) {
			cellfs_sc_cstage_set(sci, CELLFS_ST_DONE);
			return 0;
		}
		cellfs_sc_cstage_inc(sci);  /* Fall through */
	case CELLFS_ST_SR:
		if (mode == SC_LSEG_SR) {
			/* Appending a super root */
			err = cellfs_segctor_add_super_root(sci);
			if (unlikely(err))
				break;
		}
		/* End of a logical segment */
		sci->sc_curseg->sb_sum.flags |= CELLFS_SS_LOGEND;
		cellfs_sc_cstage_set(sci, CELLFS_ST_DONE);
		return 0;
	case CELLFS_ST_DSYNC:
 dsync_mode:
		sci->sc_curseg->sb_sum.flags |= CELLFS_SS_SYNDT;
		ii = sci->sc_dsync_inode;
		if (!test_bit(CELLFS_I_BUSY, &ii->i_state))
			break;

		err = cellfs_segctor_scan_file_dsync(sci, &ii->vfs_inode);
		if (unlikely(err))
			break;
		sci->sc_curseg->sb_sum.flags |= CELLFS_SS_LOGEND;
		cellfs_sc_cstage_set(sci, CELLFS_ST_DONE);
		return 0;
	case CELLFS_ST_DONE:
		return 0;
	default:
		BUG();
	}
	#pragma GCC diagnostic pop
 break_or_fail:
	return err;
}

/**
 * cellfs_segctor_begin_construction - setup segment buffer to make a new log
 * @sci: cellfs_sc_info
 * @cellfs: cellfs object
 */
static int cellfs_segctor_begin_construction(struct cellfs_sc_info *sci,
					    struct the_cellfs *cellfs)
{
	struct cellfs_segment_buffer *segbuf, *prev;
	__u64 nextnum;
	int err, alloc = 0;

	segbuf = cellfs_segbuf_new(sci->sc_super);
	if (unlikely(!segbuf))
		return -ENOMEM;

	if (list_empty(&sci->sc_write_logs)) {
		cellfs_segbuf_map(segbuf, cellfs->ns_segnum,
				 cellfs->ns_pseg_offset, cellfs);
		if (segbuf->sb_rest_blocks < CELLFS_PSEG_MIN_BLOCKS) {
			cellfs_shift_to_next_segment(cellfs);
			cellfs_segbuf_map(segbuf, cellfs->ns_segnum, 0, cellfs);
		}

		segbuf->sb_sum.seg_seq = cellfs->ns_seg_seq;
		nextnum = cellfs->ns_nextnum;

		if (cellfs->ns_segnum == cellfs->ns_nextnum)
			/* Start from the head of a new full segment */
			alloc++;
	} else {
		/* Continue logs */
		prev = CELLFS_LAST_SEGBUF(&sci->sc_write_logs);
		cellfs_segbuf_map_cont(segbuf, prev);
		segbuf->sb_sum.seg_seq = prev->sb_sum.seg_seq;
		nextnum = prev->sb_nextnum;

		if (segbuf->sb_rest_blocks < CELLFS_PSEG_MIN_BLOCKS) {
			cellfs_segbuf_map(segbuf, prev->sb_nextnum, 0, cellfs);
			segbuf->sb_sum.seg_seq++;
			alloc++;
		}
	}

	err = cellfs_sufile_mark_dirty(cellfs->ns_sufile, segbuf->sb_segnum);
	if (err)
		goto failed;

	if (alloc) {
		err = cellfs_sufile_alloc(cellfs->ns_sufile, &nextnum);
		if (err)
			goto failed;
	}
	cellfs_segbuf_set_next_segnum(segbuf, nextnum, cellfs);

	BUG_ON(!list_empty(&sci->sc_segbufs));
	list_add_tail(&segbuf->sb_list, &sci->sc_segbufs);
	sci->sc_segbuf_nblocks = segbuf->sb_rest_blocks;
	return 0;

 failed:
	cellfs_segbuf_free(segbuf);
	return err;
}

static int cellfs_segctor_extend_segments(struct cellfs_sc_info *sci,
					 struct the_cellfs *cellfs, int nadd)
{
	struct cellfs_segment_buffer *segbuf, *prev;
	struct inode *sufile = cellfs->ns_sufile;
	__u64 nextnextnum;
	LIST_HEAD(list);
	int err, ret, i;

	prev = CELLFS_LAST_SEGBUF(&sci->sc_segbufs);
	/*
	 * Since the segment specified with nextnum might be allocated during
	 * the previous construction, the buffer including its segusage may
	 * not be dirty.  The following call ensures that the buffer is dirty
	 * and will pin the buffer on memory until the sufile is written.
	 */
	err = cellfs_sufile_mark_dirty(sufile, prev->sb_nextnum);
	if (unlikely(err))
		return err;

	for (i = 0; i < nadd; i++) {
		/* extend segment info */
		err = -ENOMEM;
		segbuf = cellfs_segbuf_new(sci->sc_super);
		if (unlikely(!segbuf))
			goto failed;

		/* map this buffer to region of segment on-disk */
		cellfs_segbuf_map(segbuf, prev->sb_nextnum, 0, cellfs);
		sci->sc_segbuf_nblocks += segbuf->sb_rest_blocks;

		/* allocate the next next full segment */
		err = cellfs_sufile_alloc(sufile, &nextnextnum);
		if (unlikely(err))
			goto failed_segbuf;

		segbuf->sb_sum.seg_seq = prev->sb_sum.seg_seq + 1;
		cellfs_segbuf_set_next_segnum(segbuf, nextnextnum, cellfs);

		list_add_tail(&segbuf->sb_list, &list);
		prev = segbuf;
	}
	list_splice_tail(&list, &sci->sc_segbufs);
	return 0;

 failed_segbuf:
	cellfs_segbuf_free(segbuf);
 failed:
	list_for_each_entry(segbuf, &list, sb_list) {
		ret = cellfs_sufile_free(sufile, segbuf->sb_nextnum);
		WARN_ON(ret); /* never fails */
	}
	cellfs_destroy_logs(&list);
	return err;
}

static void cellfs_free_incomplete_logs(struct list_head *logs,
				       struct the_cellfs *cellfs)
{
	struct cellfs_segment_buffer *segbuf, *prev;
	struct inode *sufile = cellfs->ns_sufile;
	int ret;

	segbuf = CELLFS_FIRST_SEGBUF(logs);
	if (cellfs->ns_nextnum != segbuf->sb_nextnum) {
		ret = cellfs_sufile_free(sufile, segbuf->sb_nextnum);
		WARN_ON(ret); /* never fails */
	}
	if (atomic_read(&segbuf->sb_err)) {
		/* Case 1: The first segment failed */
		if (segbuf->sb_pseg_start != segbuf->sb_fseg_start)
			/*
			 * Case 1a:  Partial segment appended into an existing
			 * segment
			 */
			cellfs_terminate_segment(cellfs, segbuf->sb_fseg_start,
						segbuf->sb_fseg_end);
		else /* Case 1b:  New full segment */
			set_cellfs_discontinued(cellfs);
	}

	prev = segbuf;
	list_for_each_entry_continue(segbuf, logs, sb_list) {
		if (prev->sb_nextnum != segbuf->sb_nextnum) {
			ret = cellfs_sufile_free(sufile, segbuf->sb_nextnum);
			WARN_ON(ret); /* never fails */
		}
		if (atomic_read(&segbuf->sb_err) &&
		    segbuf->sb_segnum != cellfs->ns_nextnum)
			/* Case 2: extended segment (!= next) failed */
			cellfs_sufile_set_error(sufile, segbuf->sb_segnum);
		prev = segbuf;
	}
}

static void cellfs_segctor_update_segusage(struct cellfs_sc_info *sci,
					  struct inode *sufile)
{
	struct cellfs_segment_buffer *segbuf;
	unsigned long live_blocks;
	int ret;

	list_for_each_entry(segbuf, &sci->sc_segbufs, sb_list) {
		live_blocks = segbuf->sb_sum.nblocks +
			(segbuf->sb_pseg_start - segbuf->sb_fseg_start);
		ret = cellfs_sufile_set_segment_usage(sufile, segbuf->sb_segnum,
						     live_blocks,
						     sci->sc_seg_ctime);
		WARN_ON(ret); /* always succeed because the segusage is dirty */
	}
}

static void cellfs_cancel_segusage(struct list_head *logs, struct inode *sufile)
{
	struct cellfs_segment_buffer *segbuf;
	int ret;

	segbuf = CELLFS_FIRST_SEGBUF(logs);
	ret = cellfs_sufile_set_segment_usage(sufile, segbuf->sb_segnum,
					     segbuf->sb_pseg_start -
					     segbuf->sb_fseg_start, 0);
	WARN_ON(ret); /* always succeed because the segusage is dirty */

	list_for_each_entry_continue(segbuf, logs, sb_list) {
		ret = cellfs_sufile_set_segment_usage(sufile, segbuf->sb_segnum,
						     0, 0);
		WARN_ON(ret); /* always succeed */
	}
}

static void cellfs_segctor_truncate_segments(struct cellfs_sc_info *sci,
					    struct cellfs_segment_buffer *last,
					    struct inode *sufile)
{
	struct cellfs_segment_buffer *segbuf = last;
	int ret;

	list_for_each_entry_continue(segbuf, &sci->sc_segbufs, sb_list) {
		sci->sc_segbuf_nblocks -= segbuf->sb_rest_blocks;
		ret = cellfs_sufile_free(sufile, segbuf->sb_nextnum);
		WARN_ON(ret);
	}
	cellfs_truncate_logs(&sci->sc_segbufs, last);
}


static int cellfs_segctor_collect(struct cellfs_sc_info *sci,
				 struct the_cellfs *cellfs, int mode)
{
	struct cellfs_cstage prev_stage = sci->sc_stage;
	int err, nadd = 1;

	/* Collection retry loop */
	for (;;) {
		sci->sc_nblk_this_inc = 0;
		sci->sc_curseg = CELLFS_FIRST_SEGBUF(&sci->sc_segbufs);

		err = cellfs_segctor_reset_segment_buffer(sci);
		if (unlikely(err))
			goto failed;

		err = cellfs_segctor_collect_blocks(sci, mode);
		sci->sc_nblk_this_inc += sci->sc_curseg->sb_sum.nblocks;
		if (!err)
			break;

		if (unlikely(err != -E2BIG))
			goto failed;

		/* The current segment is filled up */
		if (mode != SC_LSEG_SR ||
		    cellfs_sc_cstage_get(sci) < CELLFS_ST_CPFILE)
			break;

		cellfs_clear_logs(&sci->sc_segbufs);

		if (sci->sc_stage.flags & CELLFS_CF_SUFREED) {
			err = cellfs_sufile_cancel_freev(cellfs->ns_sufile,
							sci->sc_freesegs,
							sci->sc_nfreesegs,
							NULL);
			WARN_ON(err); /* do not happen */
			sci->sc_stage.flags &= ~CELLFS_CF_SUFREED;
		}

		err = cellfs_segctor_extend_segments(sci, cellfs, nadd);
		if (unlikely(err))
			return err;

		nadd = min_t(int, nadd << 1, SC_MAX_SEGDELTA);
		sci->sc_stage = prev_stage;
	}
	cellfs_segctor_truncate_segments(sci, sci->sc_curseg, cellfs->ns_sufile);
	return 0;

 failed:
	return err;
}

static void cellfs_list_replace_buffer(struct buffer_head *old_bh,
				      struct buffer_head *new_bh)
{
	BUG_ON(!list_empty(&new_bh->b_assoc_buffers));

	list_replace_init(&old_bh->b_assoc_buffers, &new_bh->b_assoc_buffers);
	/* The caller must release old_bh */
}

static int
cellfs_segctor_update_payload_blocknr(struct cellfs_sc_info *sci,
				     struct cellfs_segment_buffer *segbuf,
				     int mode)
{
	struct inode *inode = NULL;
	sector_t blocknr;
	unsigned long nfinfo = segbuf->sb_sum.nfinfo;
	unsigned long nblocks = 0, ndatablk = 0;
	const struct cellfs_sc_operations *sc_op = NULL;
	struct cellfs_segsum_pointer ssp;
	struct cellfs_finfo *finfo = NULL;
	union cellfs_binfo binfo;
	struct buffer_head *bh, *bh_org;
	ino_t ino = 0;
	int err = 0;

	if (!nfinfo)
		goto out;

	blocknr = segbuf->sb_pseg_start + segbuf->sb_sum.nsumblk;
	ssp.bh = CELLFS_SEGBUF_FIRST_BH(&segbuf->sb_segsum_buffers);
	ssp.offset = sizeof(struct cellfs_segment_summary);

	list_for_each_entry(bh, &segbuf->sb_payload_buffers, b_assoc_buffers) {
		if (bh == segbuf->sb_super_root)
			break;
		if (!finfo) {
			finfo =	cellfs_segctor_map_segsum_entry(
				sci, &ssp, sizeof(*finfo));
			ino = le64_to_cpu(finfo->fi_ino);
			nblocks = le32_to_cpu(finfo->fi_nblocks);
			ndatablk = le32_to_cpu(finfo->fi_ndatablk);

			inode = bh->b_page->mapping->host;

			if (mode == SC_LSEG_DSYNC)
				sc_op = &cellfs_sc_dsync_ops;
			else if (ino == CELLFS_DAT_INO)
				sc_op = &cellfs_sc_dat_ops;
			else /* file blocks */
				sc_op = &cellfs_sc_file_ops;
		}
		bh_org = bh;
		get_bh(bh_org);
		err = cellfs_bmap_assign(CELLFS_I(inode)->i_bmap, &bh, blocknr,
					&binfo);
		if (bh != bh_org)
			cellfs_list_replace_buffer(bh_org, bh);
		brelse(bh_org);
		if (unlikely(err))
			goto failed_bmap;

		if (ndatablk > 0)
			sc_op->write_data_binfo(sci, &ssp, &binfo);
		else
			sc_op->write_node_binfo(sci, &ssp, &binfo);

		blocknr++;
		if (--nblocks == 0) {
			finfo = NULL;
			if (--nfinfo == 0)
				break;
		} else if (ndatablk > 0)
			ndatablk--;
	}
 out:
	return 0;

 failed_bmap:
	return err;
}

static int cellfs_segctor_assign(struct cellfs_sc_info *sci, int mode)
{
	struct cellfs_segment_buffer *segbuf;
	int err;

	list_for_each_entry(segbuf, &sci->sc_segbufs, sb_list) {
		err = cellfs_segctor_update_payload_blocknr(sci, segbuf, mode);
		if (unlikely(err))
			return err;
		cellfs_segbuf_fill_in_segsum(segbuf);
	}
	return 0;
}

static void cellfs_begin_page_io(struct page *page)
{
	if (!page || PageWriteback(page))
		/*
		 * For split b-tree node pages, this function may be called
		 * twice.  We ignore the 2nd or later calls by this check.
		 */
		return;

	lock_page(page);
	clear_page_dirty_for_io(page);
	set_page_writeback(page);
	unlock_page(page);
}

static void cellfs_segctor_prepare_write(struct cellfs_sc_info *sci)
{
	struct cellfs_segment_buffer *segbuf;
	struct page *bd_page = NULL, *fs_page = NULL;

	list_for_each_entry(segbuf, &sci->sc_segbufs, sb_list) {
		struct buffer_head *bh;

		list_for_each_entry(bh, &segbuf->sb_segsum_buffers,
				    b_assoc_buffers) {
			if (bh->b_page != bd_page) {
				if (bd_page) {
					lock_page(bd_page);
					clear_page_dirty_for_io(bd_page);
					set_page_writeback(bd_page);
					unlock_page(bd_page);
				}
				bd_page = bh->b_page;
			}
		}

		list_for_each_entry(bh, &segbuf->sb_payload_buffers,
				    b_assoc_buffers) {
			set_buffer_async_write(bh);
			if (bh == segbuf->sb_super_root) {
				if (bh->b_page != bd_page) {
					lock_page(bd_page);
					clear_page_dirty_for_io(bd_page);
					set_page_writeback(bd_page);
					unlock_page(bd_page);
					bd_page = bh->b_page;
				}
				break;
			}
			if (bh->b_page != fs_page) {
				cellfs_begin_page_io(fs_page);
				fs_page = bh->b_page;
			}
		}
	}
	if (bd_page) {
		lock_page(bd_page);
		clear_page_dirty_for_io(bd_page);
		set_page_writeback(bd_page);
		unlock_page(bd_page);
	}
	cellfs_begin_page_io(fs_page);
}

static int cellfs_segctor_write(struct cellfs_sc_info *sci,
			       struct the_cellfs *cellfs)
{
	int ret;

	ret = cellfs_write_logs(&sci->sc_segbufs, cellfs);
	list_splice_tail_init(&sci->sc_segbufs, &sci->sc_write_logs);
	return ret;
}

static void cellfs_end_page_io(struct page *page, int err)
{
	if (!page)
		return;

	if (buffer_cellfs_node(page_buffers(page)) && !PageWriteback(page)) {
		/*
		 * For b-tree node pages, this function may be called twice
		 * or more because they might be split in a segment.
		 */
		if (PageDirty(page)) {
			/*
			 * For pages holding split b-tree node buffers, dirty
			 * flag on the buffers may be cleared discretely.
			 * In that case, the page is once redirtied for
			 * remaining buffers, and it must be cancelled if
			 * all the buffers get cleaned later.
			 */
			lock_page(page);
			if (cellfs_page_buffers_clean(page))
				__cellfs_clear_page_dirty(page);
			unlock_page(page);
		}
		return;
	}

	if (!err) {
		if (!cellfs_page_buffers_clean(page))
			__set_page_dirty_nobuffers(page);
		ClearPageError(page);
	} else {
		__set_page_dirty_nobuffers(page);
		SetPageError(page);
	}

	end_page_writeback(page);
}

static void cellfs_abort_logs(struct list_head *logs, int err)
{
	struct cellfs_segment_buffer *segbuf;
	struct page *bd_page = NULL, *fs_page = NULL;
	struct buffer_head *bh;

	if (list_empty(logs))
		return;

	list_for_each_entry(segbuf, logs, sb_list) {
		list_for_each_entry(bh, &segbuf->sb_segsum_buffers,
				    b_assoc_buffers) {
			if (bh->b_page != bd_page) {
				if (bd_page)
					end_page_writeback(bd_page);
				bd_page = bh->b_page;
			}
		}

		list_for_each_entry(bh, &segbuf->sb_payload_buffers,
				    b_assoc_buffers) {
			clear_buffer_async_write(bh);
			if (bh == segbuf->sb_super_root) {
				if (bh->b_page != bd_page) {
					end_page_writeback(bd_page);
					bd_page = bh->b_page;
				}
				break;
			}
			if (bh->b_page != fs_page) {
				cellfs_end_page_io(fs_page, err);
				fs_page = bh->b_page;
			}
		}
	}
	if (bd_page)
		end_page_writeback(bd_page);

	cellfs_end_page_io(fs_page, err);
}

static void cellfs_segctor_abort_construction(struct cellfs_sc_info *sci,
					     struct the_cellfs *cellfs, int err)
{
	LIST_HEAD(logs);
	int ret;

	list_splice_tail_init(&sci->sc_write_logs, &logs);
	ret = cellfs_wait_on_logs(&logs);
	cellfs_abort_logs(&logs, ret ? : err);

	list_splice_tail_init(&sci->sc_segbufs, &logs);
	cellfs_cancel_segusage(&logs, cellfs->ns_sufile);
	cellfs_free_incomplete_logs(&logs, cellfs);

	if (sci->sc_stage.flags & CELLFS_CF_SUFREED) {
		ret = cellfs_sufile_cancel_freev(cellfs->ns_sufile,
						sci->sc_freesegs,
						sci->sc_nfreesegs,
						NULL);
		WARN_ON(ret); /* do not happen */
	}

	cellfs_destroy_logs(&logs);
}

static void cellfs_set_next_segment(struct the_cellfs *cellfs,
				   struct cellfs_segment_buffer *segbuf)
{
	cellfs->ns_segnum = segbuf->sb_segnum;
	cellfs->ns_nextnum = segbuf->sb_nextnum;
	cellfs->ns_pseg_offset = segbuf->sb_pseg_start - segbuf->sb_fseg_start
		+ segbuf->sb_sum.nblocks;
	cellfs->ns_seg_seq = segbuf->sb_sum.seg_seq;
	cellfs->ns_ctime = segbuf->sb_sum.ctime;
}

static void cellfs_segctor_complete_write(struct cellfs_sc_info *sci)
{
	struct cellfs_segment_buffer *segbuf;
	struct page *bd_page = NULL, *fs_page = NULL;
	struct the_cellfs *cellfs = sci->sc_super->s_fs_info;
	int update_sr = false;

	list_for_each_entry(segbuf, &sci->sc_write_logs, sb_list) {
		struct buffer_head *bh;

		list_for_each_entry(bh, &segbuf->sb_segsum_buffers,
				    b_assoc_buffers) {
			set_buffer_uptodate(bh);
			clear_buffer_dirty(bh);
			if (bh->b_page != bd_page) {
				if (bd_page)
					end_page_writeback(bd_page);
				bd_page = bh->b_page;
			}
		}
		/*
		 * We assume that the buffers which belong to the same page
		 * continue over the buffer list.
		 * Under this assumption, the last BHs of pages is
		 * identifiable by the discontinuity of bh->b_page
		 * (page != fs_page).
		 *
		 * For B-tree node blocks, however, this assumption is not
		 * guaranteed.  The cleanup code of B-tree node pages needs
		 * special care.
		 */
		list_for_each_entry(bh, &segbuf->sb_payload_buffers,
				    b_assoc_buffers) {
			const unsigned long set_bits = BIT(BH_Uptodate);
			const unsigned long clear_bits =
				(BIT(BH_Dirty) | BIT(BH_Async_Write) |
				 BIT(BH_Delay) | BIT(BH_CELLFS_Volatile) |
				 BIT(BH_CELLFS_Redirected));

			set_mask_bits(&bh->b_state, clear_bits, set_bits);
			if (bh == segbuf->sb_super_root) {
				if (bh->b_page != bd_page) {
					end_page_writeback(bd_page);
					bd_page = bh->b_page;
				}
				update_sr = true;
				break;
			}
			if (bh->b_page != fs_page) {
				cellfs_end_page_io(fs_page, 0);
				fs_page = bh->b_page;
			}
		}

		if (!cellfs_segbuf_simplex(segbuf)) {
			if (segbuf->sb_sum.flags & CELLFS_SS_LOGBGN) {
				set_bit(CELLFS_SC_UNCLOSED, &sci->sc_flags);
				sci->sc_lseg_stime = jiffies;
			}
			if (segbuf->sb_sum.flags & CELLFS_SS_LOGEND)
				clear_bit(CELLFS_SC_UNCLOSED, &sci->sc_flags);
		}
	}
	/*
	 * Since pages may continue over multiple segment buffers,
	 * end of the last page must be checked outside of the loop.
	 */
	if (bd_page)
		end_page_writeback(bd_page);

	cellfs_end_page_io(fs_page, 0);

	cellfs_drop_collected_inodes(&sci->sc_dirty_files);

	if (cellfs_doing_gc())
		cellfs_drop_collected_inodes(&sci->sc_gc_inodes);
	else
		cellfs->ns_nongc_ctime = sci->sc_seg_ctime;

	sci->sc_nblk_inc += sci->sc_nblk_this_inc;

	segbuf = CELLFS_LAST_SEGBUF(&sci->sc_write_logs);
	cellfs_set_next_segment(cellfs, segbuf);

	if (update_sr) {
		cellfs->ns_flushed_device = 0;
		cellfs_set_last_segment(cellfs, segbuf->sb_pseg_start,
				       segbuf->sb_sum.seg_seq, cellfs->ns_cno++);

		clear_bit(CELLFS_SC_HAVE_DELTA, &sci->sc_flags);
		clear_bit(CELLFS_SC_DIRTY, &sci->sc_flags);
		set_bit(CELLFS_SC_SUPER_ROOT, &sci->sc_flags);
		cellfs_segctor_clear_metadata_dirty(sci);
	} else
		clear_bit(CELLFS_SC_SUPER_ROOT, &sci->sc_flags);
}

static int cellfs_segctor_wait(struct cellfs_sc_info *sci)
{
	int ret;

	ret = cellfs_wait_on_logs(&sci->sc_write_logs);
	if (!ret) {
		cellfs_segctor_complete_write(sci);
		cellfs_destroy_logs(&sci->sc_write_logs);
	}
	return ret;
}

static int cellfs_segctor_collect_dirty_files(struct cellfs_sc_info *sci,
					     struct the_cellfs *cellfs)
{
	struct cellfs_inode_info *ii, *n;
	struct inode *ifile = sci->sc_root->ifile;

	spin_lock(&cellfs->ns_inode_lock);
 retry:
	list_for_each_entry_safe(ii, n, &cellfs->ns_dirty_files, i_dirty) {
		if (!ii->i_bh) {
			struct buffer_head *ibh;
			int err;

			spin_unlock(&cellfs->ns_inode_lock);
			err = cellfs_ifile_get_inode_block(
				ifile, ii->vfs_inode.i_ino, &ibh);
			if (unlikely(err)) {
				cellfs_msg(sci->sc_super, KERN_WARNING,
					  "log writer: error %d getting inode block (ino=%lu)",
					  err, ii->vfs_inode.i_ino);
				return err;
			}
			spin_lock(&cellfs->ns_inode_lock);
			if (likely(!ii->i_bh))
				ii->i_bh = ibh;
			else
				brelse(ibh);
			goto retry;
		}

		// Always redirty the buffer to avoid race condition
		mark_buffer_dirty(ii->i_bh);
		cellfs_mdt_mark_dirty(ifile);

		clear_bit(CELLFS_I_QUEUED, &ii->i_state);
		set_bit(CELLFS_I_BUSY, &ii->i_state);
		list_move_tail(&ii->i_dirty, &sci->sc_dirty_files);
	}
	spin_unlock(&cellfs->ns_inode_lock);

	return 0;
}

static void cellfs_segctor_drop_written_files(struct cellfs_sc_info *sci,
					     struct the_cellfs *cellfs)
{
	struct cellfs_inode_info *ii, *n;
	int during_mount = !(sci->sc_super->s_flags & SB_ACTIVE);
	int defer_iput = false;

	spin_lock(&cellfs->ns_inode_lock);
	list_for_each_entry_safe(ii, n, &sci->sc_dirty_files, i_dirty) {
		if (!test_and_clear_bit(CELLFS_I_UPDATED, &ii->i_state) ||
		    test_bit(CELLFS_I_DIRTY, &ii->i_state))
			continue;

		clear_bit(CELLFS_I_BUSY, &ii->i_state);
		brelse(ii->i_bh);
		ii->i_bh = NULL;
		list_del_init(&ii->i_dirty);
		if (!ii->vfs_inode.i_nlink || during_mount) {
			/*
			 * Defer calling iput() to avoid deadlocks if
			 * i_nlink == 0 or mount is not yet finished.
			 */
			list_add_tail(&ii->i_dirty, &sci->sc_iput_queue);
			defer_iput = true;
		} else {
			spin_unlock(&cellfs->ns_inode_lock);
			iput(&ii->vfs_inode);
			spin_lock(&cellfs->ns_inode_lock);
		}
	}
	spin_unlock(&cellfs->ns_inode_lock);

	if (defer_iput)
		schedule_work(&sci->sc_iput_work);
}

/*
 * Main procedure of segment constructor
 */
static int cellfs_segctor_do_construct(struct cellfs_sc_info *sci, int mode)
{
	struct the_cellfs *cellfs = sci->sc_super->s_fs_info;
	int err;

	cellfs_sc_cstage_set(sci, CELLFS_ST_INIT);
	sci->sc_cno = cellfs->ns_cno;

	err = cellfs_segctor_collect_dirty_files(sci, cellfs);
	if (unlikely(err))
		goto out;

	if (cellfs_test_metadata_dirty(cellfs, sci->sc_root))
		set_bit(CELLFS_SC_DIRTY, &sci->sc_flags);

	if (cellfs_segctor_clean(sci))
		goto out;

	do {
		sci->sc_stage.flags &= ~CELLFS_CF_HISTORY_MASK;

		err = cellfs_segctor_begin_construction(sci, cellfs);
		if (unlikely(err))
			goto out;

		/* Update time stamp */
		sci->sc_seg_ctime = ktime_get_real_seconds();

		err = cellfs_segctor_collect(sci, cellfs, mode);
		if (unlikely(err))
			goto failed;

		/* Avoid empty segment */
		if (cellfs_sc_cstage_get(sci) == CELLFS_ST_DONE &&
		    cellfs_segbuf_empty(sci->sc_curseg)) {
			cellfs_segctor_abort_construction(sci, cellfs, 1);
			goto out;
		}

		err = cellfs_segctor_assign(sci, mode);
		if (unlikely(err))
			goto failed;

		if (sci->sc_stage.flags & CELLFS_CF_IFILE_STARTED)
			cellfs_segctor_fill_in_file_bmap(sci);

		if (mode == SC_LSEG_SR &&
		    cellfs_sc_cstage_get(sci) >= CELLFS_ST_CPFILE) {
			err = cellfs_segctor_fill_in_checkpoint(sci);
			if (unlikely(err))
				goto failed_to_write;

			cellfs_segctor_fill_in_super_root(sci, cellfs);
		}
		cellfs_segctor_update_segusage(sci, cellfs->ns_sufile);

		/* Write partial segments */
		cellfs_segctor_prepare_write(sci);

		cellfs_add_checksums_on_logs(&sci->sc_segbufs,
					    cellfs->ns_crc_seed);

		err = cellfs_segctor_write(sci, cellfs);
		if (unlikely(err))
			goto failed_to_write;

		if (cellfs_sc_cstage_get(sci) == CELLFS_ST_DONE ||
		    cellfs->ns_blocksize_bits != PAGE_SHIFT) {
			/*
			 * At this point, we avoid double buffering
			 * for blocksize < pagesize because page dirty
			 * flag is turned off during write and dirty
			 * buffers are not properly collected for
			 * pages crossing over segments.
			 */
			err = cellfs_segctor_wait(sci);
			if (err)
				goto failed_to_write;
		}
	} while (cellfs_sc_cstage_get(sci) != CELLFS_ST_DONE);

 out:
	cellfs_segctor_drop_written_files(sci, cellfs);
	return err;

 failed_to_write:
	if (sci->sc_stage.flags & CELLFS_CF_IFILE_STARTED)
		cellfs_redirty_inodes(&sci->sc_dirty_files);

 failed:
	if (cellfs_doing_gc())
		cellfs_redirty_inodes(&sci->sc_gc_inodes);
	cellfs_segctor_abort_construction(sci, cellfs, err);
	goto out;
}

/**
 * cellfs_segctor_start_timer - set timer of background write
 * @sci: cellfs_sc_info
 *
 * If the timer has already been set, it ignores the new request.
 * This function MUST be called within a section locking the segment
 * semaphore.
 */
static void cellfs_segctor_start_timer(struct cellfs_sc_info *sci)
{
	spin_lock(&sci->sc_state_lock);
	if (!(sci->sc_state & CELLFS_SEGCTOR_COMMIT)) {
		sci->sc_timer.expires = jiffies + sci->sc_interval;
		add_timer(&sci->sc_timer);
		sci->sc_state |= CELLFS_SEGCTOR_COMMIT;
	}
	spin_unlock(&sci->sc_state_lock);
}

static void cellfs_segctor_do_flush(struct cellfs_sc_info *sci, int bn)
{
	spin_lock(&sci->sc_state_lock);
	if (!(sci->sc_flush_request & BIT(bn))) {
		unsigned long prev_req = sci->sc_flush_request;

		sci->sc_flush_request |= BIT(bn);
		if (!prev_req)
			wake_up(&sci->sc_wait_daemon);
	}
	spin_unlock(&sci->sc_state_lock);
}

/**
 * cellfs_flush_segment - trigger a segment construction for resource control
 * @sb: super block
 * @ino: inode number of the file to be flushed out.
 */
void cellfs_flush_segment(struct super_block *sb, ino_t ino)
{
	struct the_cellfs *cellfs = sb->s_fs_info;
	struct cellfs_sc_info *sci = cellfs->ns_writer;

	if (!sci || cellfs_doing_construction())
		return;
	cellfs_segctor_do_flush(sci, CELLFS_MDT_INODE(sb, ino) ? ino : 0);
					/* assign bit 0 to data files */
}

struct cellfs_segctor_wait_request {
	wait_queue_entry_t	wq;
	__u32		seq;
	int		err;
	atomic_t	done;
};

static int cellfs_segctor_sync(struct cellfs_sc_info *sci)
{
	struct cellfs_segctor_wait_request wait_req;
	int err = 0;

	spin_lock(&sci->sc_state_lock);
	init_wait(&wait_req.wq);
	wait_req.err = 0;
	atomic_set(&wait_req.done, 0);
	wait_req.seq = ++sci->sc_seq_request;
	spin_unlock(&sci->sc_state_lock);

	init_waitqueue_entry(&wait_req.wq, current);
	add_wait_queue(&sci->sc_wait_request, &wait_req.wq);
	set_current_state(TASK_INTERRUPTIBLE);
	wake_up(&sci->sc_wait_daemon);

	for (;;) {
		if (atomic_read(&wait_req.done)) {
			err = wait_req.err;
			break;
		}
		if (!signal_pending(current)) {
			schedule();
			continue;
		}
		err = -ERESTARTSYS;
		break;
	}
	finish_wait(&sci->sc_wait_request, &wait_req.wq);
	return err;
}

static void cellfs_segctor_wakeup(struct cellfs_sc_info *sci, int err)
{
	struct cellfs_segctor_wait_request *wrq, *n;
	unsigned long flags;

	spin_lock_irqsave(&sci->sc_wait_request.lock, flags);
	list_for_each_entry_safe(wrq, n, &sci->sc_wait_request.head, wq.entry) {
		if (!atomic_read(&wrq->done) &&
		    cellfs_cnt32_ge(sci->sc_seq_done, wrq->seq)) {
			wrq->err = err;
			atomic_set(&wrq->done, 1);
		}
		if (atomic_read(&wrq->done)) {
			wrq->wq.func(&wrq->wq,
				     TASK_UNINTERRUPTIBLE | TASK_INTERRUPTIBLE,
				     0, NULL);
		}
	}
	spin_unlock_irqrestore(&sci->sc_wait_request.lock, flags);
}

/**
 * cellfs_construct_segment - construct a logical segment
 * @sb: super block
 *
 * Return Value: On success, 0 is retured. On errors, one of the following
 * negative error code is returned.
 *
 * %-EROFS - Read only filesystem.
 *
 * %-EIO - I/O error
 *
 * %-ENOSPC - No space left on device (only in a panic state).
 *
 * %-ERESTARTSYS - Interrupted.
 *
 * %-ENOMEM - Insufficient memory available.
 */
int cellfs_construct_segment(struct super_block *sb)
{
	struct the_cellfs *cellfs = sb->s_fs_info;
	struct cellfs_sc_info *sci = cellfs->ns_writer;
	struct cellfs_transaction_info *ti;
	int err;

	if (!sci)
		return -EROFS;

	/* A call inside transactions causes a deadlock. */
	BUG_ON((ti = current->journal_info) && ti->ti_magic == CELLFS_TI_MAGIC);

	err = cellfs_segctor_sync(sci);
	return err;
}

/**
 * cellfs_construct_dsync_segment - construct a data-only logical segment
 * @sb: super block
 * @inode: inode whose data blocks should be written out
 * @start: start byte offset
 * @end: end byte offset (inclusive)
 *
 * Return Value: On success, 0 is retured. On errors, one of the following
 * negative error code is returned.
 *
 * %-EROFS - Read only filesystem.
 *
 * %-EIO - I/O error
 *
 * %-ENOSPC - No space left on device (only in a panic state).
 *
 * %-ERESTARTSYS - Interrupted.
 *
 * %-ENOMEM - Insufficient memory available.
 */
int cellfs_construct_dsync_segment(struct super_block *sb, struct inode *inode,
				  loff_t start, loff_t end)
{
	struct the_cellfs *cellfs = sb->s_fs_info;
	struct cellfs_sc_info *sci = cellfs->ns_writer;
	struct cellfs_inode_info *ii;
	struct cellfs_transaction_info ti;
	int err = 0;

	if (!sci)
		return -EROFS;

	cellfs_transaction_lock(sb, &ti, 0);

	ii = CELLFS_I(inode);
	if (test_bit(CELLFS_I_INODE_SYNC, &ii->i_state) ||
	    cellfs_test_opt(cellfs, STRICT_ORDER) ||
	    test_bit(CELLFS_SC_UNCLOSED, &sci->sc_flags) ||
	    cellfs_discontinued(cellfs)) {
		cellfs_transaction_unlock(sb);
		err = cellfs_segctor_sync(sci);
		return err;
	}

	spin_lock(&cellfs->ns_inode_lock);
	if (!test_bit(CELLFS_I_QUEUED, &ii->i_state) &&
	    !test_bit(CELLFS_I_BUSY, &ii->i_state)) {
		spin_unlock(&cellfs->ns_inode_lock);
		cellfs_transaction_unlock(sb);
		return 0;
	}
	spin_unlock(&cellfs->ns_inode_lock);
	sci->sc_dsync_inode = ii;
	sci->sc_dsync_start = start;
	sci->sc_dsync_end = end;

	err = cellfs_segctor_do_construct(sci, SC_LSEG_DSYNC);
	if (!err)
		cellfs->ns_flushed_device = 0;

	cellfs_transaction_unlock(sb);
	return err;
}

#define FLUSH_FILE_BIT	(0x1) /* data file only */
#define FLUSH_DAT_BIT	BIT(CELLFS_DAT_INO) /* DAT only */

/**
 * cellfs_segctor_accept - record accepted sequence count of log-write requests
 * @sci: segment constructor object
 */
static void cellfs_segctor_accept(struct cellfs_sc_info *sci)
{
	spin_lock(&sci->sc_state_lock);
	sci->sc_seq_accepted = sci->sc_seq_request;
	spin_unlock(&sci->sc_state_lock);
	del_timer_sync(&sci->sc_timer);
}

/**
 * cellfs_segctor_notify - notify the result of request to caller threads
 * @sci: segment constructor object
 * @mode: mode of log forming
 * @err: error code to be notified
 */
static void cellfs_segctor_notify(struct cellfs_sc_info *sci, int mode, int err)
{
	/* Clear requests (even when the construction failed) */
	spin_lock(&sci->sc_state_lock);

	if (mode == SC_LSEG_SR) {
		sci->sc_state &= ~CELLFS_SEGCTOR_COMMIT;
		sci->sc_seq_done = sci->sc_seq_accepted;
		cellfs_segctor_wakeup(sci, err);
		sci->sc_flush_request = 0;
	} else {
		if (mode == SC_FLUSH_FILE)
			sci->sc_flush_request &= ~FLUSH_FILE_BIT;
		else if (mode == SC_FLUSH_DAT)
			sci->sc_flush_request &= ~FLUSH_DAT_BIT;

		/* re-enable timer if checkpoint creation was not done */
		if ((sci->sc_state & CELLFS_SEGCTOR_COMMIT) &&
		    time_before(jiffies, sci->sc_timer.expires))
			add_timer(&sci->sc_timer);
	}
	spin_unlock(&sci->sc_state_lock);
}

/**
 * cellfs_segctor_construct - form logs and write them to disk
 * @sci: segment constructor object
 * @mode: mode of log forming
 */
static int cellfs_segctor_construct(struct cellfs_sc_info *sci, int mode)
{
	struct the_cellfs *cellfs = sci->sc_super->s_fs_info;
	struct cellfs_super_block **sbp;
	int err = 0;

	cellfs_segctor_accept(sci);

	if (cellfs_discontinued(cellfs))
		mode = SC_LSEG_SR;
	if (!cellfs_segctor_confirm(sci))
		err = cellfs_segctor_do_construct(sci, mode);

	if (likely(!err)) {
		if (mode != SC_FLUSH_DAT)
			atomic_set(&cellfs->ns_ndirtyblks, 0);
		if (test_bit(CELLFS_SC_SUPER_ROOT, &sci->sc_flags) &&
		    cellfs_discontinued(cellfs)) {
			down_write(&cellfs->ns_sem);
			err = -EIO;
			sbp = cellfs_prepare_super(sci->sc_super,
						  cellfs_sb_will_flip(cellfs));
			if (likely(sbp)) {
				cellfs_set_log_cursor(sbp[0], cellfs);
				err = cellfs_commit_super(sci->sc_super,
							 CELLFS_SB_COMMIT);
			}
			up_write(&cellfs->ns_sem);
		}
	}

	cellfs_segctor_notify(sci, mode, err);
	return err;
}

static void cellfs_construction_timeout(struct timer_list *t)
{
	struct cellfs_sc_info *sci = from_timer(sci, t, sc_timer);

	wake_up_process(sci->sc_timer_task);
}

static void
cellfs_remove_written_gcinodes(struct the_cellfs *cellfs, struct list_head *head)
{
	struct cellfs_inode_info *ii, *n;

	list_for_each_entry_safe(ii, n, head, i_dirty) {
		if (!test_bit(CELLFS_I_UPDATED, &ii->i_state))
			continue;
		list_del_init(&ii->i_dirty);
		truncate_inode_pages(&ii->vfs_inode.i_data, 0);
		cellfs_btnode_cache_clear(&ii->i_btnode_cache);
		iput(&ii->vfs_inode);
	}
}

int cellfs_clean_segments(struct super_block *sb, struct cellfs_argv *argv,
			 void **kbufs)
{
	struct the_cellfs *cellfs = sb->s_fs_info;
	struct cellfs_sc_info *sci = cellfs->ns_writer;
	struct cellfs_transaction_info ti;
	int err;

	if (unlikely(!sci))
		return -EROFS;

	cellfs_transaction_lock(sb, &ti, 1);

	err = cellfs_mdt_save_to_shadow_map(cellfs->ns_dat);
	if (unlikely(err))
		goto out_unlock;

	err = cellfs_ioctl_prepare_clean_segments(cellfs, argv, kbufs);
	if (unlikely(err)) {
		cellfs_mdt_restore_from_shadow_map(cellfs->ns_dat);
		goto out_unlock;
	}

	sci->sc_freesegs = kbufs[4];
	sci->sc_nfreesegs = argv[4].v_nmembs;
	list_splice_tail_init(&cellfs->ns_gc_inodes, &sci->sc_gc_inodes);

	for (;;) {
		err = cellfs_segctor_construct(sci, SC_LSEG_SR);
		cellfs_remove_written_gcinodes(cellfs, &sci->sc_gc_inodes);

		if (likely(!err))
			break;

		cellfs_msg(sb, KERN_WARNING, "error %d cleaning segments", err);
		set_current_state(TASK_INTERRUPTIBLE);
		schedule_timeout(sci->sc_interval);
	}
	if (cellfs_test_opt(cellfs, DISCARD)) {
		int ret = cellfs_discard_segments(cellfs, sci->sc_freesegs,
						 sci->sc_nfreesegs);
		if (ret) {
			cellfs_msg(sb, KERN_WARNING,
				  "error %d on discard request, turning discards off for the device",
				  ret);
			cellfs_clear_opt(cellfs, DISCARD);
		}
	}

 out_unlock:
	sci->sc_freesegs = NULL;
	sci->sc_nfreesegs = 0;
	cellfs_mdt_clear_shadow_map(cellfs->ns_dat);
	cellfs_transaction_unlock(sb);
	return err;
}

static void cellfs_segctor_thread_construct(struct cellfs_sc_info *sci, int mode)
{
	struct cellfs_transaction_info ti;

	cellfs_transaction_lock(sci->sc_super, &ti, 0);
	cellfs_segctor_construct(sci, mode);

	/*
	 * Unclosed segment should be retried.  We do this using sc_timer.
	 * Timeout of sc_timer will invoke complete construction which leads
	 * to close the current logical segment.
	 */
	if (test_bit(CELLFS_SC_UNCLOSED, &sci->sc_flags))
		cellfs_segctor_start_timer(sci);

	cellfs_transaction_unlock(sci->sc_super);
}

static void cellfs_segctor_do_immediate_flush(struct cellfs_sc_info *sci)
{
	int mode = 0;

	spin_lock(&sci->sc_state_lock);
	mode = (sci->sc_flush_request & FLUSH_DAT_BIT) ?
		SC_FLUSH_DAT : SC_FLUSH_FILE;
	spin_unlock(&sci->sc_state_lock);

	if (mode) {
		cellfs_segctor_do_construct(sci, mode);

		spin_lock(&sci->sc_state_lock);
		sci->sc_flush_request &= (mode == SC_FLUSH_FILE) ?
			~FLUSH_FILE_BIT : ~FLUSH_DAT_BIT;
		spin_unlock(&sci->sc_state_lock);
	}
	clear_bit(CELLFS_SC_PRIOR_FLUSH, &sci->sc_flags);
}

static int cellfs_segctor_flush_mode(struct cellfs_sc_info *sci)
{
	if (!test_bit(CELLFS_SC_UNCLOSED, &sci->sc_flags) ||
	    time_before(jiffies, sci->sc_lseg_stime + sci->sc_mjcp_freq)) {
		if (!(sci->sc_flush_request & ~FLUSH_FILE_BIT))
			return SC_FLUSH_FILE;
		else if (!(sci->sc_flush_request & ~FLUSH_DAT_BIT))
			return SC_FLUSH_DAT;
	}
	return SC_LSEG_SR;
}

/**
 * cellfs_segctor_thread - main loop of the segment constructor thread.
 * @arg: pointer to a struct cellfs_sc_info.
 *
 * cellfs_segctor_thread() initializes a timer and serves as a daemon
 * to execute segment constructions.
 */
static int cellfs_segctor_thread(void *arg)
{
	struct cellfs_sc_info *sci = (struct cellfs_sc_info *)arg;
	struct the_cellfs *cellfs = sci->sc_super->s_fs_info;
	int timeout = 0;

	sci->sc_timer_task = current;

	/* start sync. */
	sci->sc_task = current;
	wake_up(&sci->sc_wait_task); /* for cellfs_segctor_start_thread() */
	cellfs_msg(sci->sc_super, KERN_INFO,
		  "segctord starting. Construction interval = %lu seconds, CP frequency < %lu seconds",
		  sci->sc_interval / HZ, sci->sc_mjcp_freq / HZ);

	spin_lock(&sci->sc_state_lock);
 loop:
	for (;;) {
		int mode;

		if (sci->sc_state & CELLFS_SEGCTOR_QUIT)
			goto end_thread;

		if (timeout || sci->sc_seq_request != sci->sc_seq_done)
			mode = SC_LSEG_SR;
		else if (sci->sc_flush_request)
			mode = cellfs_segctor_flush_mode(sci);
		else
			break;

		spin_unlock(&sci->sc_state_lock);
		cellfs_segctor_thread_construct(sci, mode);
		spin_lock(&sci->sc_state_lock);
		timeout = 0;
	}


	if (freezing(current)) {
		spin_unlock(&sci->sc_state_lock);
		try_to_freeze();
		spin_lock(&sci->sc_state_lock);
	} else {
		DEFINE_WAIT(wait);
		int should_sleep = 1;

		prepare_to_wait(&sci->sc_wait_daemon, &wait,
				TASK_INTERRUPTIBLE);

		if (sci->sc_seq_request != sci->sc_seq_done)
			should_sleep = 0;
		else if (sci->sc_flush_request)
			should_sleep = 0;
		else if (sci->sc_state & CELLFS_SEGCTOR_COMMIT)
			should_sleep = time_before(jiffies,
					sci->sc_timer.expires);

		if (should_sleep) {
			spin_unlock(&sci->sc_state_lock);
			schedule();
			spin_lock(&sci->sc_state_lock);
		}
		finish_wait(&sci->sc_wait_daemon, &wait);
		timeout = ((sci->sc_state & CELLFS_SEGCTOR_COMMIT) &&
			   time_after_eq(jiffies, sci->sc_timer.expires));

		if (cellfs_sb_dirty(cellfs) && cellfs_sb_need_update(cellfs))
			set_cellfs_discontinued(cellfs);
	}
	goto loop;

 end_thread:
	spin_unlock(&sci->sc_state_lock);

	/* end sync. */
	sci->sc_task = NULL;
	wake_up(&sci->sc_wait_task); /* for cellfs_segctor_kill_thread() */
	return 0;
}

static int cellfs_segctor_start_thread(struct cellfs_sc_info *sci)
{
	struct task_struct *t;

	t = kthread_run(cellfs_segctor_thread, sci, "segctord");
	if (IS_ERR(t)) {
		int err = PTR_ERR(t);

		cellfs_msg(sci->sc_super, KERN_ERR,
			  "error %d creating segctord thread", err);
		return err;
	}
	wait_event(sci->sc_wait_task, sci->sc_task != NULL);
	return 0;
}

static void cellfs_segctor_kill_thread(struct cellfs_sc_info *sci)
	__acquires(&sci->sc_state_lock)
	__releases(&sci->sc_state_lock)
{
	sci->sc_state |= CELLFS_SEGCTOR_QUIT;

	while (sci->sc_task) {
		wake_up(&sci->sc_wait_daemon);
		spin_unlock(&sci->sc_state_lock);
		wait_event(sci->sc_wait_task, sci->sc_task == NULL);
		spin_lock(&sci->sc_state_lock);
	}
}

/*
 * Setup & clean-up functions
 */
static struct cellfs_sc_info *cellfs_segctor_new(struct super_block *sb,
					       struct cellfs_root *root)
{
	struct the_cellfs *cellfs = sb->s_fs_info;
	struct cellfs_sc_info *sci;

	sci = kzalloc(sizeof(*sci), GFP_KERNEL);
	if (!sci)
		return NULL;

	sci->sc_super = sb;

	cellfs_get_root(root);
	sci->sc_root = root;

	init_waitqueue_head(&sci->sc_wait_request);
	init_waitqueue_head(&sci->sc_wait_daemon);
	init_waitqueue_head(&sci->sc_wait_task);
	spin_lock_init(&sci->sc_state_lock);
	INIT_LIST_HEAD(&sci->sc_dirty_files);
	INIT_LIST_HEAD(&sci->sc_segbufs);
	INIT_LIST_HEAD(&sci->sc_write_logs);
	INIT_LIST_HEAD(&sci->sc_gc_inodes);
	INIT_LIST_HEAD(&sci->sc_iput_queue);
	INIT_WORK(&sci->sc_iput_work, cellfs_iput_work_func);
	timer_setup(&sci->sc_timer, cellfs_construction_timeout, 0);

	sci->sc_interval = HZ * CELLFS_SC_DEFAULT_TIMEOUT;
	sci->sc_mjcp_freq = HZ * CELLFS_SC_DEFAULT_SR_FREQ;
	sci->sc_watermark = CELLFS_SC_DEFAULT_WATERMARK;

	if (cellfs->ns_interval)
		sci->sc_interval = HZ * cellfs->ns_interval;
	if (cellfs->ns_watermark)
		sci->sc_watermark = cellfs->ns_watermark;
	return sci;
}

static void cellfs_segctor_write_out(struct cellfs_sc_info *sci)
{
	int ret, retrycount = CELLFS_SC_CLEANUP_RETRY;

	/*
	 * The segctord thread was stopped and its timer was removed.
	 * But some tasks remain.
	 */
	do {
		struct cellfs_transaction_info ti;

		cellfs_transaction_lock(sci->sc_super, &ti, 0);
		ret = cellfs_segctor_construct(sci, SC_LSEG_SR);
		cellfs_transaction_unlock(sci->sc_super);

		flush_work(&sci->sc_iput_work);

	} while (ret && retrycount-- > 0);
}

/**
 * cellfs_segctor_destroy - destroy the segment constructor.
 * @sci: cellfs_sc_info
 *
 * cellfs_segctor_destroy() kills the segctord thread and frees
 * the cellfs_sc_info struct.
 * Caller must hold the segment semaphore.
 */
static void cellfs_segctor_destroy(struct cellfs_sc_info *sci)
{
	struct the_cellfs *cellfs = sci->sc_super->s_fs_info;
	int flag;

	up_write(&cellfs->ns_segctor_sem);

	spin_lock(&sci->sc_state_lock);
	cellfs_segctor_kill_thread(sci);
	flag = ((sci->sc_state & CELLFS_SEGCTOR_COMMIT) || sci->sc_flush_request
		|| sci->sc_seq_request != sci->sc_seq_done);
	spin_unlock(&sci->sc_state_lock);

	if (flush_work(&sci->sc_iput_work))
		flag = true;

	if (flag || !cellfs_segctor_confirm(sci))
		cellfs_segctor_write_out(sci);

	if (!list_empty(&sci->sc_dirty_files)) {
		cellfs_msg(sci->sc_super, KERN_WARNING,
			  "disposed unprocessed dirty file(s) when stopping log writer");
		cellfs_dispose_list(cellfs, &sci->sc_dirty_files, 1);
	}

	if (!list_empty(&sci->sc_iput_queue)) {
		cellfs_msg(sci->sc_super, KERN_WARNING,
			  "disposed unprocessed inode(s) in iput queue when stopping log writer");
		cellfs_dispose_list(cellfs, &sci->sc_iput_queue, 1);
	}

	WARN_ON(!list_empty(&sci->sc_segbufs));
	WARN_ON(!list_empty(&sci->sc_write_logs));

	cellfs_put_root(sci->sc_root);

	down_write(&cellfs->ns_segctor_sem);

	del_timer_sync(&sci->sc_timer);
	kfree(sci);
}

/**
 * cellfs_attach_log_writer - attach log writer
 * @sb: super block instance
 * @root: root object of the current filesystem tree
 *
 * This allocates a log writer object, initializes it, and starts the
 * log writer.
 *
 * Return Value: On success, 0 is returned. On error, one of the following
 * negative error code is returned.
 *
 * %-ENOMEM - Insufficient memory available.
 */
int cellfs_attach_log_writer(struct super_block *sb, struct cellfs_root *root)
{
	struct the_cellfs *cellfs = sb->s_fs_info;
	int err;

	if (cellfs->ns_writer) {
		/*
		 * This happens if the filesystem was remounted
		 * read/write after cellfs_error degenerated it into a
		 * read-only mount.
		 */
		cellfs_detach_log_writer(sb);
	}

	cellfs->ns_writer = cellfs_segctor_new(sb, root);
	if (!cellfs->ns_writer)
		return -ENOMEM;

	inode_attach_wb(cellfs->ns_bdev->bd_inode, NULL);

	err = cellfs_segctor_start_thread(cellfs->ns_writer);
	if (err) {
		kfree(cellfs->ns_writer);
		cellfs->ns_writer = NULL;
	}
	return err;
}

/**
 * cellfs_detach_log_writer - destroy log writer
 * @sb: super block instance
 *
 * This kills log writer daemon, frees the log writer object, and
 * destroys list of dirty files.
 */
void cellfs_detach_log_writer(struct super_block *sb)
{
	struct the_cellfs *cellfs = sb->s_fs_info;
	LIST_HEAD(garbage_list);

	down_write(&cellfs->ns_segctor_sem);
	if (cellfs->ns_writer) {
		cellfs_segctor_destroy(cellfs->ns_writer);
		cellfs->ns_writer = NULL;
	}

	/* Force to free the list of dirty files */
	spin_lock(&cellfs->ns_inode_lock);
	if (!list_empty(&cellfs->ns_dirty_files)) {
		list_splice_init(&cellfs->ns_dirty_files, &garbage_list);
		cellfs_msg(sb, KERN_WARNING,
			  "disposed unprocessed dirty file(s) when detaching log writer");
	}
	spin_unlock(&cellfs->ns_inode_lock);
	up_write(&cellfs->ns_segctor_sem);

	cellfs_dispose_list(cellfs, &garbage_list, 1);
}
