// SPDX-License-Identifier: GPL-2.0+
/*
 * inode.c - CELLFS inode operations.
 *
 * Copyright (C) 2005-2008 Nippon Telegraph and Telephone Corporation.
 *
 * Written by Ryusuke Konishi.
 *
 */

#include <linux/buffer_head.h>
#include <linux/gfp.h>
#include <linux/mpage.h>
#include <linux/pagemap.h>
#include <linux/writeback.h>
#include <linux/uio.h>
#include <linux/fiemap.h>
#include "cellfs.h"
#include "btnode.h"
#include "segment.h"
#include "page.h"
#include "mdt.h"
#include "cpfile.h"
#include "ifile.h"

/**
 * struct cellfs_iget_args - arguments used during comparison between inodes
 * @ino: inode number
 * @cno: checkpoint number
 * @root: pointer on CELLFS root object (mounted checkpoint)
 * @for_gc: inode for GC flag
 */
struct cellfs_iget_args {
	u64 ino;
	__u64 cno;
	struct cellfs_root *root;
	int for_gc;
};

static int cellfs_iget_test(struct inode *inode, void *opaque);

void cellfs_inode_add_blocks(struct inode *inode, int n)
{
	struct cellfs_root *root = CELLFS_I(inode)->i_root;

	inode_add_bytes(inode, i_blocksize(inode) * n);
	if (root)
		atomic64_add(n, &root->blocks_count);
}

void cellfs_inode_sub_blocks(struct inode *inode, int n)
{
	struct cellfs_root *root = CELLFS_I(inode)->i_root;

	inode_sub_bytes(inode, i_blocksize(inode) * n);
	if (root)
		atomic64_sub(n, &root->blocks_count);
}

/**
 * cellfs_get_block() - get a file block on the filesystem (callback function)
 * @inode - inode struct of the target file
 * @blkoff - file block number
 * @bh_result - buffer head to be mapped on
 * @create - indicate whether allocating the block or not when it has not
 *      been allocated yet.
 *
 * This function does not issue actual read request of the specified data
 * block. It is done by VFS.
 */
int cellfs_get_block(struct inode *inode, sector_t blkoff,
		    struct buffer_head *bh_result, int create)
{
	struct cellfs_inode_info *ii = CELLFS_I(inode);
	struct the_cellfs *cellfs = inode->i_sb->s_fs_info;
	__u64 blknum = 0;
	int err = 0, ret;
	unsigned int maxblocks = bh_result->b_size >> inode->i_blkbits;

	down_read(&CELLFS_MDT(cellfs->ns_dat)->mi_sem);
	ret = cellfs_bmap_lookup_contig(ii->i_bmap, blkoff, &blknum, maxblocks);
	up_read(&CELLFS_MDT(cellfs->ns_dat)->mi_sem);
	if (ret >= 0) {	/* found */
		map_bh(bh_result, inode->i_sb, blknum);
		if (ret > 0)
			bh_result->b_size = (ret << inode->i_blkbits);
		goto out;
	}
	/* data block was not found */
	if (ret == -ENOENT && create) {
		struct cellfs_transaction_info ti;

		bh_result->b_blocknr = 0;
		err = cellfs_transaction_begin(inode->i_sb, &ti, 1);
		if (unlikely(err))
			goto out;
		err = cellfs_bmap_insert(ii->i_bmap, blkoff,
					(unsigned long)bh_result);
		if (unlikely(err != 0)) {
			if (err == -EEXIST) {
				/*
				 * The get_block() function could be called
				 * from multiple callers for an inode.
				 * However, the page having this block must
				 * be locked in this case.
				 */
				cellfs_msg(inode->i_sb, KERN_WARNING,
					  "%s (ino=%lu): a race condition while inserting a data block at offset=%llu",
					  __func__, inode->i_ino,
					  (unsigned long long)blkoff);
				err = 0;
			}
			cellfs_transaction_abort(inode->i_sb);
			goto out;
		}
		cellfs_mark_inode_dirty_sync(inode);
		cellfs_transaction_commit(inode->i_sb); /* never fails */
		/* Error handling should be detailed */
		set_buffer_new(bh_result);
		set_buffer_delay(bh_result);
		map_bh(bh_result, inode->i_sb, 0);
		/* Disk block number must be changed to proper value */

	} else if (ret == -ENOENT) {
		/*
		 * not found is not error (e.g. hole); must return without
		 * the mapped state flag.
		 */
		;
	} else {
		err = ret;
	}

 out:
	return err;
}

/**
 * cellfs_readpage() - implement readpage() method of cellfs_aops {}
 * address_space_operations.
 * @file - file struct of the file to be read
 * @page - the page to be read
 */
static int cellfs_readpage(struct file *file, struct page *page)
{
	return mpage_readpage(page, cellfs_get_block);
}

/* static void cellfs_readahead(struct readahead_control *rac)
{
	mpage_readahead(rac, cellfs_get_block);
}
*/

static int cellfs_readpages(struct file *file, struct address_space *mapping,
		struct list_head *pages, unsigned int nr_pages)
{
	return mpage_readpages(mapping, pages, nr_pages, cellfs_get_block);
}

static int cellfs_writepages(struct address_space *mapping,
			    struct writeback_control *wbc)
{
	struct inode *inode = mapping->host;
	int err = 0;

	if (sb_rdonly(inode->i_sb)) {
		cellfs_clear_dirty_pages(mapping, false);
		return -EROFS;
	}

	if (wbc->sync_mode == WB_SYNC_ALL)
		err = cellfs_construct_dsync_segment(inode->i_sb, inode,
						    wbc->range_start,
						    wbc->range_end);
	return err;
}

static int cellfs_writepage(struct page *page, struct writeback_control *wbc)
{
	struct inode *inode = page->mapping->host;
	int err;

	if (sb_rdonly(inode->i_sb)) {
		/*
		 * It means that filesystem was remounted in read-only
		 * mode because of error or metadata corruption. But we
		 * have dirty pages that try to be flushed in background.
		 * So, here we simply discard this dirty page.
		 */
		cellfs_clear_dirty_page(page, false);
		unlock_page(page);
		return -EROFS;
	}

	redirty_page_for_writepage(wbc, page);
	unlock_page(page);

	if (wbc->sync_mode == WB_SYNC_ALL) {
		err = cellfs_construct_segment(inode->i_sb);
		if (unlikely(err))
			return err;
	} else if (wbc->for_reclaim)
		cellfs_flush_segment(inode->i_sb, inode->i_ino);

	return 0;
}

static int cellfs_set_page_dirty(struct page *page)
{
	struct inode *inode = page->mapping->host;
	int ret = __set_page_dirty_nobuffers(page);

	if (page_has_buffers(page)) {
		unsigned int nr_dirty = 0;
		struct buffer_head *bh, *head;

		/*
		 * This page is locked by callers, and no other thread
		 * concurrently marks its buffers dirty since they are
		 * only dirtied through routines in fs/buffer.c in
		 * which call sites of mark_buffer_dirty are protected
		 * by page lock.
		 */
		bh = head = page_buffers(page);
		do {
			/* Do not mark hole blocks dirty */
			if (buffer_dirty(bh) || !buffer_mapped(bh))
				continue;

			set_buffer_dirty(bh);
			nr_dirty++;
		} while (bh = bh->b_this_page, bh != head);

		if (nr_dirty)
			cellfs_set_file_dirty(inode, nr_dirty);
	} else if (ret) {
		unsigned int nr_dirty = 1 << (PAGE_SHIFT - inode->i_blkbits);

		cellfs_set_file_dirty(inode, nr_dirty);
	}
	return ret;
}

void cellfs_write_failed(struct address_space *mapping, loff_t to)
{
	struct inode *inode = mapping->host;

	if (to > inode->i_size) {
		truncate_pagecache(inode, inode->i_size);
		cellfs_truncate(inode);
	}
}

static int cellfs_write_begin(struct file *file, struct address_space *mapping,
			     loff_t pos, unsigned len, unsigned flags,
			     struct page **pagep, void **fsdata)

{
	struct inode *inode = mapping->host;
	int err = cellfs_transaction_begin(inode->i_sb, NULL, 1);

	if (unlikely(err))
		return err;

	err = block_write_begin(mapping, pos, len, flags, pagep,
				cellfs_get_block);
	if (unlikely(err)) {
		cellfs_write_failed(mapping, pos + len);
		cellfs_transaction_abort(inode->i_sb);
	}
	return err;
}

static int cellfs_write_end(struct file *file, struct address_space *mapping,
			   loff_t pos, unsigned len, unsigned copied,
			   struct page *page, void *fsdata)
{
	struct inode *inode = mapping->host;
	unsigned int start = pos & (PAGE_SIZE - 1);
	unsigned int nr_dirty;
	int err;

	nr_dirty = cellfs_page_count_clean_buffers(page, start,
						  start + copied);
	copied = generic_write_end(file, mapping, pos, len, copied, page,
				   fsdata);
	cellfs_set_file_dirty(inode, nr_dirty);
	err = cellfs_transaction_commit(inode->i_sb);
	return err ? : copied;
}

static ssize_t
cellfs_direct_IO(struct kiocb *iocb, struct iov_iter *iter)
{
	struct inode *inode = file_inode(iocb->ki_filp);

	if (iov_iter_rw(iter) == WRITE)
		return 0;

	/* Needs synchronization with the cleaner */
	return blockdev_direct_IO(iocb, inode, iter, cellfs_get_block);
}

const struct address_space_operations cellfs_aops = {
	.writepage		= cellfs_writepage,
	.readpage		= cellfs_readpage,
	.writepages		= cellfs_writepages,
	.set_page_dirty		= cellfs_set_page_dirty,
	/* .readahead		= cellfs_readahead, */
	.readpages		= cellfs_readpages,
	.write_begin		= cellfs_write_begin,
	.write_end		= cellfs_write_end,
	/* .releasepage		= cellfs_releasepage, */
	.invalidatepage		= block_invalidatepage,
	.direct_IO		= cellfs_direct_IO,
	.is_partially_uptodate  = block_is_partially_uptodate,
};

static int cellfs_insert_inode_locked(struct inode *inode,
				     struct cellfs_root *root,
				     unsigned long ino)
{
	struct cellfs_iget_args args = {
		.ino = ino, .root = root, .cno = 0, .for_gc = 0
	};

	return insert_inode_locked4(inode, ino, cellfs_iget_test, &args);
}

struct inode *cellfs_new_inode(struct inode *dir, umode_t mode)
{
	struct super_block *sb = dir->i_sb;
	struct the_cellfs *cellfs = sb->s_fs_info;
	struct inode *inode;
	struct cellfs_inode_info *ii;
	struct cellfs_root *root;
	int err = -ENOMEM;
	ino_t ino;

	inode = new_inode(sb);
	if (unlikely(!inode))
		goto failed;

	mapping_set_gfp_mask(inode->i_mapping,
			   mapping_gfp_constraint(inode->i_mapping, ~__GFP_FS));

	root = CELLFS_I(dir)->i_root;
	ii = CELLFS_I(inode);
	ii->i_state = BIT(CELLFS_I_NEW);
	ii->i_root = root;

	err = cellfs_ifile_create_inode(root->ifile, &ino, &ii->i_bh);
	if (unlikely(err))
		goto failed_ifile_create_inode;
	/* reference count of i_bh inherits from cellfs_mdt_read_block() */

	atomic64_inc(&root->inodes_count);
	inode_init_owner(inode, dir, mode);
	inode->i_ino = ino;
	inode->i_mtime = inode->i_atime = inode->i_ctime = current_time(inode);

	if (S_ISREG(mode) || S_ISDIR(mode) || S_ISLNK(mode)) {
		err = cellfs_bmap_read(ii->i_bmap, NULL);
		if (err < 0)
			goto failed_after_creation;

		set_bit(CELLFS_I_BMAP, &ii->i_state);
		/* No lock is needed; iget() ensures it. */
	}

	ii->i_flags = cellfs_mask_flags(
		mode, CELLFS_I(dir)->i_flags & CELLFS_FL_INHERITED);

	/* ii->i_file_acl = 0; */
	/* ii->i_dir_acl = 0; */
	ii->i_dir_start_lookup = 0;
	cellfs_set_inode_flags(inode);
	spin_lock(&cellfs->ns_next_gen_lock);
	inode->i_generation = cellfs->ns_next_generation++;
	spin_unlock(&cellfs->ns_next_gen_lock);
	if (cellfs_insert_inode_locked(inode, root, ino) < 0) {
		err = -EIO;
		goto failed_after_creation;
	}

	err = cellfs_init_acl(inode, dir);
	if (unlikely(err))
		/*
		 * Never occur.  When supporting cellfs_init_acl(),
		 * proper cancellation of above jobs should be considered.
		 */
		goto failed_after_creation;

	return inode;

 failed_after_creation:
	clear_nlink(inode);
	unlock_new_inode(inode);
	iput(inode);  /*
		       * raw_inode will be deleted through
		       * cellfs_evict_inode().
		       */
	goto failed;

 failed_ifile_create_inode:
	make_bad_inode(inode);
	iput(inode);
 failed:
	return ERR_PTR(err);
}

void cellfs_set_inode_flags(struct inode *inode)
{
	unsigned int flags = CELLFS_I(inode)->i_flags;
	unsigned int new_fl = 0;

	if (flags & FS_SYNC_FL)
		new_fl |= S_SYNC;
	if (flags & FS_APPEND_FL)
		new_fl |= S_APPEND;
	if (flags & FS_IMMUTABLE_FL)
		new_fl |= S_IMMUTABLE;
	if (flags & FS_NOATIME_FL)
		new_fl |= S_NOATIME;
	if (flags & FS_DIRSYNC_FL)
		new_fl |= S_DIRSYNC;
	inode_set_flags(inode, new_fl, S_SYNC | S_APPEND | S_IMMUTABLE |
			S_NOATIME | S_DIRSYNC);
}

int cellfs_read_inode_common(struct inode *inode,
			    struct cellfs_inode *raw_inode)
{
	struct cellfs_inode_info *ii = CELLFS_I(inode);
	int err;

	inode->i_mode = le16_to_cpu(raw_inode->i_mode);
	i_uid_write(inode, le32_to_cpu(raw_inode->i_uid));
	i_gid_write(inode, le32_to_cpu(raw_inode->i_gid));
	set_nlink(inode, le16_to_cpu(raw_inode->i_links_count));
	inode->i_size = le64_to_cpu(raw_inode->i_size);
	inode->i_atime.tv_sec = le64_to_cpu(raw_inode->i_mtime);
	inode->i_ctime.tv_sec = le64_to_cpu(raw_inode->i_ctime);
	inode->i_mtime.tv_sec = le64_to_cpu(raw_inode->i_mtime);
	inode->i_atime.tv_nsec = le32_to_cpu(raw_inode->i_mtime_nsec);
	inode->i_ctime.tv_nsec = le32_to_cpu(raw_inode->i_ctime_nsec);
	inode->i_mtime.tv_nsec = le32_to_cpu(raw_inode->i_mtime_nsec);
	if (inode->i_nlink == 0)
		return -ESTALE; /* this inode is deleted */

	inode->i_blocks = le64_to_cpu(raw_inode->i_blocks);
	ii->i_flags = le32_to_cpu(raw_inode->i_flags);
#if 0
	ii->i_file_acl = le32_to_cpu(raw_inode->i_file_acl);
	ii->i_dir_acl = S_ISREG(inode->i_mode) ?
		0 : le32_to_cpu(raw_inode->i_dir_acl);
#endif
	ii->i_dir_start_lookup = 0;
	inode->i_generation = le32_to_cpu(raw_inode->i_generation);

	if (S_ISREG(inode->i_mode) || S_ISDIR(inode->i_mode) ||
	    S_ISLNK(inode->i_mode)) {
		err = cellfs_bmap_read(ii->i_bmap, raw_inode);
		if (err < 0)
			return err;
		set_bit(CELLFS_I_BMAP, &ii->i_state);
		/* No lock is needed; iget() ensures it. */
	}
	return 0;
}

static int __cellfs_read_inode(struct super_block *sb,
			      struct cellfs_root *root, unsigned long ino,
			      struct inode *inode)
{
	struct the_cellfs *cellfs = sb->s_fs_info;
	struct buffer_head *bh;
	struct cellfs_inode *raw_inode;
	int err;

	down_read(&CELLFS_MDT(cellfs->ns_dat)->mi_sem);
	err = cellfs_ifile_get_inode_block(root->ifile, ino, &bh);
	if (unlikely(err))
		goto bad_inode;

	raw_inode = cellfs_ifile_map_inode(root->ifile, ino, bh);

	err = cellfs_read_inode_common(inode, raw_inode);
	if (err)
		goto failed_unmap;

	if (S_ISREG(inode->i_mode)) {
		inode->i_op = &cellfs_file_inode_operations;
		inode->i_fop = &cellfs_file_operations;
		inode->i_mapping->a_ops = &cellfs_aops;
	} else if (S_ISDIR(inode->i_mode)) {
		inode->i_op = &cellfs_dir_inode_operations;
		inode->i_fop = &cellfs_dir_operations;
		inode->i_mapping->a_ops = &cellfs_aops;
	} else if (S_ISLNK(inode->i_mode)) {
		inode->i_op = &cellfs_symlink_inode_operations;
		inode_nohighmem(inode);
		inode->i_mapping->a_ops = &cellfs_aops;
	} else {
		inode->i_op = &cellfs_special_inode_operations;
		init_special_inode(
			inode, inode->i_mode,
			huge_decode_dev(le64_to_cpu(raw_inode->i_device_code)));
	}
	cellfs_ifile_unmap_inode(root->ifile, ino, bh);
	brelse(bh);
	up_read(&CELLFS_MDT(cellfs->ns_dat)->mi_sem);
	cellfs_set_inode_flags(inode);
	mapping_set_gfp_mask(inode->i_mapping,
			   mapping_gfp_constraint(inode->i_mapping, ~__GFP_FS));
	return 0;

 failed_unmap:
	cellfs_ifile_unmap_inode(root->ifile, ino, bh);
	brelse(bh);

 bad_inode:
	up_read(&CELLFS_MDT(cellfs->ns_dat)->mi_sem);
	return err;
}

static int cellfs_iget_test(struct inode *inode, void *opaque)
{
	struct cellfs_iget_args *args = opaque;
	struct cellfs_inode_info *ii;

	if (args->ino != inode->i_ino || args->root != CELLFS_I(inode)->i_root)
		return 0;

	ii = CELLFS_I(inode);
	if (!test_bit(CELLFS_I_GCINODE, &ii->i_state))
		return !args->for_gc;

	return args->for_gc && args->cno == ii->i_cno;
}

static int cellfs_iget_set(struct inode *inode, void *opaque)
{
	struct cellfs_iget_args *args = opaque;

	inode->i_ino = args->ino;
	if (args->for_gc) {
		CELLFS_I(inode)->i_state = BIT(CELLFS_I_GCINODE);
		CELLFS_I(inode)->i_cno = args->cno;
		CELLFS_I(inode)->i_root = NULL;
	} else {
		if (args->root && args->ino == CELLFS_ROOT_INO)
			cellfs_get_root(args->root);
		CELLFS_I(inode)->i_root = args->root;
	}
	return 0;
}

struct inode *cellfs_ilookup(struct super_block *sb, struct cellfs_root *root,
			    unsigned long ino)
{
	struct cellfs_iget_args args = {
		.ino = ino, .root = root, .cno = 0, .for_gc = 0
	};

	return ilookup5(sb, ino, cellfs_iget_test, &args);
}

struct inode *cellfs_iget_locked(struct super_block *sb, struct cellfs_root *root,
				unsigned long ino)
{
	struct cellfs_iget_args args = {
		.ino = ino, .root = root, .cno = 0, .for_gc = 0
	};

	return iget5_locked(sb, ino, cellfs_iget_test, cellfs_iget_set, &args);
}

struct inode *cellfs_iget(struct super_block *sb, struct cellfs_root *root,
			 unsigned long ino)
{
	struct inode *inode;
	int err;

	inode = cellfs_iget_locked(sb, root, ino);
	if (unlikely(!inode))
		return ERR_PTR(-ENOMEM);
	if (!(inode->i_state & I_NEW))
		return inode;

	err = __cellfs_read_inode(sb, root, ino, inode);
	if (unlikely(err)) {
		iget_failed(inode);
		return ERR_PTR(err);
	}
	unlock_new_inode(inode);
	return inode;
}

struct inode *cellfs_iget_for_gc(struct super_block *sb, unsigned long ino,
				__u64 cno)
{
	struct cellfs_iget_args args = {
		.ino = ino, .root = NULL, .cno = cno, .for_gc = 1
	};
	struct inode *inode;
	int err;

	inode = iget5_locked(sb, ino, cellfs_iget_test, cellfs_iget_set, &args);
	if (unlikely(!inode))
		return ERR_PTR(-ENOMEM);
	if (!(inode->i_state & I_NEW))
		return inode;

	err = cellfs_init_gcinode(inode);
	if (unlikely(err)) {
		iget_failed(inode);
		return ERR_PTR(err);
	}
	unlock_new_inode(inode);
	return inode;
}

void cellfs_write_inode_common(struct inode *inode,
			      struct cellfs_inode *raw_inode, int has_bmap)
{
	struct cellfs_inode_info *ii = CELLFS_I(inode);

	raw_inode->i_mode = cpu_to_le16(inode->i_mode);
	raw_inode->i_uid = cpu_to_le32(i_uid_read(inode));
	raw_inode->i_gid = cpu_to_le32(i_gid_read(inode));
	raw_inode->i_links_count = cpu_to_le16(inode->i_nlink);
	raw_inode->i_size = cpu_to_le64(inode->i_size);
	raw_inode->i_ctime = cpu_to_le64(inode->i_ctime.tv_sec);
	raw_inode->i_mtime = cpu_to_le64(inode->i_mtime.tv_sec);
	raw_inode->i_ctime_nsec = cpu_to_le32(inode->i_ctime.tv_nsec);
	raw_inode->i_mtime_nsec = cpu_to_le32(inode->i_mtime.tv_nsec);
	raw_inode->i_blocks = cpu_to_le64(inode->i_blocks);

	raw_inode->i_flags = cpu_to_le32(ii->i_flags);
	raw_inode->i_generation = cpu_to_le32(inode->i_generation);

	if (CELLFS_ROOT_METADATA_FILE(inode->i_ino)) {
		struct the_cellfs *cellfs = inode->i_sb->s_fs_info;

		/* zero-fill unused portion in the case of super root block */
		raw_inode->i_xattr = 0;
		raw_inode->i_pad = 0;
		memset((void *)raw_inode + sizeof(*raw_inode), 0,
		       cellfs->ns_inode_size - sizeof(*raw_inode));
	}

	if (has_bmap)
		cellfs_bmap_write(ii->i_bmap, raw_inode);
	else if (S_ISCHR(inode->i_mode) || S_ISBLK(inode->i_mode))
		raw_inode->i_device_code =
			cpu_to_le64(huge_encode_dev(inode->i_rdev));
	/*
	 * When extending inode, cellfs->ns_inode_size should be checked
	 * for substitutions of appended fields.
	 */
}

void cellfs_update_inode(struct inode *inode, struct buffer_head *ibh, int flags)
{
	ino_t ino = inode->i_ino;
	struct cellfs_inode_info *ii = CELLFS_I(inode);
	struct inode *ifile = ii->i_root->ifile;
	struct cellfs_inode *raw_inode;

	raw_inode = cellfs_ifile_map_inode(ifile, ino, ibh);

	if (test_and_clear_bit(CELLFS_I_NEW, &ii->i_state))
		memset(raw_inode, 0, CELLFS_MDT(ifile)->mi_entry_size);
	if (flags & I_DIRTY_DATASYNC)
		set_bit(CELLFS_I_INODE_SYNC, &ii->i_state);

	cellfs_write_inode_common(inode, raw_inode, 0);
		/*
		 * XXX: call with has_bmap = 0 is a workaround to avoid
		 * deadlock of bmap.  This delays update of i_bmap to just
		 * before writing.
		 */

	cellfs_ifile_unmap_inode(ifile, ino, ibh);
}

#define CELLFS_MAX_TRUNCATE_BLOCKS	16384  /* 64MB for 4KB block */

static void cellfs_truncate_bmap(struct cellfs_inode_info *ii,
				unsigned long from)
{
	__u64 b;
	int ret;

	if (!test_bit(CELLFS_I_BMAP, &ii->i_state))
		return;
repeat:
	ret = cellfs_bmap_last_key(ii->i_bmap, &b);
	if (ret == -ENOENT)
		return;
	else if (ret < 0)
		goto failed;

	if (b < from)
		return;

	b -= min_t(__u64, CELLFS_MAX_TRUNCATE_BLOCKS, b - from);
	ret = cellfs_bmap_truncate(ii->i_bmap, b);
	cellfs_relax_pressure_in_lock(ii->vfs_inode.i_sb);
	if (!ret || (ret == -ENOMEM &&
		     cellfs_bmap_truncate(ii->i_bmap, b) == 0))
		goto repeat;

failed:
	cellfs_msg(ii->vfs_inode.i_sb, KERN_WARNING,
		  "error %d truncating bmap (ino=%lu)", ret,
		  ii->vfs_inode.i_ino);
}

void cellfs_truncate(struct inode *inode)
{
	unsigned long blkoff;
	unsigned int blocksize;
	struct cellfs_transaction_info ti;
	struct super_block *sb = inode->i_sb;
	struct cellfs_inode_info *ii = CELLFS_I(inode);

	if (!test_bit(CELLFS_I_BMAP, &ii->i_state))
		return;
	if (IS_APPEND(inode) || IS_IMMUTABLE(inode))
		return;

	blocksize = sb->s_blocksize;
	blkoff = (inode->i_size + blocksize - 1) >> sb->s_blocksize_bits;
	cellfs_transaction_begin(sb, &ti, 0); /* never fails */

	block_truncate_page(inode->i_mapping, inode->i_size, cellfs_get_block);

	cellfs_truncate_bmap(ii, blkoff);

	inode->i_mtime = inode->i_ctime = current_time(inode);
	if (IS_SYNC(inode))
		cellfs_set_transaction_flag(CELLFS_TI_SYNC);

	cellfs_mark_inode_dirty(inode);
	cellfs_set_file_dirty(inode, 0);
	cellfs_transaction_commit(sb);
	/*
	 * May construct a logical segment and may fail in sync mode.
	 * But truncate has no return value.
	 */
}

static void cellfs_clear_inode(struct inode *inode)
{
	struct cellfs_inode_info *ii = CELLFS_I(inode);

	/*
	 * Free resources allocated in cellfs_read_inode(), here.
	 */
	BUG_ON(!list_empty(&ii->i_dirty));
	brelse(ii->i_bh);
	ii->i_bh = NULL;

	if (cellfs_is_metadata_file_inode(inode))
		cellfs_mdt_clear(inode);

	if (test_bit(CELLFS_I_BMAP, &ii->i_state))
		cellfs_bmap_clear(ii->i_bmap);

	cellfs_btnode_cache_clear(&ii->i_btnode_cache);

	if (ii->i_root && inode->i_ino == CELLFS_ROOT_INO)
		cellfs_put_root(ii->i_root);
}

void cellfs_evict_inode(struct inode *inode)
{
	struct cellfs_transaction_info ti;
	struct super_block *sb = inode->i_sb;
	struct cellfs_inode_info *ii = CELLFS_I(inode);
	int ret;

	if (inode->i_nlink || !ii->i_root || unlikely(is_bad_inode(inode))) {
		truncate_inode_pages_final(&inode->i_data);
		clear_inode(inode);
		cellfs_clear_inode(inode);
		return;
	}
	cellfs_transaction_begin(sb, &ti, 0); /* never fails */

	truncate_inode_pages_final(&inode->i_data);

	/* TODO: some of the following operations may fail.  */
	cellfs_truncate_bmap(ii, 0);
	cellfs_mark_inode_dirty(inode);
	clear_inode(inode);

	ret = cellfs_ifile_delete_inode(ii->i_root->ifile, inode->i_ino);
	if (!ret)
		atomic64_dec(&ii->i_root->inodes_count);

	cellfs_clear_inode(inode);

	if (IS_SYNC(inode))
		cellfs_set_transaction_flag(CELLFS_TI_SYNC);
	cellfs_transaction_commit(sb);
	/*
	 * May construct a logical segment and may fail in sync mode.
	 * But delete_inode has no return value.
	 */
}

int cellfs_setattr(struct dentry *dentry, struct iattr *iattr)
{
	struct cellfs_transaction_info ti;
	struct inode *inode = d_inode(dentry);
	struct super_block *sb = inode->i_sb;
	int err;

	err = setattr_prepare(dentry, iattr);
	if (err)
		return err;

	err = cellfs_transaction_begin(sb, &ti, 0);
	if (unlikely(err))
		return err;

	if ((iattr->ia_valid & ATTR_SIZE) &&
	    iattr->ia_size != i_size_read(inode)) {
		inode_dio_wait(inode);
		truncate_setsize(inode, iattr->ia_size);
		cellfs_truncate(inode);
	}

	setattr_copy(inode, iattr);
	mark_inode_dirty(inode);

	if (iattr->ia_valid & ATTR_MODE) {
		err = cellfs_acl_chmod(inode);
		if (unlikely(err))
			goto out_err;
	}

	return cellfs_transaction_commit(sb);

out_err:
	cellfs_transaction_abort(sb);
	return err;
}

int cellfs_permission(struct inode *inode, int mask)
{
	struct cellfs_root *root = CELLFS_I(inode)->i_root;

	if ((mask & MAY_WRITE) && root &&
	    root->cno != CELLFS_CPTREE_CURRENT_CNO)
		return -EROFS; /* snapshot is not writable */

	return generic_permission(inode, mask);
}

int cellfs_load_inode_block(struct inode *inode, struct buffer_head **pbh)
{
	struct the_cellfs *cellfs = inode->i_sb->s_fs_info;
	struct cellfs_inode_info *ii = CELLFS_I(inode);
	int err;

	spin_lock(&cellfs->ns_inode_lock);
	if (ii->i_bh == NULL) {
		spin_unlock(&cellfs->ns_inode_lock);
		err = cellfs_ifile_get_inode_block(ii->i_root->ifile,
						  inode->i_ino, pbh);
		if (unlikely(err))
			return err;
		spin_lock(&cellfs->ns_inode_lock);
		if (ii->i_bh == NULL)
			ii->i_bh = *pbh;
		else {
			brelse(*pbh);
			*pbh = ii->i_bh;
		}
	} else
		*pbh = ii->i_bh;

	get_bh(*pbh);
	spin_unlock(&cellfs->ns_inode_lock);
	return 0;
}

int cellfs_inode_dirty(struct inode *inode)
{
	struct cellfs_inode_info *ii = CELLFS_I(inode);
	struct the_cellfs *cellfs = inode->i_sb->s_fs_info;
	int ret = 0;

	if (!list_empty(&ii->i_dirty)) {
		spin_lock(&cellfs->ns_inode_lock);
		ret = test_bit(CELLFS_I_DIRTY, &ii->i_state) ||
			test_bit(CELLFS_I_BUSY, &ii->i_state);
		spin_unlock(&cellfs->ns_inode_lock);
	}
	return ret;
}

int cellfs_set_file_dirty(struct inode *inode, unsigned int nr_dirty)
{
	struct cellfs_inode_info *ii = CELLFS_I(inode);
	struct the_cellfs *cellfs = inode->i_sb->s_fs_info;

	atomic_add(nr_dirty, &cellfs->ns_ndirtyblks);

	if (test_and_set_bit(CELLFS_I_DIRTY, &ii->i_state))
		return 0;

	spin_lock(&cellfs->ns_inode_lock);
	if (!test_bit(CELLFS_I_QUEUED, &ii->i_state) &&
	    !test_bit(CELLFS_I_BUSY, &ii->i_state)) {
		/*
		 * Because this routine may race with cellfs_dispose_list(),
		 * we have to check CELLFS_I_QUEUED here, too.
		 */
		if (list_empty(&ii->i_dirty) && igrab(inode) == NULL) {
			/*
			 * This will happen when somebody is freeing
			 * this inode.
			 */
			cellfs_msg(inode->i_sb, KERN_WARNING,
				  "cannot set file dirty (ino=%lu): the file is being freed",
				  inode->i_ino);
			spin_unlock(&cellfs->ns_inode_lock);
			return -EINVAL; /*
					 * CELLFS_I_DIRTY may remain for
					 * freeing inode.
					 */
		}
		list_move_tail(&ii->i_dirty, &cellfs->ns_dirty_files);
		set_bit(CELLFS_I_QUEUED, &ii->i_state);
	}
	spin_unlock(&cellfs->ns_inode_lock);
	return 0;
}

int __cellfs_mark_inode_dirty(struct inode *inode, int flags)
{
	struct buffer_head *ibh;
	int err;

	err = cellfs_load_inode_block(inode, &ibh);
	if (unlikely(err)) {
		cellfs_msg(inode->i_sb, KERN_WARNING,
			  "cannot mark inode dirty (ino=%lu): error %d loading inode block",
			  inode->i_ino, err);
		return err;
	}
	cellfs_update_inode(inode, ibh, flags);
	mark_buffer_dirty(ibh);
	cellfs_mdt_mark_dirty(CELLFS_I(inode)->i_root->ifile);
	brelse(ibh);
	return 0;
}

/**
 * cellfs_dirty_inode - reflect changes on given inode to an inode block.
 * @inode: inode of the file to be registered.
 *
 * cellfs_dirty_inode() loads a inode block containing the specified
 * @inode and copies data from a cellfs_inode to a corresponding inode
 * entry in the inode block. This operation is excluded from the segment
 * construction. This function can be called both as a single operation
 * and as a part of indivisible file operations.
 */
void cellfs_dirty_inode(struct inode *inode, int flags)
{
	struct cellfs_transaction_info ti;
	struct cellfs_mdt_info *mdi = CELLFS_MDT(inode);

	if (is_bad_inode(inode)) {
		cellfs_msg(inode->i_sb, KERN_WARNING,
			  "tried to mark bad_inode dirty. ignored.");
		dump_stack();
		return;
	}
	if (mdi) {
		cellfs_mdt_mark_dirty(inode);
		return;
	}
	cellfs_transaction_begin(inode->i_sb, &ti, 0);
	__cellfs_mark_inode_dirty(inode, flags);
	cellfs_transaction_commit(inode->i_sb); /* never fails */
}

int cellfs_fiemap(struct inode *inode, struct fiemap_extent_info *fieinfo,
		 __u64 start, __u64 len)
{
	struct the_cellfs *cellfs = inode->i_sb->s_fs_info;
	__u64 logical = 0, phys = 0, size = 0;
	__u32 flags = 0;
	loff_t isize;
	sector_t blkoff, end_blkoff;
	sector_t delalloc_blkoff;
	unsigned long delalloc_blklen;
	unsigned int blkbits = inode->i_blkbits;
	int ret, n;

	// ret = fiemap_prep(inode, fieinfo, start, &len, 0);
	ret = fiemap_check_flags(fieinfo, FIEMAP_FLAG_SYNC);
	if (ret)
		return ret;

	inode_lock(inode);

	isize = i_size_read(inode);

	blkoff = start >> blkbits;
	end_blkoff = (start + len - 1) >> blkbits;

	delalloc_blklen = cellfs_find_uncommitted_extent(inode, blkoff,
							&delalloc_blkoff);

	do {
		__u64 blkphy;
		unsigned int maxblocks;

		if (delalloc_blklen && blkoff == delalloc_blkoff) {
			if (size) {
				/* End of the current extent */
				ret = fiemap_fill_next_extent(
					fieinfo, logical, phys, size, flags);
				if (ret)
					break;
			}
			if (blkoff > end_blkoff)
				break;

			flags = FIEMAP_EXTENT_MERGED | FIEMAP_EXTENT_DELALLOC;
			logical = blkoff << blkbits;
			phys = 0;
			size = delalloc_blklen << blkbits;

			blkoff = delalloc_blkoff + delalloc_blklen;
			delalloc_blklen = cellfs_find_uncommitted_extent(
				inode, blkoff, &delalloc_blkoff);
			continue;
		}

		/*
		 * Limit the number of blocks that we look up so as
		 * not to get into the next delayed allocation extent.
		 */
		maxblocks = INT_MAX;
		if (delalloc_blklen)
			maxblocks = min_t(sector_t, delalloc_blkoff - blkoff,
					  maxblocks);
		blkphy = 0;

		down_read(&CELLFS_MDT(cellfs->ns_dat)->mi_sem);
		n = cellfs_bmap_lookup_contig(
			CELLFS_I(inode)->i_bmap, blkoff, &blkphy, maxblocks);
		up_read(&CELLFS_MDT(cellfs->ns_dat)->mi_sem);

		if (n < 0) {
			int past_eof;

			if (unlikely(n != -ENOENT))
				break; /* error */

			/* HOLE */
			blkoff++;
			past_eof = ((blkoff << blkbits) >= isize);

			if (size) {
				/* End of the current extent */

				if (past_eof)
					flags |= FIEMAP_EXTENT_LAST;

				ret = fiemap_fill_next_extent(
					fieinfo, logical, phys, size, flags);
				if (ret)
					break;
				size = 0;
			}
			if (blkoff > end_blkoff || past_eof)
				break;
		} else {
			if (size) {
				if (phys && blkphy << blkbits == phys + size) {
					/* The current extent goes on */
					size += n << blkbits;
				} else {
					/* Terminate the current extent */
					ret = fiemap_fill_next_extent(
						fieinfo, logical, phys, size,
						flags);
					if (ret || blkoff > end_blkoff)
						break;

					/* Start another extent */
					flags = FIEMAP_EXTENT_MERGED;
					logical = blkoff << blkbits;
					phys = blkphy << blkbits;
					size = n << blkbits;
				}
			} else {
				/* Start a new extent */
				flags = FIEMAP_EXTENT_MERGED;
				logical = blkoff << blkbits;
				phys = blkphy << blkbits;
				size = n << blkbits;
			}
			blkoff += n;
		}
		cond_resched();
	} while (true);

	/* If ret is 1 then we just hit the end of the extent array */
	if (ret == 1)
		ret = 0;

	inode_unlock(inode);
	return ret;
}
