/* SPDX-License-Identifier: GPL-2.0+ */
/*
 * CELLFS block mapping.
 *
 * Copyright (C) 2006-2008 Nippon Telegraph and Telephone Corporation.
 *
 * Written by Koji Sato.
 */

#ifndef _CELLFS_BMAP_H
#define _CELLFS_BMAP_H

#include <linux/types.h>
#include <linux/fs.h>
#include <linux/buffer_head.h>
#include <linux/cellfs_ondisk.h>	/* cellfs_binfo, cellfs_inode, etc */
#include "alloc.h"
#include "dat.h"

#define CELLFS_BMAP_INVALID_PTR	0

#define cellfs_bmap_keydiff_abs(diff)	((diff) < 0 ? -(diff) : (diff))


struct cellfs_bmap;

/**
 * union cellfs_bmap_ptr_req - request for bmap ptr
 * @bpr_ptr: bmap pointer
 * @bpr_req: request for persistent allocator
 */
union cellfs_bmap_ptr_req {
	__u64 bpr_ptr;
	struct cellfs_palloc_req bpr_req;
};

/**
 * struct cellfs_bmap_stats - bmap statistics
 * @bs_nblocks: number of blocks created or deleted
 */
struct cellfs_bmap_stats {
	unsigned int bs_nblocks;
};

/**
 * struct cellfs_bmap_operations - bmap operation table
 */
struct cellfs_bmap_operations {
	int (*bop_lookup)(const struct cellfs_bmap *, __u64, int, __u64 *);
	int (*bop_lookup_contig)(const struct cellfs_bmap *, __u64, __u64 *,
				 unsigned int);
	int (*bop_insert)(struct cellfs_bmap *, __u64, __u64);
	int (*bop_delete)(struct cellfs_bmap *, __u64);
	void (*bop_clear)(struct cellfs_bmap *);

	int (*bop_propagate)(struct cellfs_bmap *, struct buffer_head *);
	void (*bop_lookup_dirty_buffers)(struct cellfs_bmap *,
					 struct list_head *);

	int (*bop_assign)(struct cellfs_bmap *,
			  struct buffer_head **,
			  sector_t,
			  union cellfs_binfo *);
	int (*bop_mark)(struct cellfs_bmap *, __u64, int);

	int (*bop_seek_key)(const struct cellfs_bmap *, __u64, __u64 *);
	int (*bop_last_key)(const struct cellfs_bmap *, __u64 *);

	/* The following functions are internal use only. */
	int (*bop_check_insert)(const struct cellfs_bmap *, __u64);
	int (*bop_check_delete)(struct cellfs_bmap *, __u64);
	int (*bop_gather_data)(struct cellfs_bmap *, __u64 *, __u64 *, int);
};


#define CELLFS_BMAP_SIZE		(CELLFS_INODE_BMAP_SIZE * sizeof(__le64))
#define CELLFS_BMAP_KEY_BIT	(sizeof(unsigned long) * 8 /* CHAR_BIT */)
#define CELLFS_BMAP_NEW_PTR_INIT	\
	(1UL << (sizeof(unsigned long) * 8 /* CHAR_BIT */ - 1))

static inline int cellfs_bmap_is_new_ptr(unsigned long ptr)
{
	return !!(ptr & CELLFS_BMAP_NEW_PTR_INIT);
}


/**
 * struct cellfs_bmap - bmap structure
 * @b_u: raw data
 * @b_sem: semaphore
 * @b_inode: owner of bmap
 * @b_ops: bmap operation table
 * @b_last_allocated_key: last allocated key for data block
 * @b_last_allocated_ptr: last allocated ptr for data block
 * @b_ptr_type: pointer type
 * @b_state: state
 * @b_nchildren_per_block: maximum number of child nodes for non-root nodes
 */
struct cellfs_bmap {
	union {
		__u8 u_flags;
		__le64 u_data[CELLFS_BMAP_SIZE / sizeof(__le64)];
	} b_u;
	struct rw_semaphore b_sem;
	struct inode *b_inode;
	const struct cellfs_bmap_operations *b_ops;
	__u64 b_last_allocated_key;
	__u64 b_last_allocated_ptr;
	int b_ptr_type;
	int b_state;
	__u16 b_nchildren_per_block;
};

/* pointer type */
#define CELLFS_BMAP_PTR_P	0	/* physical block number (i.e. LBN) */
#define CELLFS_BMAP_PTR_VS	1	/*
					 * virtual block number (single
					 * version)
					 */
#define CELLFS_BMAP_PTR_VM	2	/*
					 * virtual block number (has multiple
					 * versions)
					 */
#define CELLFS_BMAP_PTR_U	(-1)	/* never perform pointer operations */

#define CELLFS_BMAP_USE_VBN(bmap)	((bmap)->b_ptr_type > 0)

/* state */
#define CELLFS_BMAP_DIRTY	0x00000001

/**
 * struct cellfs_bmap_store - shadow copy of bmap state
 * @data: cached raw block mapping of on-disk inode
 * @last_allocated_key: cached value of last allocated key for data block
 * @last_allocated_ptr: cached value of last allocated ptr for data block
 * @state: cached value of state field of bmap structure
 */
struct cellfs_bmap_store {
	__le64 data[CELLFS_BMAP_SIZE / sizeof(__le64)];
	__u64 last_allocated_key;
	__u64 last_allocated_ptr;
	int state;
};

int cellfs_bmap_test_and_clear_dirty(struct cellfs_bmap *);
int cellfs_bmap_read(struct cellfs_bmap *, struct cellfs_inode *);
void cellfs_bmap_write(struct cellfs_bmap *, struct cellfs_inode *);
int cellfs_bmap_lookup_contig(struct cellfs_bmap *, __u64, __u64 *, unsigned int);
int cellfs_bmap_insert(struct cellfs_bmap *bmap, __u64 key, unsigned long rec);
int cellfs_bmap_delete(struct cellfs_bmap *bmap, __u64 key);
int cellfs_bmap_seek_key(struct cellfs_bmap *bmap, __u64 start, __u64 *keyp);
int cellfs_bmap_last_key(struct cellfs_bmap *bmap, __u64 *keyp);
int cellfs_bmap_truncate(struct cellfs_bmap *bmap, __u64 key);
void cellfs_bmap_clear(struct cellfs_bmap *);
int cellfs_bmap_propagate(struct cellfs_bmap *, struct buffer_head *);
void cellfs_bmap_lookup_dirty_buffers(struct cellfs_bmap *, struct list_head *);
int cellfs_bmap_assign(struct cellfs_bmap *, struct buffer_head **,
		      unsigned long, union cellfs_binfo *);
int cellfs_bmap_lookup_at_level(struct cellfs_bmap *, __u64, int, __u64 *);
int cellfs_bmap_mark(struct cellfs_bmap *, __u64, int);

void cellfs_bmap_init_gc(struct cellfs_bmap *);

void cellfs_bmap_save(const struct cellfs_bmap *, struct cellfs_bmap_store *);
void cellfs_bmap_restore(struct cellfs_bmap *, const struct cellfs_bmap_store *);

static inline int cellfs_bmap_lookup(struct cellfs_bmap *bmap, __u64 key,
				    __u64 *ptr)
{
	return cellfs_bmap_lookup_at_level(bmap, key, 1, ptr);
}

/*
 * Internal use only
 */
struct inode *cellfs_bmap_get_dat(const struct cellfs_bmap *);

static inline int cellfs_bmap_prepare_alloc_ptr(struct cellfs_bmap *bmap,
					       union cellfs_bmap_ptr_req *req,
					       struct inode *dat)
{
	if (dat)
		return cellfs_dat_prepare_alloc(dat, &req->bpr_req);
	/* ignore target ptr */
	req->bpr_ptr = bmap->b_last_allocated_ptr++;
	return 0;
}

static inline void cellfs_bmap_commit_alloc_ptr(struct cellfs_bmap *bmap,
					       union cellfs_bmap_ptr_req *req,
					       struct inode *dat)
{
	if (dat)
		cellfs_dat_commit_alloc(dat, &req->bpr_req);
}

static inline void cellfs_bmap_abort_alloc_ptr(struct cellfs_bmap *bmap,
					      union cellfs_bmap_ptr_req *req,
					      struct inode *dat)
{
	if (dat)
		cellfs_dat_abort_alloc(dat, &req->bpr_req);
	else
		bmap->b_last_allocated_ptr--;
}

static inline int cellfs_bmap_prepare_end_ptr(struct cellfs_bmap *bmap,
					     union cellfs_bmap_ptr_req *req,
					     struct inode *dat)
{
	return dat ? cellfs_dat_prepare_end(dat, &req->bpr_req) : 0;
}

static inline void cellfs_bmap_commit_end_ptr(struct cellfs_bmap *bmap,
					     union cellfs_bmap_ptr_req *req,
					     struct inode *dat)
{
	if (dat)
		cellfs_dat_commit_end(dat, &req->bpr_req,
				     bmap->b_ptr_type == CELLFS_BMAP_PTR_VS);
}

static inline void cellfs_bmap_abort_end_ptr(struct cellfs_bmap *bmap,
					    union cellfs_bmap_ptr_req *req,
					    struct inode *dat)
{
	if (dat)
		cellfs_dat_abort_end(dat, &req->bpr_req);
}

static inline void cellfs_bmap_set_target_v(struct cellfs_bmap *bmap, __u64 key,
					   __u64 ptr)
{
	bmap->b_last_allocated_key = key;
	bmap->b_last_allocated_ptr = ptr;
}

__u64 cellfs_bmap_data_get_key(const struct cellfs_bmap *,
			      const struct buffer_head *);

__u64 cellfs_bmap_find_target_seq(const struct cellfs_bmap *, __u64);
__u64 cellfs_bmap_find_target_in_group(const struct cellfs_bmap *);


/* Assume that bmap semaphore is locked. */
static inline int cellfs_bmap_dirty(const struct cellfs_bmap *bmap)
{
	return !!(bmap->b_state & CELLFS_BMAP_DIRTY);
}

/* Assume that bmap semaphore is locked. */
static inline void cellfs_bmap_set_dirty(struct cellfs_bmap *bmap)
{
	bmap->b_state |= CELLFS_BMAP_DIRTY;
}

/* Assume that bmap semaphore is locked. */
static inline void cellfs_bmap_clear_dirty(struct cellfs_bmap *bmap)
{
	bmap->b_state &= ~CELLFS_BMAP_DIRTY;
}


#define CELLFS_BMAP_LARGE	0x1

#define CELLFS_BMAP_SMALL_LOW	CELLFS_DIRECT_KEY_MIN
#define CELLFS_BMAP_SMALL_HIGH	CELLFS_DIRECT_KEY_MAX
#define CELLFS_BMAP_LARGE_LOW	CELLFS_BTREE_ROOT_NCHILDREN_MAX
#define CELLFS_BMAP_LARGE_HIGH	CELLFS_BTREE_KEY_MAX

#endif	/* _CELLFS_BMAP_H */
