/*
 * namei.c - CELLFS pathname lookup operations.
 *
 * Copyright (C) 2005-2008 Nippon Telegraph and Telephone Corporation.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * Modified for CELLFS by Amagai Yoshiji and Ryusuke Konishi.
 */
/*
 *  linux/fs/ext2/namei.c
 *
 * Copyright (C) 1992, 1993, 1994, 1995
 * Remy Card (card@masi.ibp.fr)
 * Laboratoire MASI - Institut Blaise Pascal
 * Universite Pierre et Marie Curie (Paris VI)
 *
 *  from
 *
 *  linux/fs/minix/namei.c
 *
 *  Copyright (C) 1991, 1992  Linus Torvalds
 *
 *  Big-endian to little-endian byte-swapping/bitmaps by
 *        David S. Miller (davem@caip.rutgers.edu), 1995
 */

#include <linux/pagemap.h>
#include "cellfs.h"
#include "export.h"

#define CELLFS_FID_SIZE_NON_CONNECTABLE \
	(offsetof(struct cellfs_fid, parent_gen) / 4)
#define CELLFS_FID_SIZE_CONNECTABLE	(sizeof(struct cellfs_fid) / 4)

static inline int cellfs_add_nondir(struct dentry *dentry, struct inode *inode)
{
	int err = cellfs_add_link(dentry, inode);

	if (!err) {
		d_instantiate(dentry, inode);
		unlock_new_inode(inode);
		return 0;
	}
	inode_dec_link_count(inode);
	unlock_new_inode(inode);
	iput(inode);
	return err;
}

/*
 * Methods themselves.
 */

static struct dentry *
cellfs_lookup(struct inode *dir, struct dentry *dentry, unsigned int flags)
{
	struct inode *inode;
	ino_t ino;

	if (dentry->d_name.len > CELLFS_NAME_LEN)
		return ERR_PTR(-ENAMETOOLONG);

	ino = cellfs_inode_by_name(dir, &dentry->d_name);
	inode = ino ? cellfs_iget(dir->i_sb, CELLFS_I(dir)->i_root, ino) : NULL;
	return d_splice_alias(inode, dentry);
}

/*
 * By the time this is called, we already have created
 * the directory cache entry for the new file, but it
 * is so far negative - it has no inode.
 *
 * If the create succeeds, we fill in the inode information
 * with d_instantiate().
 */
static int cellfs_create(struct inode *dir, struct dentry *dentry, umode_t mode,
			bool excl)
{
	struct inode *inode;
	struct cellfs_transaction_info ti;
	int err;

	err = cellfs_transaction_begin(dir->i_sb, &ti, 1);
	if (err)
		return err;
	inode = cellfs_new_inode(dir, mode);
	err = PTR_ERR(inode);
	if (!IS_ERR(inode)) {
		inode->i_op = &cellfs_file_inode_operations;
		inode->i_fop = &cellfs_file_operations;
		inode->i_mapping->a_ops = &cellfs_aops;
		cellfs_mark_inode_dirty(inode);
		err = cellfs_add_nondir(dentry, inode);
	}
	if (!err)
		err = cellfs_transaction_commit(dir->i_sb);
	else
		cellfs_transaction_abort(dir->i_sb);

	return err;
}

static int
cellfs_mknod(struct inode *dir, struct dentry *dentry, umode_t mode, dev_t rdev)
{
	struct inode *inode;
	struct cellfs_transaction_info ti;
	int err;

	err = cellfs_transaction_begin(dir->i_sb, &ti, 1);
	if (err)
		return err;
	inode = cellfs_new_inode(dir, mode);
	err = PTR_ERR(inode);
	if (!IS_ERR(inode)) {
		init_special_inode(inode, inode->i_mode, rdev);
		cellfs_mark_inode_dirty(inode);
		err = cellfs_add_nondir(dentry, inode);
	}
	if (!err)
		err = cellfs_transaction_commit(dir->i_sb);
	else
		cellfs_transaction_abort(dir->i_sb);

	return err;
}

static int cellfs_symlink(struct inode *dir, struct dentry *dentry,
			 const char *symname)
{
	struct cellfs_transaction_info ti;
	struct super_block *sb = dir->i_sb;
	unsigned int l = strlen(symname) + 1;
	struct inode *inode;
	int err;

	if (l > sb->s_blocksize)
		return -ENAMETOOLONG;

	err = cellfs_transaction_begin(dir->i_sb, &ti, 1);
	if (err)
		return err;

	inode = cellfs_new_inode(dir, S_IFLNK | S_IRWXUGO);
	err = PTR_ERR(inode);
	if (IS_ERR(inode))
		goto out;

	/* slow symlink */
	inode->i_op = &cellfs_symlink_inode_operations;
	inode_nohighmem(inode);
	inode->i_mapping->a_ops = &cellfs_aops;
	err = page_symlink(inode, symname, l);
	if (err)
		goto out_fail;

	/* mark_inode_dirty(inode); */
	/* page_symlink() do this */

	err = cellfs_add_nondir(dentry, inode);
out:
	if (!err)
		err = cellfs_transaction_commit(dir->i_sb);
	else
		cellfs_transaction_abort(dir->i_sb);

	return err;

out_fail:
	drop_nlink(inode);
	cellfs_mark_inode_dirty(inode);
	unlock_new_inode(inode);
	iput(inode);
	goto out;
}

static int cellfs_link(struct dentry *old_dentry, struct inode *dir,
		      struct dentry *dentry)
{
	struct inode *inode = d_inode(old_dentry);
	struct cellfs_transaction_info ti;
	int err;

	err = cellfs_transaction_begin(dir->i_sb, &ti, 1);
	if (err)
		return err;

	inode->i_ctime = current_time(inode);
	inode_inc_link_count(inode);
	ihold(inode);

	err = cellfs_add_link(dentry, inode);
	if (!err) {
		d_instantiate(dentry, inode);
		err = cellfs_transaction_commit(dir->i_sb);
	} else {
		inode_dec_link_count(inode);
		iput(inode);
		cellfs_transaction_abort(dir->i_sb);
	}

	return err;
}

static int cellfs_mkdir(struct inode *dir, struct dentry *dentry, umode_t mode)
{
	struct inode *inode;
	struct cellfs_transaction_info ti;
	int err;

	err = cellfs_transaction_begin(dir->i_sb, &ti, 1);
	if (err)
		return err;

	inc_nlink(dir);

	inode = cellfs_new_inode(dir, S_IFDIR | mode);
	err = PTR_ERR(inode);
	if (IS_ERR(inode))
		goto out_dir;

	inode->i_op = &cellfs_dir_inode_operations;
	inode->i_fop = &cellfs_dir_operations;
	inode->i_mapping->a_ops = &cellfs_aops;

	inc_nlink(inode);

	err = cellfs_make_empty(inode, dir);
	if (err)
		goto out_fail;

	err = cellfs_add_link(dentry, inode);
	if (err)
		goto out_fail;

	cellfs_mark_inode_dirty(inode);
	d_instantiate(dentry, inode);
	unlock_new_inode(inode);
out:
	if (!err)
		err = cellfs_transaction_commit(dir->i_sb);
	else
		cellfs_transaction_abort(dir->i_sb);

	return err;

out_fail:
	drop_nlink(inode);
	drop_nlink(inode);
	cellfs_mark_inode_dirty(inode);
	unlock_new_inode(inode);
	iput(inode);
out_dir:
	drop_nlink(dir);
	cellfs_mark_inode_dirty(dir);
	goto out;
}

static int cellfs_do_unlink(struct inode *dir, struct dentry *dentry)
{
	struct inode *inode;
	struct cellfs_dir_entry *de;
	struct page *page;
	int err;

	err = -ENOENT;
	de = cellfs_find_entry(dir, &dentry->d_name, &page);
	if (!de)
		goto out;

	inode = d_inode(dentry);
	err = -EIO;
	if (le64_to_cpu(de->inode) != inode->i_ino)
		goto out;

	if (!inode->i_nlink) {
		cellfs_msg(inode->i_sb, KERN_WARNING,
			  "deleting nonexistent file (ino=%lu), %d",
			  inode->i_ino, inode->i_nlink);
		set_nlink(inode, 1);
	}
	err = cellfs_delete_entry(de, page);
	if (err)
		goto out;

	inode->i_ctime = dir->i_ctime;
	drop_nlink(inode);
	err = 0;
out:
	return err;
}

static int cellfs_unlink(struct inode *dir, struct dentry *dentry)
{
	struct cellfs_transaction_info ti;
	int err;

	err = cellfs_transaction_begin(dir->i_sb, &ti, 0);
	if (err)
		return err;

	err = cellfs_do_unlink(dir, dentry);

	if (!err) {
		cellfs_mark_inode_dirty(dir);
		cellfs_mark_inode_dirty(d_inode(dentry));
		err = cellfs_transaction_commit(dir->i_sb);
	} else
		cellfs_transaction_abort(dir->i_sb);

	return err;
}

static int cellfs_rmdir(struct inode *dir, struct dentry *dentry)
{
	struct inode *inode = d_inode(dentry);
	struct cellfs_transaction_info ti;
	int err;

	err = cellfs_transaction_begin(dir->i_sb, &ti, 0);
	if (err)
		return err;

	err = -ENOTEMPTY;
	if (cellfs_empty_dir(inode)) {
		err = cellfs_do_unlink(dir, dentry);
		if (!err) {
			inode->i_size = 0;
			drop_nlink(inode);
			cellfs_mark_inode_dirty(inode);
			drop_nlink(dir);
			cellfs_mark_inode_dirty(dir);
		}
	}
	if (!err)
		err = cellfs_transaction_commit(dir->i_sb);
	else
		cellfs_transaction_abort(dir->i_sb);

	return err;
}

static int cellfs_rename(struct inode *old_dir, struct dentry *old_dentry,
			struct inode *new_dir,	struct dentry *new_dentry,
			unsigned int flags)
{
	struct inode *old_inode = d_inode(old_dentry);
	struct inode *new_inode = d_inode(new_dentry);
	struct page *dir_page = NULL;
	struct cellfs_dir_entry *dir_de = NULL;
	struct page *old_page;
	struct cellfs_dir_entry *old_de;
	struct cellfs_transaction_info ti;
	int err;

	if (flags & ~RENAME_NOREPLACE)
		return -EINVAL;

	err = cellfs_transaction_begin(old_dir->i_sb, &ti, 1);
	if (unlikely(err))
		return err;

	err = -ENOENT;
	old_de = cellfs_find_entry(old_dir, &old_dentry->d_name, &old_page);
	if (!old_de)
		goto out;

	if (S_ISDIR(old_inode->i_mode)) {
		err = -EIO;
		dir_de = cellfs_dotdot(old_inode, &dir_page);
		if (!dir_de)
			goto out_old;
	}

	if (new_inode) {
		struct page *new_page;
		struct cellfs_dir_entry *new_de;

		err = -ENOTEMPTY;
		if (dir_de && !cellfs_empty_dir(new_inode))
			goto out_dir;

		err = -ENOENT;
		new_de = cellfs_find_entry(new_dir, &new_dentry->d_name, &new_page);
		if (!new_de)
			goto out_dir;
		cellfs_set_link(new_dir, new_de, new_page, old_inode);
		cellfs_mark_inode_dirty(new_dir);
		new_inode->i_ctime = current_time(new_inode);
		if (dir_de)
			drop_nlink(new_inode);
		drop_nlink(new_inode);
		cellfs_mark_inode_dirty(new_inode);
	} else {
		err = cellfs_add_link(new_dentry, old_inode);
		if (err)
			goto out_dir;
		if (dir_de) {
			inc_nlink(new_dir);
			cellfs_mark_inode_dirty(new_dir);
		}
	}

	/*
	 * Like most other Unix systems, set the ctime for inodes on a
	 * rename.
	 */
	old_inode->i_ctime = current_time(old_inode);

	cellfs_delete_entry(old_de, old_page);

	if (dir_de) {
		cellfs_set_link(old_inode, dir_de, dir_page, new_dir);
		drop_nlink(old_dir);
	}
	cellfs_mark_inode_dirty(old_dir);
	cellfs_mark_inode_dirty(old_inode);

	err = cellfs_transaction_commit(old_dir->i_sb);
	return err;

out_dir:
	if (dir_de) {
		kunmap(dir_page);
		put_page(dir_page);
	}
out_old:
	kunmap(old_page);
	put_page(old_page);
out:
	cellfs_transaction_abort(old_dir->i_sb);
	return err;
}

/*
 * Export operations
 */
static struct dentry *cellfs_get_parent(struct dentry *child)
{
	unsigned long ino;
	struct inode *inode;
	struct qstr dotdot = QSTR_INIT("..", 2);
	struct cellfs_root *root;

	ino = cellfs_inode_by_name(d_inode(child), &dotdot);
	if (!ino)
		return ERR_PTR(-ENOENT);

	root = CELLFS_I(d_inode(child))->i_root;

	inode = cellfs_iget(child->d_sb, root, ino);
	if (IS_ERR(inode))
		return ERR_CAST(inode);

	return d_obtain_alias(inode);
}

static struct dentry *cellfs_get_dentry(struct super_block *sb, u64 cno,
				       u64 ino, u32 gen)
{
	struct cellfs_root *root;
	struct inode *inode;

	if (ino < CELLFS_FIRST_INO(sb) && ino != CELLFS_ROOT_INO)
		return ERR_PTR(-ESTALE);

	root = cellfs_lookup_root(sb->s_fs_info, cno);
	if (!root)
		return ERR_PTR(-ESTALE);

	inode = cellfs_iget(sb, root, ino);
	cellfs_put_root(root);

	if (IS_ERR(inode))
		return ERR_CAST(inode);
	if (gen && inode->i_generation != gen) {
		iput(inode);
		return ERR_PTR(-ESTALE);
	}
	return d_obtain_alias(inode);
}

static struct dentry *cellfs_fh_to_dentry(struct super_block *sb, struct fid *fh,
					 int fh_len, int fh_type)
{
	struct cellfs_fid *fid = (struct cellfs_fid *)fh;

	if (fh_len < CELLFS_FID_SIZE_NON_CONNECTABLE ||
	    (fh_type != FILEID_CELLFS_WITH_PARENT &&
	     fh_type != FILEID_CELLFS_WITHOUT_PARENT))
		return NULL;

	return cellfs_get_dentry(sb, fid->cno, fid->ino, fid->gen);
}

static struct dentry *cellfs_fh_to_parent(struct super_block *sb, struct fid *fh,
					 int fh_len, int fh_type)
{
	struct cellfs_fid *fid = (struct cellfs_fid *)fh;

	if (fh_len < CELLFS_FID_SIZE_CONNECTABLE ||
	    fh_type != FILEID_CELLFS_WITH_PARENT)
		return NULL;

	return cellfs_get_dentry(sb, fid->cno, fid->parent_ino, fid->parent_gen);
}

static int cellfs_encode_fh(struct inode *inode, __u32 *fh, int *lenp,
			   struct inode *parent)
{
	struct cellfs_fid *fid = (struct cellfs_fid *)fh;
	struct cellfs_root *root = CELLFS_I(inode)->i_root;
	int type;

	if (parent && *lenp < CELLFS_FID_SIZE_CONNECTABLE) {
		*lenp = CELLFS_FID_SIZE_CONNECTABLE;
		return FILEID_INVALID;
	}
	if (*lenp < CELLFS_FID_SIZE_NON_CONNECTABLE) {
		*lenp = CELLFS_FID_SIZE_NON_CONNECTABLE;
		return FILEID_INVALID;
	}

	fid->cno = root->cno;
	fid->ino = inode->i_ino;
	fid->gen = inode->i_generation;

	if (parent) {
		fid->parent_ino = parent->i_ino;
		fid->parent_gen = parent->i_generation;
		type = FILEID_CELLFS_WITH_PARENT;
		*lenp = CELLFS_FID_SIZE_CONNECTABLE;
	} else {
		type = FILEID_CELLFS_WITHOUT_PARENT;
		*lenp = CELLFS_FID_SIZE_NON_CONNECTABLE;
	}

	return type;
}

const struct inode_operations cellfs_dir_inode_operations = {
	.create		= cellfs_create,
	.lookup		= cellfs_lookup,
	.link		= cellfs_link,
	.unlink		= cellfs_unlink,
	.symlink	= cellfs_symlink,
	.mkdir		= cellfs_mkdir,
	.rmdir		= cellfs_rmdir,
	.mknod		= cellfs_mknod,
	.rename		= cellfs_rename,
	.setattr	= cellfs_setattr,
	.permission	= cellfs_permission,
	.fiemap		= cellfs_fiemap,
};

const struct inode_operations cellfs_special_inode_operations = {
	.setattr	= cellfs_setattr,
	.permission	= cellfs_permission,
};

const struct inode_operations cellfs_symlink_inode_operations = {
	.get_link	= page_get_link,
	.permission     = cellfs_permission,
};

const struct export_operations cellfs_export_ops = {
	.encode_fh = cellfs_encode_fh,
	.fh_to_dentry = cellfs_fh_to_dentry,
	.fh_to_parent = cellfs_fh_to_parent,
	.get_parent = cellfs_get_parent,
};
