/*
 * sysfs.c - sysfs support implementation.
 *
 * Copyright (C) 2005-2014 Nippon Telegraph and Telephone Corporation.
 * Copyright (C) 2014 HGST, Inc., a Western Digital Company.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * Written by Vyacheslav Dubeyko <Vyacheslav.Dubeyko@hgst.com>
 */

#include <linux/kobject.h>

#include "cellfs.h"
#include "mdt.h"
#include "sufile.h"
#include "cpfile.h"
#include "sysfs.h"

/* /sys/fs/<cellfs>/ */
static struct kset *cellfs_kset;

#define CELLFS_SHOW_TIME(time_t_val, buf) ({ \
		struct tm res; \
		int count = 0; \
		time_to_tm(time_t_val, 0, &res); \
		res.tm_year += 1900; \
		res.tm_mon += 1; \
		count = scnprintf(buf, PAGE_SIZE, \
				    "%ld-%.2d-%.2d %.2d:%.2d:%.2d\n", \
				    res.tm_year, res.tm_mon, res.tm_mday, \
				    res.tm_hour, res.tm_min, res.tm_sec);\
		count; \
})

#define CELLFS_DEV_INT_GROUP_OPS(name, parent_name) \
static ssize_t cellfs_##name##_attr_show(struct kobject *kobj, \
					struct attribute *attr, char *buf) \
{ \
	struct the_cellfs *cellfs = container_of(kobj->parent, \
						struct the_cellfs, \
						ns_##parent_name##_kobj); \
	struct cellfs_##name##_attr *a = container_of(attr, \
						struct cellfs_##name##_attr, \
						attr); \
	return a->show ? a->show(a, cellfs, buf) : 0; \
} \
static ssize_t cellfs_##name##_attr_store(struct kobject *kobj, \
					 struct attribute *attr, \
					 const char *buf, size_t len) \
{ \
	struct the_cellfs *cellfs = container_of(kobj->parent, \
						struct the_cellfs, \
						ns_##parent_name##_kobj); \
	struct cellfs_##name##_attr *a = container_of(attr, \
						struct cellfs_##name##_attr, \
						attr); \
	return a->store ? a->store(a, cellfs, buf, len) : 0; \
} \
static const struct sysfs_ops cellfs_##name##_attr_ops = { \
	.show	= cellfs_##name##_attr_show, \
	.store	= cellfs_##name##_attr_store, \
}

#define CELLFS_DEV_INT_GROUP_TYPE(name, parent_name) \
static void cellfs_##name##_attr_release(struct kobject *kobj) \
{ \
	struct cellfs_sysfs_##parent_name##_subgroups *subgroups; \
	struct the_cellfs *cellfs = container_of(kobj->parent, \
						struct the_cellfs, \
						ns_##parent_name##_kobj); \
	subgroups = cellfs->ns_##parent_name##_subgroups; \
	complete(&subgroups->sg_##name##_kobj_unregister); \
} \
static struct kobj_type cellfs_##name##_ktype = { \
	.default_attrs	= cellfs_##name##_attrs, \
	.sysfs_ops	= &cellfs_##name##_attr_ops, \
	.release	= cellfs_##name##_attr_release, \
}

#define CELLFS_DEV_INT_GROUP_FNS(name, parent_name) \
static int cellfs_sysfs_create_##name##_group(struct the_cellfs *cellfs) \
{ \
	struct kobject *parent; \
	struct kobject *kobj; \
	struct completion *kobj_unregister; \
	struct cellfs_sysfs_##parent_name##_subgroups *subgroups; \
	int err; \
	subgroups = cellfs->ns_##parent_name##_subgroups; \
	kobj = &subgroups->sg_##name##_kobj; \
	kobj_unregister = &subgroups->sg_##name##_kobj_unregister; \
	parent = &cellfs->ns_##parent_name##_kobj; \
	kobj->kset = cellfs_kset; \
	init_completion(kobj_unregister); \
	err = kobject_init_and_add(kobj, &cellfs_##name##_ktype, parent, \
				    #name); \
	if (err) \
		return err; \
	return 0; \
} \
static void cellfs_sysfs_delete_##name##_group(struct the_cellfs *cellfs) \
{ \
	kobject_del(&cellfs->ns_##parent_name##_subgroups->sg_##name##_kobj); \
}

/************************************************************************
 *                        CELLFS snapshot attrs                          *
 ************************************************************************/

static ssize_t
cellfs_snapshot_inodes_count_show(struct cellfs_snapshot_attr *attr,
				 struct cellfs_root *root, char *buf)
{
	return snprintf(buf, PAGE_SIZE, "%llu\n",
			(unsigned long long)atomic64_read(&root->inodes_count));
}

static ssize_t
cellfs_snapshot_blocks_count_show(struct cellfs_snapshot_attr *attr,
				 struct cellfs_root *root, char *buf)
{
	return snprintf(buf, PAGE_SIZE, "%llu\n",
			(unsigned long long)atomic64_read(&root->blocks_count));
}

static const char snapshot_readme_str[] =
	"The group contains details about mounted snapshot.\n\n"
	"(1) inodes_count\n\tshow number of inodes for snapshot.\n\n"
	"(2) blocks_count\n\tshow number of blocks for snapshot.\n\n";

static ssize_t
cellfs_snapshot_README_show(struct cellfs_snapshot_attr *attr,
			    struct cellfs_root *root, char *buf)
{
	return snprintf(buf, PAGE_SIZE, snapshot_readme_str);
}

CELLFS_SNAPSHOT_RO_ATTR(inodes_count);
CELLFS_SNAPSHOT_RO_ATTR(blocks_count);
CELLFS_SNAPSHOT_RO_ATTR(README);

static struct attribute *cellfs_snapshot_attrs[] = {
	CELLFS_SNAPSHOT_ATTR_LIST(inodes_count),
	CELLFS_SNAPSHOT_ATTR_LIST(blocks_count),
	CELLFS_SNAPSHOT_ATTR_LIST(README),
	NULL,
};

static ssize_t cellfs_snapshot_attr_show(struct kobject *kobj,
					struct attribute *attr, char *buf)
{
	struct cellfs_root *root =
			container_of(kobj, struct cellfs_root, snapshot_kobj);
	struct cellfs_snapshot_attr *a =
			container_of(attr, struct cellfs_snapshot_attr, attr);

	return a->show ? a->show(a, root, buf) : 0;
}

static ssize_t cellfs_snapshot_attr_store(struct kobject *kobj,
					 struct attribute *attr,
					 const char *buf, size_t len)
{
	struct cellfs_root *root =
			container_of(kobj, struct cellfs_root, snapshot_kobj);
	struct cellfs_snapshot_attr *a =
			container_of(attr, struct cellfs_snapshot_attr, attr);

	return a->store ? a->store(a, root, buf, len) : 0;
}

static void cellfs_snapshot_attr_release(struct kobject *kobj)
{
	struct cellfs_root *root = container_of(kobj, struct cellfs_root,
						snapshot_kobj);
	complete(&root->snapshot_kobj_unregister);
}

static const struct sysfs_ops cellfs_snapshot_attr_ops = {
	.show	= cellfs_snapshot_attr_show,
	.store	= cellfs_snapshot_attr_store,
};

static struct kobj_type cellfs_snapshot_ktype = {
	.default_attrs	= cellfs_snapshot_attrs,
	.sysfs_ops	= &cellfs_snapshot_attr_ops,
	.release	= cellfs_snapshot_attr_release,
};

int cellfs_sysfs_create_snapshot_group(struct cellfs_root *root)
{
	struct the_cellfs *cellfs;
	struct kobject *parent;
	int err;

	cellfs = root->cellfs;
	parent = &cellfs->ns_dev_subgroups->sg_mounted_snapshots_kobj;
	root->snapshot_kobj.kset = cellfs_kset;
	init_completion(&root->snapshot_kobj_unregister);

	if (root->cno == CELLFS_CPTREE_CURRENT_CNO) {
		err = kobject_init_and_add(&root->snapshot_kobj,
					    &cellfs_snapshot_ktype,
					    &cellfs->ns_dev_kobj,
					    "current_checkpoint");
	} else {
		err = kobject_init_and_add(&root->snapshot_kobj,
					    &cellfs_snapshot_ktype,
					    parent,
					    "%llu", root->cno);
	}

	if (err)
		return err;

	return 0;
}

void cellfs_sysfs_delete_snapshot_group(struct cellfs_root *root)
{
	kobject_del(&root->snapshot_kobj);
}

/************************************************************************
 *                    CELLFS mounted snapshots attrs                     *
 ************************************************************************/

static const char mounted_snapshots_readme_str[] =
	"The mounted_snapshots group contains group for\n"
	"every mounted snapshot.\n";

static ssize_t
cellfs_mounted_snapshots_README_show(struct cellfs_mounted_snapshots_attr *attr,
				    struct the_cellfs *cellfs, char *buf)
{
	return snprintf(buf, PAGE_SIZE, mounted_snapshots_readme_str);
}

CELLFS_MOUNTED_SNAPSHOTS_RO_ATTR(README);

static struct attribute *cellfs_mounted_snapshots_attrs[] = {
	CELLFS_MOUNTED_SNAPSHOTS_ATTR_LIST(README),
	NULL,
};

CELLFS_DEV_INT_GROUP_OPS(mounted_snapshots, dev);
CELLFS_DEV_INT_GROUP_TYPE(mounted_snapshots, dev);
CELLFS_DEV_INT_GROUP_FNS(mounted_snapshots, dev);

/************************************************************************
 *                      CELLFS checkpoints attrs                         *
 ************************************************************************/

static ssize_t
cellfs_checkpoints_checkpoints_number_show(struct cellfs_checkpoints_attr *attr,
					    struct the_cellfs *cellfs,
					    char *buf)
{
	__u64 ncheckpoints;
	struct cellfs_cpstat cpstat;
	int err;

	down_read(&cellfs->ns_segctor_sem);
	err = cellfs_cpfile_get_stat(cellfs->ns_cpfile, &cpstat);
	up_read(&cellfs->ns_segctor_sem);
	if (err < 0) {
		cellfs_msg(cellfs->ns_sb, KERN_ERR,
			  "unable to get checkpoint stat: err=%d", err);
		return err;
	}

	ncheckpoints = cpstat.cs_ncps;

	return snprintf(buf, PAGE_SIZE, "%llu\n", ncheckpoints);
}

static ssize_t
cellfs_checkpoints_snapshots_number_show(struct cellfs_checkpoints_attr *attr,
					struct the_cellfs *cellfs,
					char *buf)
{
	__u64 nsnapshots;
	struct cellfs_cpstat cpstat;
	int err;

	down_read(&cellfs->ns_segctor_sem);
	err = cellfs_cpfile_get_stat(cellfs->ns_cpfile, &cpstat);
	up_read(&cellfs->ns_segctor_sem);
	if (err < 0) {
		cellfs_msg(cellfs->ns_sb, KERN_ERR,
			  "unable to get checkpoint stat: err=%d", err);
		return err;
	}

	nsnapshots = cpstat.cs_nsss;

	return snprintf(buf, PAGE_SIZE, "%llu\n", nsnapshots);
}

static ssize_t
cellfs_checkpoints_last_seg_checkpoint_show(struct cellfs_checkpoints_attr *attr,
					    struct the_cellfs *cellfs,
					    char *buf)
{
	__u64 last_cno;

	spin_lock(&cellfs->ns_last_segment_lock);
	last_cno = cellfs->ns_last_cno;
	spin_unlock(&cellfs->ns_last_segment_lock);

	return snprintf(buf, PAGE_SIZE, "%llu\n", last_cno);
}

static ssize_t
cellfs_checkpoints_next_checkpoint_show(struct cellfs_checkpoints_attr *attr,
					struct the_cellfs *cellfs,
					char *buf)
{
	__u64 cno;

	down_read(&cellfs->ns_segctor_sem);
	cno = cellfs->ns_cno;
	up_read(&cellfs->ns_segctor_sem);

	return snprintf(buf, PAGE_SIZE, "%llu\n", cno);
}

static const char checkpoints_readme_str[] =
	"The checkpoints group contains attributes that describe\n"
	"details about volume's checkpoints.\n\n"
	"(1) checkpoints_number\n\tshow number of checkpoints on volume.\n\n"
	"(2) snapshots_number\n\tshow number of snapshots on volume.\n\n"
	"(3) last_seg_checkpoint\n"
	"\tshow checkpoint number of the latest segment.\n\n"
	"(4) next_checkpoint\n\tshow next checkpoint number.\n\n";

static ssize_t
cellfs_checkpoints_README_show(struct cellfs_checkpoints_attr *attr,
				struct the_cellfs *cellfs, char *buf)
{
	return snprintf(buf, PAGE_SIZE, checkpoints_readme_str);
}

CELLFS_CHECKPOINTS_RO_ATTR(checkpoints_number);
CELLFS_CHECKPOINTS_RO_ATTR(snapshots_number);
CELLFS_CHECKPOINTS_RO_ATTR(last_seg_checkpoint);
CELLFS_CHECKPOINTS_RO_ATTR(next_checkpoint);
CELLFS_CHECKPOINTS_RO_ATTR(README);

static struct attribute *cellfs_checkpoints_attrs[] = {
	CELLFS_CHECKPOINTS_ATTR_LIST(checkpoints_number),
	CELLFS_CHECKPOINTS_ATTR_LIST(snapshots_number),
	CELLFS_CHECKPOINTS_ATTR_LIST(last_seg_checkpoint),
	CELLFS_CHECKPOINTS_ATTR_LIST(next_checkpoint),
	CELLFS_CHECKPOINTS_ATTR_LIST(README),
	NULL,
};

CELLFS_DEV_INT_GROUP_OPS(checkpoints, dev);
CELLFS_DEV_INT_GROUP_TYPE(checkpoints, dev);
CELLFS_DEV_INT_GROUP_FNS(checkpoints, dev);

/************************************************************************
 *                        CELLFS segments attrs                          *
 ************************************************************************/

static ssize_t
cellfs_segments_segments_number_show(struct cellfs_segments_attr *attr,
				     struct the_cellfs *cellfs,
				     char *buf)
{
	return snprintf(buf, PAGE_SIZE, "%lu\n", cellfs->ns_nsegments);
}

static ssize_t
cellfs_segments_blocks_per_segment_show(struct cellfs_segments_attr *attr,
					struct the_cellfs *cellfs,
					char *buf)
{
	return snprintf(buf, PAGE_SIZE, "%lu\n", cellfs->ns_blocks_per_segment);
}

static ssize_t
cellfs_segments_clean_segments_show(struct cellfs_segments_attr *attr,
				    struct the_cellfs *cellfs,
				    char *buf)
{
	unsigned long ncleansegs;

	down_read(&CELLFS_MDT(cellfs->ns_dat)->mi_sem);
	ncleansegs = cellfs_sufile_get_ncleansegs(cellfs->ns_sufile);
	up_read(&CELLFS_MDT(cellfs->ns_dat)->mi_sem);

	return snprintf(buf, PAGE_SIZE, "%lu\n", ncleansegs);
}

static ssize_t
cellfs_segments_dirty_segments_show(struct cellfs_segments_attr *attr,
				    struct the_cellfs *cellfs,
				    char *buf)
{
	struct cellfs_sustat sustat;
	int err;

	down_read(&cellfs->ns_segctor_sem);
	err = cellfs_sufile_get_stat(cellfs->ns_sufile, &sustat);
	up_read(&cellfs->ns_segctor_sem);
	if (err < 0) {
		cellfs_msg(cellfs->ns_sb, KERN_ERR,
			  "unable to get segment stat: err=%d", err);
		return err;
	}

	return snprintf(buf, PAGE_SIZE, "%llu\n", sustat.ss_ndirtysegs);
}

static const char segments_readme_str[] =
	"The segments group contains attributes that describe\n"
	"details about volume's segments.\n\n"
	"(1) segments_number\n\tshow number of segments on volume.\n\n"
	"(2) blocks_per_segment\n\tshow number of blocks in segment.\n\n"
	"(3) clean_segments\n\tshow count of clean segments.\n\n"
	"(4) dirty_segments\n\tshow count of dirty segments.\n\n";

static ssize_t
cellfs_segments_README_show(struct cellfs_segments_attr *attr,
			    struct the_cellfs *cellfs,
			    char *buf)
{
	return snprintf(buf, PAGE_SIZE, segments_readme_str);
}

CELLFS_SEGMENTS_RO_ATTR(segments_number);
CELLFS_SEGMENTS_RO_ATTR(blocks_per_segment);
CELLFS_SEGMENTS_RO_ATTR(clean_segments);
CELLFS_SEGMENTS_RO_ATTR(dirty_segments);
CELLFS_SEGMENTS_RO_ATTR(README);

static struct attribute *cellfs_segments_attrs[] = {
	CELLFS_SEGMENTS_ATTR_LIST(segments_number),
	CELLFS_SEGMENTS_ATTR_LIST(blocks_per_segment),
	CELLFS_SEGMENTS_ATTR_LIST(clean_segments),
	CELLFS_SEGMENTS_ATTR_LIST(dirty_segments),
	CELLFS_SEGMENTS_ATTR_LIST(README),
	NULL,
};

CELLFS_DEV_INT_GROUP_OPS(segments, dev);
CELLFS_DEV_INT_GROUP_TYPE(segments, dev);
CELLFS_DEV_INT_GROUP_FNS(segments, dev);

/************************************************************************
 *                        CELLFS segctor attrs                           *
 ************************************************************************/

static ssize_t
cellfs_segctor_last_pseg_block_show(struct cellfs_segctor_attr *attr,
				    struct the_cellfs *cellfs,
				    char *buf)
{
	sector_t last_pseg;

	spin_lock(&cellfs->ns_last_segment_lock);
	last_pseg = cellfs->ns_last_pseg;
	spin_unlock(&cellfs->ns_last_segment_lock);

	return snprintf(buf, PAGE_SIZE, "%llu\n",
			(unsigned long long)last_pseg);
}

static ssize_t
cellfs_segctor_last_seg_sequence_show(struct cellfs_segctor_attr *attr,
					struct the_cellfs *cellfs,
					char *buf)
{
	u64 last_seq;

	spin_lock(&cellfs->ns_last_segment_lock);
	last_seq = cellfs->ns_last_seq;
	spin_unlock(&cellfs->ns_last_segment_lock);

	return snprintf(buf, PAGE_SIZE, "%llu\n", last_seq);
}

static ssize_t
cellfs_segctor_last_seg_checkpoint_show(struct cellfs_segctor_attr *attr,
					struct the_cellfs *cellfs,
					char *buf)
{
	__u64 last_cno;

	spin_lock(&cellfs->ns_last_segment_lock);
	last_cno = cellfs->ns_last_cno;
	spin_unlock(&cellfs->ns_last_segment_lock);

	return snprintf(buf, PAGE_SIZE, "%llu\n", last_cno);
}

static ssize_t
cellfs_segctor_current_seg_sequence_show(struct cellfs_segctor_attr *attr,
					struct the_cellfs *cellfs,
					char *buf)
{
	u64 seg_seq;

	down_read(&cellfs->ns_segctor_sem);
	seg_seq = cellfs->ns_seg_seq;
	up_read(&cellfs->ns_segctor_sem);

	return snprintf(buf, PAGE_SIZE, "%llu\n", seg_seq);
}

static ssize_t
cellfs_segctor_current_last_full_seg_show(struct cellfs_segctor_attr *attr,
					 struct the_cellfs *cellfs,
					 char *buf)
{
	__u64 segnum;

	down_read(&cellfs->ns_segctor_sem);
	segnum = cellfs->ns_segnum;
	up_read(&cellfs->ns_segctor_sem);

	return snprintf(buf, PAGE_SIZE, "%llu\n", segnum);
}

static ssize_t
cellfs_segctor_next_full_seg_show(struct cellfs_segctor_attr *attr,
				 struct the_cellfs *cellfs,
				 char *buf)
{
	__u64 nextnum;

	down_read(&cellfs->ns_segctor_sem);
	nextnum = cellfs->ns_nextnum;
	up_read(&cellfs->ns_segctor_sem);

	return snprintf(buf, PAGE_SIZE, "%llu\n", nextnum);
}

static ssize_t
cellfs_segctor_next_pseg_offset_show(struct cellfs_segctor_attr *attr,
					struct the_cellfs *cellfs,
					char *buf)
{
	unsigned long pseg_offset;

	down_read(&cellfs->ns_segctor_sem);
	pseg_offset = cellfs->ns_pseg_offset;
	up_read(&cellfs->ns_segctor_sem);

	return snprintf(buf, PAGE_SIZE, "%lu\n", pseg_offset);
}

static ssize_t
cellfs_segctor_next_checkpoint_show(struct cellfs_segctor_attr *attr,
					struct the_cellfs *cellfs,
					char *buf)
{
	__u64 cno;

	down_read(&cellfs->ns_segctor_sem);
	cno = cellfs->ns_cno;
	up_read(&cellfs->ns_segctor_sem);

	return snprintf(buf, PAGE_SIZE, "%llu\n", cno);
}

static ssize_t
cellfs_segctor_last_seg_write_time_show(struct cellfs_segctor_attr *attr,
					struct the_cellfs *cellfs,
					char *buf)
{
	time_t ctime;

	down_read(&cellfs->ns_segctor_sem);
	ctime = cellfs->ns_ctime;
	up_read(&cellfs->ns_segctor_sem);

	return CELLFS_SHOW_TIME(ctime, buf);
}

static ssize_t
cellfs_segctor_last_seg_write_time_secs_show(struct cellfs_segctor_attr *attr,
					    struct the_cellfs *cellfs,
					    char *buf)
{
	time_t ctime;

	down_read(&cellfs->ns_segctor_sem);
	ctime = cellfs->ns_ctime;
	up_read(&cellfs->ns_segctor_sem);

	return snprintf(buf, PAGE_SIZE, "%llu\n", (unsigned long long)ctime);
}

static ssize_t
cellfs_segctor_last_nongc_write_time_show(struct cellfs_segctor_attr *attr,
					 struct the_cellfs *cellfs,
					 char *buf)
{
	time_t nongc_ctime;

	down_read(&cellfs->ns_segctor_sem);
	nongc_ctime = cellfs->ns_nongc_ctime;
	up_read(&cellfs->ns_segctor_sem);

	return CELLFS_SHOW_TIME(nongc_ctime, buf);
}

static ssize_t
cellfs_segctor_last_nongc_write_time_secs_show(struct cellfs_segctor_attr *attr,
						struct the_cellfs *cellfs,
						char *buf)
{
	time_t nongc_ctime;

	down_read(&cellfs->ns_segctor_sem);
	nongc_ctime = cellfs->ns_nongc_ctime;
	up_read(&cellfs->ns_segctor_sem);

	return snprintf(buf, PAGE_SIZE, "%llu\n",
			(unsigned long long)nongc_ctime);
}

static ssize_t
cellfs_segctor_dirty_data_blocks_count_show(struct cellfs_segctor_attr *attr,
					    struct the_cellfs *cellfs,
					    char *buf)
{
	u32 ndirtyblks;

	down_read(&cellfs->ns_segctor_sem);
	ndirtyblks = atomic_read(&cellfs->ns_ndirtyblks);
	up_read(&cellfs->ns_segctor_sem);

	return snprintf(buf, PAGE_SIZE, "%u\n", ndirtyblks);
}

static const char segctor_readme_str[] =
	"The segctor group contains attributes that describe\n"
	"segctor thread activity details.\n\n"
	"(1) last_pseg_block\n"
	"\tshow start block number of the latest segment.\n\n"
	"(2) last_seg_sequence\n"
	"\tshow sequence value of the latest segment.\n\n"
	"(3) last_seg_checkpoint\n"
	"\tshow checkpoint number of the latest segment.\n\n"
	"(4) current_seg_sequence\n\tshow segment sequence counter.\n\n"
	"(5) current_last_full_seg\n"
	"\tshow index number of the latest full segment.\n\n"
	"(6) next_full_seg\n"
	"\tshow index number of the full segment index to be used next.\n\n"
	"(7) next_pseg_offset\n"
	"\tshow offset of next partial segment in the current full segment.\n\n"
	"(8) next_checkpoint\n\tshow next checkpoint number.\n\n"
	"(9) last_seg_write_time\n"
	"\tshow write time of the last segment in human-readable format.\n\n"
	"(10) last_seg_write_time_secs\n"
	"\tshow write time of the last segment in seconds.\n\n"
	"(11) last_nongc_write_time\n"
	"\tshow write time of the last segment not for cleaner operation "
	"in human-readable format.\n\n"
	"(12) last_nongc_write_time_secs\n"
	"\tshow write time of the last segment not for cleaner operation "
	"in seconds.\n\n"
	"(13) dirty_data_blocks_count\n"
	"\tshow number of dirty data blocks.\n\n";

static ssize_t
cellfs_segctor_README_show(struct cellfs_segctor_attr *attr,
			  struct the_cellfs *cellfs, char *buf)
{
	return snprintf(buf, PAGE_SIZE, segctor_readme_str);
}

CELLFS_SEGCTOR_RO_ATTR(last_pseg_block);
CELLFS_SEGCTOR_RO_ATTR(last_seg_sequence);
CELLFS_SEGCTOR_RO_ATTR(last_seg_checkpoint);
CELLFS_SEGCTOR_RO_ATTR(current_seg_sequence);
CELLFS_SEGCTOR_RO_ATTR(current_last_full_seg);
CELLFS_SEGCTOR_RO_ATTR(next_full_seg);
CELLFS_SEGCTOR_RO_ATTR(next_pseg_offset);
CELLFS_SEGCTOR_RO_ATTR(next_checkpoint);
CELLFS_SEGCTOR_RO_ATTR(last_seg_write_time);
CELLFS_SEGCTOR_RO_ATTR(last_seg_write_time_secs);
CELLFS_SEGCTOR_RO_ATTR(last_nongc_write_time);
CELLFS_SEGCTOR_RO_ATTR(last_nongc_write_time_secs);
CELLFS_SEGCTOR_RO_ATTR(dirty_data_blocks_count);
CELLFS_SEGCTOR_RO_ATTR(README);

static struct attribute *cellfs_segctor_attrs[] = {
	CELLFS_SEGCTOR_ATTR_LIST(last_pseg_block),
	CELLFS_SEGCTOR_ATTR_LIST(last_seg_sequence),
	CELLFS_SEGCTOR_ATTR_LIST(last_seg_checkpoint),
	CELLFS_SEGCTOR_ATTR_LIST(current_seg_sequence),
	CELLFS_SEGCTOR_ATTR_LIST(current_last_full_seg),
	CELLFS_SEGCTOR_ATTR_LIST(next_full_seg),
	CELLFS_SEGCTOR_ATTR_LIST(next_pseg_offset),
	CELLFS_SEGCTOR_ATTR_LIST(next_checkpoint),
	CELLFS_SEGCTOR_ATTR_LIST(last_seg_write_time),
	CELLFS_SEGCTOR_ATTR_LIST(last_seg_write_time_secs),
	CELLFS_SEGCTOR_ATTR_LIST(last_nongc_write_time),
	CELLFS_SEGCTOR_ATTR_LIST(last_nongc_write_time_secs),
	CELLFS_SEGCTOR_ATTR_LIST(dirty_data_blocks_count),
	CELLFS_SEGCTOR_ATTR_LIST(README),
	NULL,
};

CELLFS_DEV_INT_GROUP_OPS(segctor, dev);
CELLFS_DEV_INT_GROUP_TYPE(segctor, dev);
CELLFS_DEV_INT_GROUP_FNS(segctor, dev);

/************************************************************************
 *                        CELLFS superblock attrs                        *
 ************************************************************************/

static ssize_t
cellfs_superblock_sb_write_time_show(struct cellfs_superblock_attr *attr,
				     struct the_cellfs *cellfs,
				     char *buf)
{
	time_t sbwtime;

	down_read(&cellfs->ns_sem);
	sbwtime = cellfs->ns_sbwtime;
	up_read(&cellfs->ns_sem);

	return CELLFS_SHOW_TIME(sbwtime, buf);
}

static ssize_t
cellfs_superblock_sb_write_time_secs_show(struct cellfs_superblock_attr *attr,
					 struct the_cellfs *cellfs,
					 char *buf)
{
	time_t sbwtime;

	down_read(&cellfs->ns_sem);
	sbwtime = cellfs->ns_sbwtime;
	up_read(&cellfs->ns_sem);

	return snprintf(buf, PAGE_SIZE, "%llu\n", (unsigned long long)sbwtime);
}

static ssize_t
cellfs_superblock_sb_write_count_show(struct cellfs_superblock_attr *attr,
				      struct the_cellfs *cellfs,
				      char *buf)
{
	unsigned int sbwcount;

	down_read(&cellfs->ns_sem);
	sbwcount = cellfs->ns_sbwcount;
	up_read(&cellfs->ns_sem);

	return snprintf(buf, PAGE_SIZE, "%u\n", sbwcount);
}

static ssize_t
cellfs_superblock_sb_update_frequency_show(struct cellfs_superblock_attr *attr,
					    struct the_cellfs *cellfs,
					    char *buf)
{
	unsigned int sb_update_freq;

	down_read(&cellfs->ns_sem);
	sb_update_freq = cellfs->ns_sb_update_freq;
	up_read(&cellfs->ns_sem);

	return snprintf(buf, PAGE_SIZE, "%u\n", sb_update_freq);
}

static ssize_t
cellfs_superblock_sb_update_frequency_store(struct cellfs_superblock_attr *attr,
					    struct the_cellfs *cellfs,
					    const char *buf, size_t count)
{
	unsigned int val;
	int err;

	err = kstrtouint(skip_spaces(buf), 0, &val);
	if (err) {
		cellfs_msg(cellfs->ns_sb, KERN_ERR,
			  "unable to convert string: err=%d", err);
		return err;
	}

	if (val < CELLFS_SB_FREQ) {
		val = CELLFS_SB_FREQ;
		cellfs_msg(cellfs->ns_sb, KERN_WARNING,
			  "superblock update frequency cannot be lesser than 10 seconds");
	}

	down_write(&cellfs->ns_sem);
	cellfs->ns_sb_update_freq = val;
	up_write(&cellfs->ns_sem);

	return count;
}

static const char sb_readme_str[] =
	"The superblock group contains attributes that describe\n"
	"superblock's details.\n\n"
	"(1) sb_write_time\n\tshow previous write time of super block "
	"in human-readable format.\n\n"
	"(2) sb_write_time_secs\n\tshow previous write time of super block "
	"in seconds.\n\n"
	"(3) sb_write_count\n\tshow write count of super block.\n\n"
	"(4) sb_update_frequency\n"
	"\tshow/set interval of periodical update of superblock (in seconds).\n\n"
	"\tYou can set preferable frequency of superblock update by command:\n\n"
	"\t'echo <val> > /sys/fs/<cellfs>/<dev>/superblock/sb_update_frequency'\n";

static ssize_t
cellfs_superblock_README_show(struct cellfs_superblock_attr *attr,
				struct the_cellfs *cellfs, char *buf)
{
	return snprintf(buf, PAGE_SIZE, sb_readme_str);
}

CELLFS_SUPERBLOCK_RO_ATTR(sb_write_time);
CELLFS_SUPERBLOCK_RO_ATTR(sb_write_time_secs);
CELLFS_SUPERBLOCK_RO_ATTR(sb_write_count);
CELLFS_SUPERBLOCK_RW_ATTR(sb_update_frequency);
CELLFS_SUPERBLOCK_RO_ATTR(README);

static struct attribute *cellfs_superblock_attrs[] = {
	CELLFS_SUPERBLOCK_ATTR_LIST(sb_write_time),
	CELLFS_SUPERBLOCK_ATTR_LIST(sb_write_time_secs),
	CELLFS_SUPERBLOCK_ATTR_LIST(sb_write_count),
	CELLFS_SUPERBLOCK_ATTR_LIST(sb_update_frequency),
	CELLFS_SUPERBLOCK_ATTR_LIST(README),
	NULL,
};

CELLFS_DEV_INT_GROUP_OPS(superblock, dev);
CELLFS_DEV_INT_GROUP_TYPE(superblock, dev);
CELLFS_DEV_INT_GROUP_FNS(superblock, dev);

/************************************************************************
 *                        CELLFS device attrs                            *
 ************************************************************************/

static
ssize_t cellfs_dev_revision_show(struct cellfs_dev_attr *attr,
				struct the_cellfs *cellfs,
				char *buf)
{
	struct cellfs_super_block **sbp = cellfs->ns_sbp;
	u32 major = le32_to_cpu(sbp[0]->s_rev_level);
	u16 minor = le16_to_cpu(sbp[0]->s_minor_rev_level);

	return snprintf(buf, PAGE_SIZE, "%d.%d\n", major, minor);
}

static
ssize_t cellfs_dev_blocksize_show(struct cellfs_dev_attr *attr,
				 struct the_cellfs *cellfs,
				 char *buf)
{
	return snprintf(buf, PAGE_SIZE, "%u\n", cellfs->ns_blocksize);
}

static
ssize_t cellfs_dev_device_size_show(struct cellfs_dev_attr *attr,
				    struct the_cellfs *cellfs,
				    char *buf)
{
	struct cellfs_super_block **sbp = cellfs->ns_sbp;
	u64 dev_size = le64_to_cpu(sbp[0]->s_dev_size);

	return snprintf(buf, PAGE_SIZE, "%llu\n", dev_size);
}

static
ssize_t cellfs_dev_free_blocks_show(struct cellfs_dev_attr *attr,
				   struct the_cellfs *cellfs,
				   char *buf)
{
	sector_t free_blocks = 0;

	cellfs_count_free_blocks(cellfs, &free_blocks);
	return snprintf(buf, PAGE_SIZE, "%llu\n",
			(unsigned long long)free_blocks);
}

static
ssize_t cellfs_dev_uuid_show(struct cellfs_dev_attr *attr,
			    struct the_cellfs *cellfs,
			    char *buf)
{
	struct cellfs_super_block **sbp = cellfs->ns_sbp;

	return snprintf(buf, PAGE_SIZE, "%pUb\n", sbp[0]->s_uuid);
}

static
ssize_t cellfs_dev_volume_name_show(struct cellfs_dev_attr *attr,
				    struct the_cellfs *cellfs,
				    char *buf)
{
	struct cellfs_super_block **sbp = cellfs->ns_sbp;

	return scnprintf(buf, sizeof(sbp[0]->s_volume_name), "%s\n",
			 sbp[0]->s_volume_name);
}

static const char dev_readme_str[] =
	"The <device> group contains attributes that describe file system\n"
	"partition's details.\n\n"
	"(1) revision\n\tshow CELLFS file system revision.\n\n"
	"(2) blocksize\n\tshow volume block size in bytes.\n\n"
	"(3) device_size\n\tshow volume size in bytes.\n\n"
	"(4) free_blocks\n\tshow count of free blocks on volume.\n\n"
	"(5) uuid\n\tshow volume's UUID.\n\n"
	"(6) volume_name\n\tshow volume's name.\n\n";

static ssize_t cellfs_dev_README_show(struct cellfs_dev_attr *attr,
				     struct the_cellfs *cellfs,
				     char *buf)
{
	return snprintf(buf, PAGE_SIZE, dev_readme_str);
}

CELLFS_DEV_RO_ATTR(revision);
CELLFS_DEV_RO_ATTR(blocksize);
CELLFS_DEV_RO_ATTR(device_size);
CELLFS_DEV_RO_ATTR(free_blocks);
CELLFS_DEV_RO_ATTR(uuid);
CELLFS_DEV_RO_ATTR(volume_name);
CELLFS_DEV_RO_ATTR(README);

static struct attribute *cellfs_dev_attrs[] = {
	CELLFS_DEV_ATTR_LIST(revision),
	CELLFS_DEV_ATTR_LIST(blocksize),
	CELLFS_DEV_ATTR_LIST(device_size),
	CELLFS_DEV_ATTR_LIST(free_blocks),
	CELLFS_DEV_ATTR_LIST(uuid),
	CELLFS_DEV_ATTR_LIST(volume_name),
	CELLFS_DEV_ATTR_LIST(README),
	NULL,
};

static ssize_t cellfs_dev_attr_show(struct kobject *kobj,
				    struct attribute *attr, char *buf)
{
	struct the_cellfs *cellfs = container_of(kobj, struct the_cellfs,
						ns_dev_kobj);
	struct cellfs_dev_attr *a = container_of(attr, struct cellfs_dev_attr,
						attr);

	return a->show ? a->show(a, cellfs, buf) : 0;
}

static ssize_t cellfs_dev_attr_store(struct kobject *kobj,
				    struct attribute *attr,
				    const char *buf, size_t len)
{
	struct the_cellfs *cellfs = container_of(kobj, struct the_cellfs,
						ns_dev_kobj);
	struct cellfs_dev_attr *a = container_of(attr, struct cellfs_dev_attr,
						attr);

	return a->store ? a->store(a, cellfs, buf, len) : 0;
}

static void cellfs_dev_attr_release(struct kobject *kobj)
{
	struct the_cellfs *cellfs = container_of(kobj, struct the_cellfs,
						ns_dev_kobj);
	complete(&cellfs->ns_dev_kobj_unregister);
}

static const struct sysfs_ops cellfs_dev_attr_ops = {
	.show	= cellfs_dev_attr_show,
	.store	= cellfs_dev_attr_store,
};

static struct kobj_type cellfs_dev_ktype = {
	.default_attrs	= cellfs_dev_attrs,
	.sysfs_ops	= &cellfs_dev_attr_ops,
	.release	= cellfs_dev_attr_release,
};

int cellfs_sysfs_create_device_group(struct super_block *sb)
{
	struct the_cellfs *cellfs = sb->s_fs_info;
	size_t devgrp_size = sizeof(struct cellfs_sysfs_dev_subgroups);
	int err;

	cellfs->ns_dev_subgroups = kzalloc(devgrp_size, GFP_KERNEL);
	if (unlikely(!cellfs->ns_dev_subgroups)) {
		err = -ENOMEM;
		cellfs_msg(sb, KERN_ERR,
			  "unable to allocate memory for device group");
		goto failed_create_device_group;
	}

	cellfs->ns_dev_kobj.kset = cellfs_kset;
	init_completion(&cellfs->ns_dev_kobj_unregister);
	err = kobject_init_and_add(&cellfs->ns_dev_kobj, &cellfs_dev_ktype, NULL,
				    "%s", sb->s_id);
	if (err)
		goto free_dev_subgroups;

	err = cellfs_sysfs_create_mounted_snapshots_group(cellfs);
	if (err)
		goto cleanup_dev_kobject;

	err = cellfs_sysfs_create_checkpoints_group(cellfs);
	if (err)
		goto delete_mounted_snapshots_group;

	err = cellfs_sysfs_create_segments_group(cellfs);
	if (err)
		goto delete_checkpoints_group;

	err = cellfs_sysfs_create_superblock_group(cellfs);
	if (err)
		goto delete_segments_group;

	err = cellfs_sysfs_create_segctor_group(cellfs);
	if (err)
		goto delete_superblock_group;

	return 0;

delete_superblock_group:
	cellfs_sysfs_delete_superblock_group(cellfs);

delete_segments_group:
	cellfs_sysfs_delete_segments_group(cellfs);

delete_checkpoints_group:
	cellfs_sysfs_delete_checkpoints_group(cellfs);

delete_mounted_snapshots_group:
	cellfs_sysfs_delete_mounted_snapshots_group(cellfs);

cleanup_dev_kobject:
	kobject_del(&cellfs->ns_dev_kobj);

free_dev_subgroups:
	kfree(cellfs->ns_dev_subgroups);

failed_create_device_group:
	return err;
}

void cellfs_sysfs_delete_device_group(struct the_cellfs *cellfs)
{
	cellfs_sysfs_delete_mounted_snapshots_group(cellfs);
	cellfs_sysfs_delete_checkpoints_group(cellfs);
	cellfs_sysfs_delete_segments_group(cellfs);
	cellfs_sysfs_delete_superblock_group(cellfs);
	cellfs_sysfs_delete_segctor_group(cellfs);
	kobject_del(&cellfs->ns_dev_kobj);
	kfree(cellfs->ns_dev_subgroups);
}

/************************************************************************
 *                        CELLFS feature attrs                           *
 ************************************************************************/

static ssize_t cellfs_feature_revision_show(struct kobject *kobj,
					    struct attribute *attr, char *buf)
{
	return snprintf(buf, PAGE_SIZE, "%d.%d\n",
			CELLFS_CURRENT_REV, CELLFS_MINOR_REV);
}

static const char features_readme_str[] =
	"The features group contains attributes that describe CELLFS file\n"
	"system driver features.\n\n"
	"(1) revision\n\tshow current revision of CELLFS file system driver.\n";

static ssize_t cellfs_feature_README_show(struct kobject *kobj,
					 struct attribute *attr,
					 char *buf)
{
	return snprintf(buf, PAGE_SIZE, features_readme_str);
}

CELLFS_FEATURE_RO_ATTR(revision);
CELLFS_FEATURE_RO_ATTR(README);

static struct attribute *cellfs_feature_attrs[] = {
	CELLFS_FEATURE_ATTR_LIST(revision),
	CELLFS_FEATURE_ATTR_LIST(README),
	NULL,
};

static const struct attribute_group cellfs_feature_attr_group = {
	.name = "features",
	.attrs = cellfs_feature_attrs,
};

int __init cellfs_sysfs_init(void)
{
	int err;

	cellfs_kset = kset_create_and_add(CELLFS_ROOT_GROUP_NAME, NULL, fs_kobj);
	if (!cellfs_kset) {
		err = -ENOMEM;
		cellfs_msg(NULL, KERN_ERR,
			  "unable to create sysfs entry: err=%d", err);
		goto failed_sysfs_init;
	}

	err = sysfs_create_group(&cellfs_kset->kobj, &cellfs_feature_attr_group);
	if (unlikely(err)) {
		cellfs_msg(NULL, KERN_ERR,
			  "unable to create feature group: err=%d", err);
		goto cleanup_sysfs_init;
	}

	return 0;

cleanup_sysfs_init:
	kset_unregister(cellfs_kset);

failed_sysfs_init:
	return err;
}

void cellfs_sysfs_exit(void)
{
	sysfs_remove_group(&cellfs_kset->kobj, &cellfs_feature_attr_group);
	kset_unregister(cellfs_kset);
}
