/*
 * file.c - CELLFS regular file handling primitives including fsync().
 *
 * Copyright (C) 2005-2008 Nippon Telegraph and Telephone Corporation.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * Written by Amagai Yoshiji and Ryusuke Konishi.
 */

#include <linux/fs.h>
#include <linux/mm.h>
#include <linux/writeback.h>
#include "kern_feature.h"
#include "cellfs.h"
#include "segment.h"

int cellfs_sync_file(struct file *file, loff_t start, loff_t end, int datasync)
{
	/*
	 * Called from fsync() system call
	 * This is the only entry point that can catch write and synch
	 * timing for both data blocks and intermediate blocks.
	 *
	 * This function should be implemented when the writeback function
	 * will be implemented.
	 */
	struct the_cellfs *cellfs;
	struct inode *inode = file->f_mapping->host;
	int err;

	err = filemap_write_and_wait_range(inode->i_mapping, start, end);
	if (err)
		return err;
	mutex_lock(&inode->i_mutex);

	if (cellfs_inode_dirty(inode)) {
		if (datasync)
			err = cellfs_construct_dsync_segment(inode->i_sb, inode,
							    0, LLONG_MAX);
		else
			err = cellfs_construct_segment(inode->i_sb);
	}
	mutex_unlock(&inode->i_mutex);

	cellfs = inode->i_sb->s_fs_info;
	if (!err)
		err = cellfs_flush_device(cellfs);

	return err;
}

static int cellfs_page_mkwrite(struct vm_area_struct *vma, struct vm_fault *vmf)
{
	struct page *page = vmf->page;
	struct inode *inode = file_inode(vma->vm_file);
	struct cellfs_transaction_info ti;
	int ret = 0;

	if (unlikely(cellfs_near_disk_full(inode->i_sb->s_fs_info)))
		return VM_FAULT_SIGBUS; /* -ENOSPC */

	sb_start_pagefault(inode->i_sb);
	lock_page(page);
	if (page->mapping != inode->i_mapping ||
	    page_offset(page) >= i_size_read(inode) || !PageUptodate(page)) {
		unlock_page(page);
		ret = -EFAULT;	/* make the VM retry the fault */
		goto out;
	}

	/*
	 * check to see if the page is mapped already (no holes)
	 */
	if (PageMappedToDisk(page))
		goto mapped;

	if (page_has_buffers(page)) {
		struct buffer_head *bh, *head;
		int fully_mapped = 1;

		bh = head = page_buffers(page);
		do {
			if (!buffer_mapped(bh)) {
				fully_mapped = 0;
				break;
			}
		} while (bh = bh->b_this_page, bh != head);

		if (fully_mapped) {
			SetPageMappedToDisk(page);
			goto mapped;
		}
	}
	unlock_page(page);

	/*
	 * fill hole blocks
	 */
	ret = cellfs_transaction_begin(inode->i_sb, &ti, 1);
	/* never returns -ENOMEM, but may return -ENOSPC */
	if (unlikely(ret))
		goto out;

#if !HAVE_FILE_UPDATE_TIME_IN_BLOCK_PAGE_MKWRITE
	file_update_time(vma->vm_file);
#endif
	ret = __block_page_mkwrite(vma, vmf, cellfs_get_block);
	if (ret) {
		cellfs_transaction_abort(inode->i_sb);
		goto out;
	}
	cellfs_set_file_dirty(inode, 1 << (PAGE_SHIFT - inode->i_blkbits));
	cellfs_transaction_commit(inode->i_sb);

 mapped:
	wait_for_stable_page(page);
 out:
	sb_end_pagefault(inode->i_sb);
	return block_page_mkwrite_return(ret);
}

static const struct vm_operations_struct cellfs_file_vm_ops = {
	.fault		= filemap_fault,
	.page_mkwrite	= cellfs_page_mkwrite,
#if HAVE_VM_OPS_REMAP_PAGES
	.remap_pages	= generic_file_remap_pages,
#endif
};

static int cellfs_file_mmap(struct file *file, struct vm_area_struct *vma)
{
	file_accessed(file);
	vma->vm_ops = &cellfs_file_vm_ops;
#if HAVE_VM_CAN_NONLINEAR
	vma->vm_flags |= VM_CAN_NONLINEAR;
#endif
	return 0;
}

/*
 * We have mostly NULL's here: the current defaults are ok for
 * the cellfs filesystem.
 */
const struct file_operations cellfs_file_operations = {
	.llseek		= generic_file_llseek,
#if HAVE_IOV_ITER
	.read		= new_sync_read,
	.write		= new_sync_write,
	.read_iter	= generic_file_read_iter,
	.write_iter	= generic_file_write_iter,
#else
	.read		= do_sync_read,
	.write		= do_sync_write,
	.aio_read	= generic_file_aio_read,
	.aio_write	= generic_file_aio_write,
#endif
	.unlocked_ioctl	= cellfs_ioctl,
#ifdef CONFIG_COMPAT
	.compat_ioctl	= cellfs_compat_ioctl,
#endif	/* CONFIG_COMPAT */
	.mmap		= cellfs_file_mmap,
	.open		= generic_file_open,
	/* .release	= cellfs_release_file, */
	.fsync		= cellfs_sync_file,
	.splice_read	= generic_file_splice_read,
};

const struct inode_operations cellfs_file_inode_operations = {
	.setattr	= cellfs_setattr,
	.permission     = cellfs_permission,
	.fiemap		= cellfs_fiemap,
};

/* end of file */
