/*
 * the_cellfs.c - the_cellfs shared structure.
 *
 * Copyright (C) 2005-2008 Nippon Telegraph and Telephone Corporation.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * Written by Ryusuke Konishi.
 *
 */

#include <linux/buffer_head.h>
#include <linux/slab.h>
#include <linux/blkdev.h>
#include <linux/backing-dev.h>
#include <linux/random.h>
#include <linux/crc32.h>
#include "cellfs.h"
#include "segment.h"
#include "alloc.h"
#include "cpfile.h"
#include "sufile.h"
#include "dat.h"
#include "segbuf.h"


static int cellfs_valid_sb(struct cellfs_super_block *sbp);

void cellfs_set_last_segment(struct the_cellfs *cellfs,
			    sector_t start_blocknr, u64 seq, __u64 cno)
{
	spin_lock(&cellfs->ns_last_segment_lock);
	cellfs->ns_last_pseg = start_blocknr;
	cellfs->ns_last_seq = seq;
	cellfs->ns_last_cno = cno;

	if (!cellfs_sb_dirty(cellfs)) {
		if (cellfs->ns_prev_seq == cellfs->ns_last_seq)
			goto stay_cursor;

		set_cellfs_sb_dirty(cellfs);
	}
	cellfs->ns_prev_seq = cellfs->ns_last_seq;

 stay_cursor:
	spin_unlock(&cellfs->ns_last_segment_lock);
}

/**
 * alloc_cellfs - allocate a cellfs object
 * @bdev: block device to which the_cellfs is related
 *
 * Return Value: On success, pointer to the_cellfs is returned.
 * On error, NULL is returned.
 */
struct the_cellfs *alloc_cellfs(struct block_device *bdev)
{
	struct the_cellfs *cellfs;

	cellfs = kzalloc(sizeof(*cellfs), GFP_KERNEL);
	if (!cellfs)
		return NULL;

	cellfs->ns_bdev = bdev;
	atomic_set(&cellfs->ns_ndirtyblks, 0);
	init_rwsem(&cellfs->ns_sem);
	mutex_init(&cellfs->ns_snapshot_mount_mutex);
	INIT_LIST_HEAD(&cellfs->ns_dirty_files);
	INIT_LIST_HEAD(&cellfs->ns_gc_inodes);
	spin_lock_init(&cellfs->ns_inode_lock);
	spin_lock_init(&cellfs->ns_next_gen_lock);
	spin_lock_init(&cellfs->ns_last_segment_lock);
	cellfs->ns_cptree = RB_ROOT;
	spin_lock_init(&cellfs->ns_cptree_lock);
	init_rwsem(&cellfs->ns_segctor_sem);

	return cellfs;
}

/**
 * destroy_cellfs - destroy cellfs object
 * @cellfs: cellfs object to be released
 */
void destroy_cellfs(struct the_cellfs *cellfs)
{
	might_sleep();
	if (cellfs_init(cellfs)) {
		brelse(cellfs->ns_sbh[0]);
		brelse(cellfs->ns_sbh[1]);
	}
	kfree(cellfs);
}

static int cellfs_load_super_root(struct the_cellfs *cellfs,
				 struct super_block *sb, sector_t sr_block)
{
	struct buffer_head *bh_sr;
	struct cellfs_super_root *raw_sr;
	struct cellfs_super_block **sbp = cellfs->ns_sbp;
	struct cellfs_inode *rawi;
	unsigned dat_entry_size, segment_usage_size, checkpoint_size;
	unsigned inode_size;
	int err;

	err = cellfs_read_super_root_block(cellfs, sr_block, &bh_sr, 1);
	if (unlikely(err))
		return err;

	down_read(&cellfs->ns_sem);
	dat_entry_size = le16_to_cpu(sbp[0]->s_dat_entry_size);
	checkpoint_size = le16_to_cpu(sbp[0]->s_checkpoint_size);
	segment_usage_size = le16_to_cpu(sbp[0]->s_segment_usage_size);
	up_read(&cellfs->ns_sem);

	inode_size = cellfs->ns_inode_size;

	rawi = (void *)bh_sr->b_data + CELLFS_SR_DAT_OFFSET(inode_size);
	err = cellfs_dat_read(sb, dat_entry_size, rawi, &cellfs->ns_dat);
	if (err)
		goto failed;

	rawi = (void *)bh_sr->b_data + CELLFS_SR_CPFILE_OFFSET(inode_size);
	err = cellfs_cpfile_read(sb, checkpoint_size, rawi, &cellfs->ns_cpfile);
	if (err)
		goto failed_dat;

	rawi = (void *)bh_sr->b_data + CELLFS_SR_SUFILE_OFFSET(inode_size);
	err = cellfs_sufile_read(sb, segment_usage_size, rawi,
				&cellfs->ns_sufile);
	if (err)
		goto failed_cpfile;

	raw_sr = (struct cellfs_super_root *)bh_sr->b_data;
	cellfs->ns_nongc_ctime = le64_to_cpu(raw_sr->sr_nongc_ctime);

 failed:
	brelse(bh_sr);
	return err;

 failed_cpfile:
	iput(cellfs->ns_cpfile);

 failed_dat:
	iput(cellfs->ns_dat);
	goto failed;
}

static void cellfs_init_recovery_info(struct cellfs_recovery_info *ri)
{
	memset(ri, 0, sizeof(*ri));
	INIT_LIST_HEAD(&ri->ri_used_segments);
}

static void cellfs_clear_recovery_info(struct cellfs_recovery_info *ri)
{
	cellfs_dispose_segment_list(&ri->ri_used_segments);
}

/**
 * cellfs_store_log_cursor - load log cursor from a super block
 * @cellfs: cellfs object
 * @sbp: buffer storing super block to be read
 *
 * cellfs_store_log_cursor() reads the last position of the log
 * containing a super root from a given super block, and initializes
 * relevant information on the cellfs object preparatory for log
 * scanning and recovery.
 */
static int cellfs_store_log_cursor(struct the_cellfs *cellfs,
				  struct cellfs_super_block *sbp)
{
	int ret = 0;

	cellfs->ns_last_pseg = le64_to_cpu(sbp->s_last_pseg);
	cellfs->ns_last_cno = le64_to_cpu(sbp->s_last_cno);
	cellfs->ns_last_seq = le64_to_cpu(sbp->s_last_seq);

	cellfs->ns_prev_seq = cellfs->ns_last_seq;
	cellfs->ns_seg_seq = cellfs->ns_last_seq;
	cellfs->ns_segnum =
		cellfs_get_segnum_of_block(cellfs, cellfs->ns_last_pseg);
	cellfs->ns_cno = cellfs->ns_last_cno + 1;
	if (cellfs->ns_segnum >= cellfs->ns_nsegments) {
		printk(KERN_ERR "CELLFS invalid last segment number.\n");
		ret = -EINVAL;
	}
	return ret;
}

/**
 * load_cellfs - load and recover the cellfs
 * @cellfs: the_cellfs structure to be released
 * @sb: super block isntance used to recover past segment
 *
 * load_cellfs() searches and load the latest super root,
 * attaches the last segment, and does recovery if needed.
 * The caller must call this exclusively for simultaneous mounts.
 */
int load_cellfs(struct the_cellfs *cellfs, struct super_block *sb)
{
	struct cellfs_recovery_info ri;
	unsigned int s_flags = sb->s_flags;
	int really_read_only = bdev_read_only(cellfs->ns_bdev);
	int valid_fs = cellfs_valid_fs(cellfs);
	int err;

	if (!valid_fs) {
		printk(KERN_WARNING "CELLFS warning: mounting unchecked fs\n");
		if (s_flags & MS_RDONLY) {
			printk(KERN_INFO "CELLFS: INFO: recovery "
			       "required for readonly filesystem.\n");
			printk(KERN_INFO "CELLFS: write access will "
			       "be enabled during recovery.\n");
		}
	}

	cellfs_init_recovery_info(&ri);

	err = cellfs_search_super_root(cellfs, &ri);
	if (unlikely(err)) {
		struct cellfs_super_block **sbp = cellfs->ns_sbp;
		int blocksize;

		if (err != -EINVAL)
			goto scan_error;

		if (!cellfs_valid_sb(sbp[1])) {
			printk(KERN_WARNING
			       "CELLFS warning: unable to fall back to spare"
			       "super block\n");
			goto scan_error;
		}
		printk(KERN_INFO
		       "CELLFS: try rollback from an earlier position\n");

		/*
		 * restore super block with its spare and reconfigure
		 * relevant states of the cellfs object.
		 */
		memcpy(sbp[0], sbp[1], cellfs->ns_sbsize);
		cellfs->ns_crc_seed = le32_to_cpu(sbp[0]->s_crc_seed);
		cellfs->ns_sbwtime = le64_to_cpu(sbp[0]->s_wtime);

		/* verify consistency between two super blocks */
		blocksize = BLOCK_SIZE << le32_to_cpu(sbp[0]->s_log_block_size);
		if (blocksize != cellfs->ns_blocksize) {
			printk(KERN_WARNING
			       "CELLFS warning: blocksize differs between "
			       "two super blocks (%d != %d)\n",
			       blocksize, cellfs->ns_blocksize);
			goto scan_error;
		}

		err = cellfs_store_log_cursor(cellfs, sbp[0]);
		if (err)
			goto scan_error;

		/* drop clean flag to allow roll-forward and recovery */
		cellfs->ns_mount_state &= ~CELLFS_VALID_FS;
		valid_fs = 0;

		err = cellfs_search_super_root(cellfs, &ri);
		if (err)
			goto scan_error;
	}

	err = cellfs_load_super_root(cellfs, sb, ri.ri_super_root);
	if (unlikely(err)) {
		printk(KERN_ERR "CELLFS: error loading super root.\n");
		goto failed;
	}

	if (valid_fs)
		goto skip_recovery;

	if (s_flags & MS_RDONLY) {
		__u64 features;

		if (cellfs_test_opt(cellfs, NORECOVERY)) {
			printk(KERN_INFO "CELLFS: norecovery option specified. "
			       "skipping roll-forward recovery\n");
			goto skip_recovery;
		}
		features = le64_to_cpu(cellfs->ns_sbp[0]->s_feature_compat_ro) &
			~CELLFS_FEATURE_COMPAT_RO_SUPP;
		if (features) {
			printk(KERN_ERR "CELLFS: couldn't proceed with "
			       "recovery because of unsupported optional "
			       "features (%llx)\n",
			       (unsigned long long)features);
			err = -EROFS;
			goto failed_unload;
		}
		if (really_read_only) {
			printk(KERN_ERR "CELLFS: write access "
			       "unavailable, cannot proceed.\n");
			err = -EROFS;
			goto failed_unload;
		}
		sb->s_flags &= ~MS_RDONLY;
	} else if (cellfs_test_opt(cellfs, NORECOVERY)) {
		printk(KERN_ERR "CELLFS: recovery cancelled because norecovery "
		       "option was specified for a read/write mount\n");
		err = -EINVAL;
		goto failed_unload;
	}

	err = cellfs_salvage_orphan_logs(cellfs, sb, &ri);
	if (err)
		goto failed_unload;

	down_write(&cellfs->ns_sem);
	cellfs->ns_mount_state |= CELLFS_VALID_FS; /* set "clean" flag */
	err = cellfs_cleanup_super(sb);
	up_write(&cellfs->ns_sem);

	if (err) {
		printk(KERN_ERR "CELLFS: failed to update super block. "
		       "recovery unfinished.\n");
		goto failed_unload;
	}
	printk(KERN_INFO "CELLFS: recovery complete.\n");

 skip_recovery:
	cellfs_clear_recovery_info(&ri);
	sb->s_flags = s_flags;
	return 0;

 scan_error:
	printk(KERN_ERR "CELLFS: error searching super root.\n");
	goto failed;

 failed_unload:
	iput(cellfs->ns_cpfile);
	iput(cellfs->ns_sufile);
	iput(cellfs->ns_dat);

 failed:
	cellfs_clear_recovery_info(&ri);
	sb->s_flags = s_flags;
	return err;
}

static unsigned long long cellfs_max_size(unsigned int blkbits)
{
	unsigned int max_bits;
	unsigned long long res = MAX_LFS_FILESIZE; /* page cache limit */

	max_bits = blkbits + CELLFS_BMAP_KEY_BIT; /* bmap size limit */
	if (max_bits < 64)
		res = min_t(unsigned long long, res, (1ULL << max_bits) - 1);
	return res;
}

/**
 * cellfs_nrsvsegs - calculate the number of reserved segments
 * @cellfs: cellfs object
 * @nsegs: total number of segments
 */
unsigned long cellfs_nrsvsegs(struct the_cellfs *cellfs, unsigned long nsegs)
{
	return max_t(unsigned long, CELLFS_MIN_NRSVSEGS,
		     DIV_ROUND_UP(nsegs * cellfs->ns_r_segments_percentage,
				  100));
}

void cellfs_set_nsegments(struct the_cellfs *cellfs, unsigned long nsegs)
{
	cellfs->ns_nsegments = nsegs;
	cellfs->ns_nrsvsegs = cellfs_nrsvsegs(cellfs, nsegs);
}

static int cellfs_store_disk_layout(struct the_cellfs *cellfs,
				   struct cellfs_super_block *sbp)
{
	if (le32_to_cpu(sbp->s_rev_level) < CELLFS_MIN_SUPP_REV) {
		printk(KERN_ERR "CELLFS: unsupported revision "
		       "(superblock rev.=%d.%d, current rev.=%d.%d). "
		       "Please check the version of mkfs.cellfs.\n",
		       le32_to_cpu(sbp->s_rev_level),
		       le16_to_cpu(sbp->s_minor_rev_level),
		       CELLFS_CURRENT_REV, CELLFS_MINOR_REV);
		return -EINVAL;
	}
	cellfs->ns_sbsize = le16_to_cpu(sbp->s_bytes);
	if (cellfs->ns_sbsize > BLOCK_SIZE)
		return -EINVAL;

	cellfs->ns_inode_size = le16_to_cpu(sbp->s_inode_size);
	if (cellfs->ns_inode_size > cellfs->ns_blocksize) {
		printk(KERN_ERR "CELLFS: too large inode size: %d bytes.\n",
		       cellfs->ns_inode_size);
		return -EINVAL;
	} else if (cellfs->ns_inode_size < CELLFS_MIN_INODE_SIZE) {
		printk(KERN_ERR "CELLFS: too small inode size: %d bytes.\n",
		       cellfs->ns_inode_size);
		return -EINVAL;
	}

	cellfs->ns_first_ino = le32_to_cpu(sbp->s_first_ino);

	cellfs->ns_blocks_per_segment = le32_to_cpu(sbp->s_blocks_per_segment);
	if (cellfs->ns_blocks_per_segment < CELLFS_SEG_MIN_BLOCKS) {
		printk(KERN_ERR "CELLFS: too short segment.\n");
		return -EINVAL;
	}

	cellfs->ns_first_data_block = le64_to_cpu(sbp->s_first_data_block);
	cellfs->ns_r_segments_percentage =
		le32_to_cpu(sbp->s_r_segments_percentage);
	if (cellfs->ns_r_segments_percentage < 1 ||
	    cellfs->ns_r_segments_percentage > 99) {
		printk(KERN_ERR "CELLFS: invalid reserved segments percentage.\n");
		return -EINVAL;
	}

	cellfs_set_nsegments(cellfs, le64_to_cpu(sbp->s_nsegments));
	cellfs->ns_crc_seed = le32_to_cpu(sbp->s_crc_seed);
	return 0;
}

static int cellfs_valid_sb(struct cellfs_super_block *sbp)
{
	static unsigned char sum[4];
	const int sumoff = offsetof(struct cellfs_super_block, s_sum);
	size_t bytes;
	u32 crc;

	if (!sbp || le16_to_cpu(sbp->s_magic) != CELLFS_SUPER_MAGIC)
		return 0;
	bytes = le16_to_cpu(sbp->s_bytes);
	if (bytes < sumoff + 4 || bytes > BLOCK_SIZE)
		return 0;
	crc = crc32_le(le32_to_cpu(sbp->s_crc_seed), (unsigned char *)sbp,
		       sumoff);
	crc = crc32_le(crc, sum, 4);
	crc = crc32_le(crc, (unsigned char *)sbp + sumoff + 4,
		       bytes - sumoff - 4);
	return crc == le32_to_cpu(sbp->s_sum);
}

static int cellfs_sb2_bad_offset(struct cellfs_super_block *sbp, u64 offset)
{
	return offset < ((le64_to_cpu(sbp->s_nsegments) *
			  le32_to_cpu(sbp->s_blocks_per_segment)) <<
			 (le32_to_cpu(sbp->s_log_block_size) + 10));
}

static void cellfs_release_super_block(struct the_cellfs *cellfs)
{
	int i;

	for (i = 0; i < 2; i++) {
		if (cellfs->ns_sbp[i]) {
			brelse(cellfs->ns_sbh[i]);
			cellfs->ns_sbh[i] = NULL;
			cellfs->ns_sbp[i] = NULL;
		}
	}
}

void cellfs_fall_back_super_block(struct the_cellfs *cellfs)
{
	brelse(cellfs->ns_sbh[0]);
	cellfs->ns_sbh[0] = cellfs->ns_sbh[1];
	cellfs->ns_sbp[0] = cellfs->ns_sbp[1];
	cellfs->ns_sbh[1] = NULL;
	cellfs->ns_sbp[1] = NULL;
}

void cellfs_swap_super_block(struct the_cellfs *cellfs)
{
	struct buffer_head *tsbh = cellfs->ns_sbh[0];
	struct cellfs_super_block *tsbp = cellfs->ns_sbp[0];

	cellfs->ns_sbh[0] = cellfs->ns_sbh[1];
	cellfs->ns_sbp[0] = cellfs->ns_sbp[1];
	cellfs->ns_sbh[1] = tsbh;
	cellfs->ns_sbp[1] = tsbp;
}

static int cellfs_load_super_block(struct the_cellfs *cellfs,
				  struct super_block *sb, int blocksize,
				  struct cellfs_super_block **sbpp)
{
	struct cellfs_super_block **sbp = cellfs->ns_sbp;
	struct buffer_head **sbh = cellfs->ns_sbh;
	u64 sb2off = CELLFS_SB2_OFFSET_BYTES(cellfs->ns_bdev->bd_inode->i_size);
	int valid[2], swp = 0;

	sbp[0] = cellfs_read_super_block(sb, CELLFS_SB_OFFSET_BYTES, blocksize,
					&sbh[0]);
	sbp[1] = cellfs_read_super_block(sb, sb2off, blocksize, &sbh[1]);

	if (!sbp[0]) {
		if (!sbp[1]) {
			printk(KERN_ERR "CELLFS: unable to read superblock\n");
			return -EIO;
		}
		printk(KERN_WARNING
		       "CELLFS warning: unable to read primary superblock "
		       "(blocksize = %d)\n", blocksize);
	} else if (!sbp[1]) {
		printk(KERN_WARNING
		       "CELLFS warning: unable to read secondary superblock "
		       "(blocksize = %d)\n", blocksize);
	}

	/*
	 * Compare two super blocks and set 1 in swp if the secondary
	 * super block is valid and newer.  Otherwise, set 0 in swp.
	 */
	valid[0] = cellfs_valid_sb(sbp[0]);
	valid[1] = cellfs_valid_sb(sbp[1]);
	swp = valid[1] && (!valid[0] ||
			   le64_to_cpu(sbp[1]->s_last_cno) >
			   le64_to_cpu(sbp[0]->s_last_cno));

	if (valid[swp] && cellfs_sb2_bad_offset(sbp[swp], sb2off)) {
		brelse(sbh[1]);
		sbh[1] = NULL;
		sbp[1] = NULL;
		valid[1] = 0;
		swp = 0;
	}
	if (!valid[swp]) {
		cellfs_release_super_block(cellfs);
		printk(KERN_ERR "CELLFS: Can't find cellfs on dev %s.\n",
		       sb->s_id);
		return -EINVAL;
	}

	if (!valid[!swp])
		printk(KERN_WARNING "CELLFS warning: broken superblock. "
		       "using spare superblock (blocksize = %d).\n", blocksize);
	if (swp)
		cellfs_swap_super_block(cellfs);

	cellfs->ns_sbwcount = 0;
	cellfs->ns_sbwtime = le64_to_cpu(sbp[0]->s_wtime);
	cellfs->ns_prot_seq = le64_to_cpu(sbp[valid[1] & !swp]->s_last_seq);
	*sbpp = sbp[0];
	return 0;
}

/**
 * init_cellfs - initialize a CELLFS instance.
 * @cellfs: the_cellfs structure
 * @sb: super block
 * @data: mount options
 *
 * init_cellfs() performs common initialization per block device (e.g.
 * reading the super block, getting disk layout information, initializing
 * shared fields in the_cellfs).
 *
 * Return Value: On success, 0 is returned. On error, a negative error
 * code is returned.
 */
int init_cellfs(struct the_cellfs *cellfs, struct super_block *sb, char *data)
{
	struct cellfs_super_block *sbp;
	int blocksize;
	int err;

	down_write(&cellfs->ns_sem);

	blocksize = sb_min_blocksize(sb, CELLFS_MIN_BLOCK_SIZE);
	if (!blocksize) {
		printk(KERN_ERR "CELLFS: unable to set blocksize\n");
		err = -EINVAL;
		goto out;
	}
	err = cellfs_load_super_block(cellfs, sb, blocksize, &sbp);
	if (err)
		goto out;

	err = cellfs_store_magic_and_option(sb, sbp, data);
	if (err)
		goto failed_sbh;

	err = cellfs_check_feature_compatibility(sb, sbp);
	if (err)
		goto failed_sbh;

	blocksize = BLOCK_SIZE << le32_to_cpu(sbp->s_log_block_size);
	if (blocksize < CELLFS_MIN_BLOCK_SIZE ||
	    blocksize > CELLFS_MAX_BLOCK_SIZE) {
		printk(KERN_ERR "CELLFS: couldn't mount because of unsupported "
		       "filesystem blocksize %d\n", blocksize);
		err = -EINVAL;
		goto failed_sbh;
	}
	if (sb->s_blocksize != blocksize) {
		int hw_blocksize = bdev_logical_block_size(sb->s_bdev);

		if (blocksize < hw_blocksize) {
			printk(KERN_ERR
			       "CELLFS: blocksize %d too small for device "
			       "(sector-size = %d).\n",
			       blocksize, hw_blocksize);
			err = -EINVAL;
			goto failed_sbh;
		}
		cellfs_release_super_block(cellfs);
		sb_set_blocksize(sb, blocksize);

		err = cellfs_load_super_block(cellfs, sb, blocksize, &sbp);
		if (err)
			goto out;
			/* not failed_sbh; sbh is released automatically
			   when reloading fails. */
	}
	cellfs->ns_blocksize_bits = sb->s_blocksize_bits;
	cellfs->ns_blocksize = blocksize;

	get_random_bytes(&cellfs->ns_next_generation,
			 sizeof(cellfs->ns_next_generation));

	err = cellfs_store_disk_layout(cellfs, sbp);
	if (err)
		goto failed_sbh;

	sb->s_maxbytes = cellfs_max_size(sb->s_blocksize_bits);

	cellfs->ns_mount_state = le16_to_cpu(sbp->s_state);

	err = cellfs_store_log_cursor(cellfs, sbp);
	if (err)
		goto failed_sbh;

	set_cellfs_init(cellfs);
	err = 0;
 out:
	up_write(&cellfs->ns_sem);
	return err;

 failed_sbh:
	cellfs_release_super_block(cellfs);
	goto out;
}

int cellfs_discard_segments(struct the_cellfs *cellfs, __u64 *segnump,
			    size_t nsegs)
{
	sector_t seg_start, seg_end;
	sector_t start = 0, nblocks = 0;
	unsigned int sects_per_block;
	__u64 *sn;
	int ret = 0;

	sects_per_block = (1 << cellfs->ns_blocksize_bits) /
		bdev_logical_block_size(cellfs->ns_bdev);
	for (sn = segnump; sn < segnump + nsegs; sn++) {
		cellfs_get_segment_range(cellfs, *sn, &seg_start, &seg_end);

		if (!nblocks) {
			start = seg_start;
			nblocks = seg_end - seg_start + 1;
		} else if (start + nblocks == seg_start) {
			nblocks += seg_end - seg_start + 1;
		} else {
			ret = blkdev_issue_discard(cellfs->ns_bdev,
						   start * sects_per_block,
						   nblocks * sects_per_block,
						   GFP_NOFS, 0);
			if (ret < 0)
				return ret;
			nblocks = 0;
		}
	}
	if (nblocks)
		ret = blkdev_issue_discard(cellfs->ns_bdev,
					   start * sects_per_block,
					   nblocks * sects_per_block,
					   GFP_NOFS, 0);
	return ret;
}

int cellfs_count_free_blocks(struct the_cellfs *cellfs, sector_t *nblocks)
{
	unsigned long ncleansegs;

	down_read(&CELLFS_MDT(cellfs->ns_dat)->mi_sem);
	ncleansegs = cellfs_sufile_get_ncleansegs(cellfs->ns_sufile);
	up_read(&CELLFS_MDT(cellfs->ns_dat)->mi_sem);
	*nblocks = (sector_t)ncleansegs * cellfs->ns_blocks_per_segment;
	return 0;
}

int cellfs_near_disk_full(struct the_cellfs *cellfs)
{
	unsigned long ncleansegs, nincsegs;

	ncleansegs = cellfs_sufile_get_ncleansegs(cellfs->ns_sufile);
	nincsegs = atomic_read(&cellfs->ns_ndirtyblks) /
		cellfs->ns_blocks_per_segment + 1;

	return ncleansegs <= cellfs->ns_nrsvsegs + nincsegs;
}

struct cellfs_root *cellfs_lookup_root(struct the_cellfs *cellfs, __u64 cno)
{
	struct rb_node *n;
	struct cellfs_root *root;

	spin_lock(&cellfs->ns_cptree_lock);
	n = cellfs->ns_cptree.rb_node;
	while (n) {
		root = rb_entry(n, struct cellfs_root, rb_node);

		if (cno < root->cno) {
			n = n->rb_left;
		} else if (cno > root->cno) {
			n = n->rb_right;
		} else {
			atomic_inc(&root->count);
			spin_unlock(&cellfs->ns_cptree_lock);
			return root;
		}
	}
	spin_unlock(&cellfs->ns_cptree_lock);

	return NULL;
}

struct cellfs_root *
cellfs_find_or_create_root(struct the_cellfs *cellfs, __u64 cno)
{
	struct rb_node **p, *parent;
	struct cellfs_root *root, *new;

	root = cellfs_lookup_root(cellfs, cno);
	if (root)
		return root;

	new = kmalloc(sizeof(*root), GFP_KERNEL);
	if (!new)
		return NULL;

	spin_lock(&cellfs->ns_cptree_lock);

	p = &cellfs->ns_cptree.rb_node;
	parent = NULL;

	while (*p) {
		parent = *p;
		root = rb_entry(parent, struct cellfs_root, rb_node);

		if (cno < root->cno) {
			p = &(*p)->rb_left;
		} else if (cno > root->cno) {
			p = &(*p)->rb_right;
		} else {
			atomic_inc(&root->count);
			spin_unlock(&cellfs->ns_cptree_lock);
			kfree(new);
			return root;
		}
	}

	new->cno = cno;
	new->ifile = NULL;
	new->cellfs = cellfs;
	atomic_set(&new->count, 1);
	atomic64_set(&new->inodes_count, 0);
	atomic64_set(&new->blocks_count, 0);

	rb_link_node(&new->rb_node, parent, p);
	rb_insert_color(&new->rb_node, &cellfs->ns_cptree);

	spin_unlock(&cellfs->ns_cptree_lock);

	return new;
}

void cellfs_put_root(struct cellfs_root *root)
{
	if (atomic_dec_and_test(&root->count)) {
		struct the_cellfs *cellfs = root->cellfs;

		spin_lock(&cellfs->ns_cptree_lock);
		rb_erase(&root->rb_node, &cellfs->ns_cptree);
		spin_unlock(&cellfs->ns_cptree_lock);
		if (root->ifile)
			iput(root->ifile);

		kfree(root);
	}
}
