/*
 * the_cellfs.h - the_cellfs shared structure.
 *
 * Copyright (C) 2005-2008 Nippon Telegraph and Telephone Corporation.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * Written by Ryusuke Konishi.
 *
 */

#ifndef _THE_CELLFS_H
#define _THE_CELLFS_H

#include <linux/types.h>
#include <linux/buffer_head.h>
#include <linux/rbtree.h>
#include <linux/fs.h>
#include <linux/blkdev.h>
#include <linux/backing-dev.h>
#include <linux/slab.h>

struct cellfs_sc_info;

/* the_cellfs struct */
enum {
	THE_CELLFS_INIT = 0,     /* Information from super_block is set */
	THE_CELLFS_DISCONTINUED,	/* 'next' pointer chain has broken */
	THE_CELLFS_GC_RUNNING,	/* gc process is running */
	THE_CELLFS_SB_DIRTY,	/* super block is dirty */
};

/**
 * struct the_cellfs - struct to supervise multiple cellfs mount points
 * @ns_flags: flags
 * @ns_flushed_device: flag indicating if all volatile data was flushed
 * @ns_bdev: block device
 * @ns_sem: semaphore for shared states
 * @ns_snapshot_mount_mutex: mutex to protect snapshot mounts
 * @ns_sbh: buffer heads of on-disk super blocks
 * @ns_sbp: pointers to super block data
 * @ns_sbwtime: previous write time of super block
 * @ns_sbwcount: write count of super block
 * @ns_sbsize: size of valid data in super block
 * @ns_mount_state: file system state
 * @ns_seg_seq: segment sequence counter
 * @ns_segnum: index number of the latest full segment.
 * @ns_nextnum: index number of the full segment index to be used next
 * @ns_pseg_offset: offset of next partial segment in the current full segment
 * @ns_cno: next checkpoint number
 * @ns_ctime: write time of the last segment
 * @ns_nongc_ctime: write time of the last segment not for cleaner operation
 * @ns_ndirtyblks: Number of dirty data blocks
 * @ns_last_segment_lock: lock protecting fields for the latest segment
 * @ns_last_pseg: start block number of the latest segment
 * @ns_last_seq: sequence value of the latest segment
 * @ns_last_cno: checkpoint number of the latest segment
 * @ns_prot_seq: least sequence number of segments which must not be reclaimed
 * @ns_prev_seq: base sequence number used to decide if advance log cursor
 * @ns_writer: log writer
 * @ns_segctor_sem: semaphore protecting log write
 * @ns_dat: DAT file inode
 * @ns_cpfile: checkpoint file inode
 * @ns_sufile: segusage file inode
 * @ns_cptree: rb-tree of all mounted checkpoints (cellfs_root)
 * @ns_cptree_lock: lock protecting @ns_cptree
 * @ns_dirty_files: list of dirty files
 * @ns_inode_lock: lock protecting @ns_dirty_files
 * @ns_gc_inodes: dummy inodes to keep live blocks
 * @ns_next_generation: next generation number for inodes
 * @ns_next_gen_lock: lock protecting @ns_next_generation
 * @ns_mount_opt: mount options
 * @ns_resuid: uid for reserved blocks
 * @ns_resgid: gid for reserved blocks
 * @ns_interval: checkpoint creation interval
 * @ns_watermark: watermark for the number of dirty buffers
 * @ns_blocksize_bits: bit length of block size
 * @ns_blocksize: block size
 * @ns_nsegments: number of segments in filesystem
 * @ns_blocks_per_segment: number of blocks per segment
 * @ns_r_segments_percentage: reserved segments percentage
 * @ns_nrsvsegs: number of reserved segments
 * @ns_first_data_block: block number of first data block
 * @ns_inode_size: size of on-disk inode
 * @ns_first_ino: first not-special inode number
 * @ns_crc_seed: seed value of CRC32 calculation
 */
struct the_cellfs {
	unsigned long		ns_flags;
	int			ns_flushed_device;

	struct block_device    *ns_bdev;
	struct rw_semaphore	ns_sem;
	struct mutex		ns_snapshot_mount_mutex;

	/*
	 * used for
	 * - loading the latest checkpoint exclusively.
	 * - allocating a new full segment.
	 */
	struct buffer_head     *ns_sbh[2];
	struct cellfs_super_block *ns_sbp[2];
	time_t			ns_sbwtime;
	unsigned		ns_sbwcount;
	unsigned		ns_sbsize;
	unsigned		ns_mount_state;

	/*
	 * Following fields are dedicated to a writable FS-instance.
	 * Except for the period seeking checkpoint, code outside the segment
	 * constructor must lock a segment semaphore while accessing these
	 * fields.
	 * The writable FS-instance is sole during a lifetime of the_cellfs.
	 */
	u64			ns_seg_seq;
	__u64			ns_segnum;
	__u64			ns_nextnum;
	unsigned long		ns_pseg_offset;
	__u64			ns_cno;
	time_t			ns_ctime;
	time_t			ns_nongc_ctime;
	atomic_t		ns_ndirtyblks;

	/*
	 * The following fields hold information on the latest partial segment
	 * written to disk with a super root.  These fields are protected by
	 * ns_last_segment_lock.
	 */
	spinlock_t		ns_last_segment_lock;
	sector_t		ns_last_pseg;
	u64			ns_last_seq;
	__u64			ns_last_cno;
	u64			ns_prot_seq;
	u64			ns_prev_seq;

	struct cellfs_sc_info   *ns_writer;
	struct rw_semaphore	ns_segctor_sem;

	/*
	 * Following fields are lock free except for the period before
	 * the_cellfs is initialized.
	 */
	struct inode	       *ns_dat;
	struct inode	       *ns_cpfile;
	struct inode	       *ns_sufile;

	/* Checkpoint tree */
	struct rb_root		ns_cptree;
	spinlock_t		ns_cptree_lock;

	/* Dirty inode list */
	struct list_head	ns_dirty_files;
	spinlock_t		ns_inode_lock;

	/* GC inode list */
	struct list_head	ns_gc_inodes;

	/* Inode allocator */
	u32			ns_next_generation;
	spinlock_t		ns_next_gen_lock;

	/* Mount options */
	unsigned long		ns_mount_opt;

	uid_t			ns_resuid;
	gid_t			ns_resgid;
	unsigned long		ns_interval;
	unsigned long		ns_watermark;

	/* Disk layout information (static) */
	unsigned int		ns_blocksize_bits;
	unsigned int		ns_blocksize;
	unsigned long		ns_nsegments;
	unsigned long		ns_blocks_per_segment;
	unsigned long		ns_r_segments_percentage;
	unsigned long		ns_nrsvsegs;
	unsigned long		ns_first_data_block;
	int			ns_inode_size;
	int			ns_first_ino;
	u32			ns_crc_seed;
};

#define THE_CELLFS_FNS(bit, name)					\
static inline void set_cellfs_##name(struct the_cellfs *cellfs)		\
{									\
	set_bit(THE_CELLFS_##bit, &(cellfs)->ns_flags);			\
}									\
static inline void clear_cellfs_##name(struct the_cellfs *cellfs)		\
{									\
	clear_bit(THE_CELLFS_##bit, &(cellfs)->ns_flags);			\
}									\
static inline int cellfs_##name(struct the_cellfs *cellfs)			\
{									\
	return test_bit(THE_CELLFS_##bit, &(cellfs)->ns_flags);		\
}

THE_CELLFS_FNS(INIT, init)
THE_CELLFS_FNS(DISCONTINUED, discontinued)
THE_CELLFS_FNS(GC_RUNNING, gc_running)
THE_CELLFS_FNS(SB_DIRTY, sb_dirty)

/*
 * Mount option operations
 */
#define cellfs_clear_opt(cellfs, opt)  \
	((cellfs)->ns_mount_opt &= ~CELLFS_MOUNT_##opt)
#define cellfs_set_opt(cellfs, opt)  \
	((cellfs)->ns_mount_opt |= CELLFS_MOUNT_##opt)
#define cellfs_test_opt(cellfs, opt) ((cellfs)->ns_mount_opt & CELLFS_MOUNT_##opt)
#define cellfs_write_opt(cellfs, mask, opt)				\
	((cellfs)->ns_mount_opt =					\
		(((cellfs)->ns_mount_opt & ~CELLFS_MOUNT_##mask) |	\
		 CELLFS_MOUNT_##opt))					\

/**
 * struct cellfs_root - cellfs root object
 * @cno: checkpoint number
 * @rb_node: red-black tree node
 * @count: refcount of this structure
 * @cellfs: cellfs object
 * @ifile: inode file
 * @inodes_count: number of inodes
 * @blocks_count: number of blocks
 */
struct cellfs_root {
	__u64 cno;
	struct rb_node rb_node;

	atomic_t count;
	struct the_cellfs *cellfs;
	struct inode *ifile;

	atomic64_t inodes_count;
	atomic64_t blocks_count;
};

/* Special checkpoint number */
#define CELLFS_CPTREE_CURRENT_CNO	0

/* Minimum interval of periodical update of superblocks (in seconds) */
#define CELLFS_SB_FREQ		10

static inline int cellfs_sb_need_update(struct the_cellfs *cellfs)
{
	u64 t = get_seconds();

	return t < cellfs->ns_sbwtime || t > cellfs->ns_sbwtime + CELLFS_SB_FREQ;
}

static inline int cellfs_sb_will_flip(struct the_cellfs *cellfs)
{
	int flip_bits = cellfs->ns_sbwcount & 0x0FL;

	return (flip_bits != 0x08 && flip_bits != 0x0F);
}

void cellfs_set_last_segment(struct the_cellfs *, sector_t, u64, __u64);
struct the_cellfs *alloc_cellfs(struct block_device *bdev);
void destroy_cellfs(struct the_cellfs *cellfs);
int init_cellfs(struct the_cellfs *cellfs, struct super_block *sb, char *data);
int load_cellfs(struct the_cellfs *cellfs, struct super_block *sb);
unsigned long cellfs_nrsvsegs(struct the_cellfs *cellfs, unsigned long nsegs);
void cellfs_set_nsegments(struct the_cellfs *cellfs, unsigned long nsegs);
int cellfs_discard_segments(struct the_cellfs *, __u64 *, size_t);
int cellfs_count_free_blocks(struct the_cellfs *, sector_t *);
struct cellfs_root *cellfs_lookup_root(struct the_cellfs *cellfs, __u64 cno);
struct cellfs_root *cellfs_find_or_create_root(struct the_cellfs *cellfs,
					     __u64 cno);
void cellfs_put_root(struct cellfs_root *root);
int cellfs_near_disk_full(struct the_cellfs *);
void cellfs_fall_back_super_block(struct the_cellfs *);
void cellfs_swap_super_block(struct the_cellfs *);


static inline void cellfs_get_root(struct cellfs_root *root)
{
	atomic_inc(&root->count);
}

static inline int cellfs_valid_fs(struct the_cellfs *cellfs)
{
	unsigned valid_fs;

	down_read(&cellfs->ns_sem);
	valid_fs = (cellfs->ns_mount_state & CELLFS_VALID_FS);
	up_read(&cellfs->ns_sem);
	return valid_fs;
}

static inline void
cellfs_get_segment_range(struct the_cellfs *cellfs, __u64 segnum,
			sector_t *seg_start, sector_t *seg_end)
{
	*seg_start = (sector_t)cellfs->ns_blocks_per_segment * segnum;
	*seg_end = *seg_start + cellfs->ns_blocks_per_segment - 1;
	if (segnum == 0)
		*seg_start = cellfs->ns_first_data_block;
}

static inline sector_t
cellfs_get_segment_start_blocknr(struct the_cellfs *cellfs, __u64 segnum)
{
	return (segnum == 0) ? cellfs->ns_first_data_block :
		(sector_t)cellfs->ns_blocks_per_segment * segnum;
}

static inline __u64
cellfs_get_segnum_of_block(struct the_cellfs *cellfs, sector_t blocknr)
{
	sector_t segnum = blocknr;

	sector_div(segnum, cellfs->ns_blocks_per_segment);
	return segnum;
}

static inline void
cellfs_terminate_segment(struct the_cellfs *cellfs, sector_t seg_start,
			sector_t seg_end)
{
	/* terminate the current full segment (used in case of I/O-error) */
	cellfs->ns_pseg_offset = seg_end - seg_start + 1;
}

static inline void cellfs_shift_to_next_segment(struct the_cellfs *cellfs)
{
	/* move forward with a full segment */
	cellfs->ns_segnum = cellfs->ns_nextnum;
	cellfs->ns_pseg_offset = 0;
	cellfs->ns_seg_seq++;
}

static inline __u64 cellfs_last_cno(struct the_cellfs *cellfs)
{
	__u64 cno;

	spin_lock(&cellfs->ns_last_segment_lock);
	cno = cellfs->ns_last_cno;
	spin_unlock(&cellfs->ns_last_segment_lock);
	return cno;
}

static inline int cellfs_segment_is_active(struct the_cellfs *cellfs, __u64 n)
{
	return n == cellfs->ns_segnum || n == cellfs->ns_nextnum;
}

static inline int cellfs_flush_device(struct the_cellfs *cellfs)
{
	int err;

	if (!cellfs_test_opt(cellfs, BARRIER) || cellfs->ns_flushed_device)
		return 0;

	cellfs->ns_flushed_device = 1;
	/*
	 * the store to ns_flushed_device must not be reordered after
	 * blkdev_issue_flush().
	 */
	smp_wmb();

	err = blkdev_issue_flush(cellfs->ns_bdev, GFP_KERNEL, NULL);
	if (err != -EIO)
		err = 0;
	return err;
}

#endif /* _THE_CELLFS_H */
